// Copyright (C) 1997 by Ritual Entertainment, Inc.

// All rights reserved.

//

// This source is may not be distributed and/or modified without

// expressly written permission by Ritual Entertainment, Inc.

// 

// DESCRIPTION:

// Subclass of script that preprocesses labels

//



#include "g_local.h"

#include "script.h"

#include "gamescript.h"



ScriptLibrarian ScriptLib;



CLASS_DECLARATION( Class, GameScriptMarker, NULL );



ResponseDef GameScriptMarker::Responses[] =

	{

		{ NULL, NULL }

	};



CLASS_DECLARATION( Class, ScriptLibrarian, NULL );



ResponseDef ScriptLibrarian::Responses[] =

	{

		{ NULL, NULL }

	};



ScriptLibrarian::~ScriptLibrarian()

	{

	int i;

	int num;



   num = scripts.NumObjects();

   for( i = 1; i <= num; i++ )

      {

      delete scripts.ObjectAt( i );

      }

	}



void ScriptLibrarian::CloseScripts

	(

	void

	)



	{

	int i;

	int num;

	GameScript *scr;



   // Clear out the game and dialog scripts

   SetGameScript( "" );

   SetDialogScript( "" );



   num = scripts.NumObjects();

   for( i = num; i > 0; i-- )

      {

		scr = scripts.ObjectAt( i );

		scripts.RemoveObjectAt( i );

      delete scr;

      }

	}



void ScriptLibrarian::SetDialogScript

	(

   Burger::String scriptname

	)

   {

   dialog_script = scriptname;

   }



void ScriptLibrarian::SetGameScript

	(

   Burger::String scriptname

	)

   {

   game_script = scriptname;

   }



const char *ScriptLibrarian::GetGameScript

	(

   void

	)

   {

   return game_script.c_str();

   }



GameScript *ScriptLibrarian::FindScript

	(

	const char *name

	)



	{

	int i;

	int num;

	GameScript *scr;

	Burger::String n;



	// Convert all forward slashes to back slashes

	n = name;

   G_FixSlashes(n.c_str());



   num = scripts.NumObjects();

   for( i = 1; i <= num; i++ )

      {

		scr = scripts.ObjectAt( i );



		if ( scr->Filename() == n )

			{

			return scr;

			}

      }



   return NULL;

	}



GameScript *ScriptLibrarian::GetScript

	(

	const char *name

	)



	{

	GameScript *scr;

   Burger::String n;



	n=name;

   G_FixSlashes( n.c_str());

   scr = FindScript( n.c_str() );

   if ( !scr && ( gi.LoadFile( name, NULL, 0 ) != -1 ) )

		{

		scr = new GameScript();

		scr->LoadFile( n.c_str() );

		scripts.AddObject( scr );

		}

	

	return scr;

	}



qboolean ScriptLibrarian::Goto(GameScript *scr,const char *name)

{

	const char *p;

	GameScript *s;

	const char *TextPtr;

	qboolean Result;

	

	p = strstr( name, "::" );

	if (!p) {

		return scr->Goto(name);

	}

	Burger::String n( name, 0, p - name );

	if (!strcmp(n.c_str(),"dialog")) {

		TextPtr = dialog_script.c_str();

	} else {

		TextPtr = n.c_str();

	}

	s = GetScript(TextPtr);

	Result = FALSE;

	if (s) {

		p += 2;

		if ( s->labelExists( p ) ) {

			scr->SetSourceScript( s );

			Result = scr->Goto(p);

		}

	}

	return Result;

}

	

qboolean ScriptLibrarian::labelExists(GameScript *scr,const char *name)

{

	const char *p;

	GameScript *s;

	Burger::String n;



	p = strstr( name, "::" );

	if ( !p )

		{

		return scr->labelExists( name );

		}

	else

		{

		n = Burger::String( name, 0, p - name );

      if ( n == Burger::String( "dialog" ) )

         {

         n = dialog_script;

         }

		s = GetScript( n.c_str() );

		if ( !s )

			{

			return FALSE;

			}



		p += 2;

		return s->labelExists( p );

		}



	return FALSE;

	}



CLASS_DECLARATION( Script, GameScript, NULL );



ResponseDef GameScript::Responses[] =

	{

		{ NULL, NULL }

	};



GameScript::GameScript()

	{

	sourcescript = this;

	labelList = NULL;

   crc = 0;

	}



GameScript::GameScript

	(

	GameScript *scr

	)



	{

   crc = 0;

	labelList = NULL;

	SetSourceScript( scr );

	}



GameScript::~GameScript()

	{

	Close();

	}



void GameScript::Close

	(

	void

	)



	{

	FreeLabels();

	Script::Close();

	sourcescript = this;

   crc = 0;

	}



void GameScript::SetSourceScript

	(

	GameScript *scr 

	)



	{

	if ( scr != this )

		{

		Close();



		sourcescript = scr->sourcescript;

      crc = sourcescript->crc;

		Parse( scr->buffer, scr->length, scr->Filename() );

		}

	}



void GameScript::FreeLabels

	(

	void

	)



	{

   int i;

	int num;



	if ( labelList )

		{

		num = labelList->NumObjects();

		for( i = 1; i <= num; i++ )

			{

			delete labelList->ObjectAt( i );

			}



		delete labelList;

		}



	labelList = NULL;

	}



void GameScript::LoadFile(const char *name)

{

	Burger::String n;

	

	// Convert all forward slashes to back slashes

	n = name;

	G_FixSlashes( n.c_str() );

	

	sourcescript = this;

	Script::LoadFile( n.c_str() );

	FindLabels();



   crc = gi.CalcCRC( buffer, length );

	}



void GameScript::FindLabels

	(

   void

   )



   {

   scriptmarker_t mark;

   const char		*tok;

   script_label_t *label;

	int				len;



	FreeLabels();



	labelList = new Container<script_label_t *>;



	MarkPosition( &mark );

   

   Reset();



	while( TokenAvailable( TRUE ) )

		{

      tok = GetToken( TRUE );

      // see if it is a label

      if ( tok )

			{

			len = Burger::StringLength( tok );

			if ( len && tok[ len - 1 ] == ':' )

				{

				if ( !labelExists( tok ) )

					{

					label = new script_label_t;

         		MarkPosition( &label->pos );

					label->labelname = tok;

					labelList->AddObject( label );

					}

				else

					{

					warning( "FindLabels", "Duplicate labels %s\n", tok );

					}

				}

         }

		}



   RestorePosition( &mark );

   }



EXPORT_FROM_DLL qboolean GameScript::labelExists

	(

	const char *name

	)



	{

   Burger::String				labelname;

   script_label_t *label;

   int				i;

	int				num;



	if ( !sourcescript->labelList )

		{

		return FALSE;

		}



   labelname = name;

	if ( !labelname.length() )

		{

		return FALSE;

		}



   if ( labelname[ labelname.length() - 1 ] != ':' )

		{

      labelname += ":";

		}



   num = sourcescript->labelList->NumObjects();

   for( i = 1; i <= num; i++ )

      {

      label = sourcescript->labelList->ObjectAt( i );

      if ( labelname == label->labelname )

			{

         return TRUE;

			}

      }



   return FALSE;

	}



qboolean GameScript::Goto(const char *name)

{

	script_label_t *label;

	int i;

	int	num;

	

	if (sourcescript->labelList ) {

		Burger::String labelname(name);

		i = labelname.length();

		if (i) {

			if (labelname[i- 1] != ':' ) {

				labelname += ":";

			}

	

			num = sourcescript->labelList->NumObjects();

			for( i = 1; i <= num; i++ ) {

				label = sourcescript->labelList->ObjectAt( i );

				if (!strcmp(labelname.c_str(),label->labelname.c_str())) {

					RestorePosition( &label->pos );

					return TRUE;

				}

			}

		}

	}

	return FALSE;

}



EXPORT_FROM_DLL void GameScript::Mark

	( 

	GameScriptMarker *mark 

	)



	{

	assert( mark );

   assert( sourcescript );



	mark->filename = sourcescript->Filename();

	MarkPosition( &mark->scriptmarker );

	}



EXPORT_FROM_DLL void GameScript::Restore

	(

	GameScriptMarker *mark

	)



	{

   // If we change this function, we must update the unarchive function as well

   GameScript *scr;



	assert( mark );



   scr = ScriptLib.FindScript( mark->filename.c_str() );

   if ( scr )

      {

      SetSourceScript( scr );

      }

   else

      {

      LoadFile( mark->filename.c_str() );

      }



	RestorePosition( &mark->scriptmarker );

	}





#ifdef EXPORT_TEMPLATE

template class EXPORT_FROM_DLL Container<script_label_t *>;

template class EXPORT_FROM_DLL Container<GameScript *>;

#endif



