//-----------------------------------------------------------------------------
//
//  $Logfile:: /2015/worldspawn.h                                             $
// $Revision:: 1                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 12/27/00 5:22 PM                                               $
//
// Copyright (C) 1997 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// Base class for worldspawn objects.  This should be subclassed whenever
// a DLL has new game behaviour that needs to be initialized before any other
// entities are created, or before any entity thinks each frame.  Also controls
// spawning of clients.
//

#ifndef __WORLDSPAWN_H__
#define __WORLDSPAWN_H__

#include "entity.h"

class EXPORT_FROM_DLL TargetList : public Class
{
public:
	CLASS_PROTOTYPE( TargetList );
	Container<Entity *>  list;
	char *TargetNamePtr;
	
	TargetList();
	TargetList(const char *tname );
	~TargetList();
	void AddEntity( Entity * ent );
	void RemoveEntity( Entity * ent );
	Entity *GetNextEntity( Entity * ent );
};


class EXPORT_FROM_DLL World : public Entity
{
private:
	Container<TargetList *> targetList;
public:
	CLASS_PROTOTYPE( World );

	World();
	~World();
	char *SkipThreadPtr;

	void        FreeTargetList( void );
	TargetList  *GetTargetList(const char *targetname );
	void        AddTargetEntity(const char *targetname, Entity * ent );
	void        RemoveTargetEntity(const char *targetname, Entity * ent );
	Entity      *GetNextEntity(const char *targetname, Entity * ent );
	void        ActivateTarget(const char *targetname); //###
	virtual void Archive( Archiver &arc );
	virtual void Unarchive( Archiver &arc );
};


typedef SafePtr<World> WorldPtr;
extern WorldPtr world;

#endif /* worldspawn.h */
