//-----------------------------------------------------------------------------
//
//  $Logfile:: /game/actor.h                                                  $
// $Revision:: 2                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 12/27/00 6:34 PM                                               $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
// 
// DESCRIPTION:
// Base class for character AI.
//

#ifndef __ACTOR_H__
#define __ACTOR_H__

#include "g_local.h"
#include "weapon.h"
#include "sentient.h"
#include "container.h"
#include "stack.h"
#include "navigate.h"
#include "behavior.h"
#include "scriptmaster.h"
#include "prioritystack.h"

extern Event EV_Actor_Start;
extern Event EV_Actor_Dead;
extern Event EV_Actor_PopAnim;

extern Event EV_Actor_LookAt;
extern Event EV_Actor_TurnTo;
extern Event EV_Actor_FinishedBehavior;
extern Event EV_Actor_NotifyBehavior;
extern Event EV_Actor_FinishedMove;
extern Event EV_Actor_FinishedAnim;
extern Event EV_Actor_WalkTo;
extern Event EV_Actor_RunTo;
extern Event EV_Actor_Anim;
extern Event EV_Actor_Attack;
extern Event EV_Actor_InPain;
extern Event EV_Actor_Gib;
extern Event EV_Actor_ForwardToBehavior;
extern Event EV_Actor_Aim;
extern Event EV_Actor_MeleeRange;
extern Event EV_Actor_MeleeDamage;
extern Event EV_Actor_AttackFinished;
extern Event EV_Actor_Attack;
extern Event EV_Actor_AttackPlayer;

typedef enum
	{
	RANGE_MELEE,
	RANGE_NEAR,
	RANGE_MID,
	RANGE_FAR
	} range_t;

/*typedef*/ enum
	{
	ON_SIGHT,
	ON_ATTACK
	};

typedef enum
	{
	IGNORE,
	ATTACK,
	FLEE
	} reaction_t;

typedef enum
	{
	LIKES,			// Will not fire at, even if attacked
	TOLERATES,		// Will attack when attacked
	HATES,			// Will attack on sight
	WARY,				// Will flee when attacked
	FEARS,			// Will flee on sight
	NUM_DISPOSITIONS
	} disposition_t;

typedef enum
	{
   IS_INANIMATE,
	IS_MONSTER,
	IS_ENEMY,
	IS_CIVILIAN,
	IS_FRIEND,
   IS_ANIMAL,
	NUM_ACTORTYPES
	} actortype_t;

#define AI_CANWALK   0x00000001
#define AI_CANSWIM   0x00000002
#define AI_CANFLY    0x00000004

#define WEAK_WEIGHT		0.5
#define WEAK_HEALTH		20
#define STRONGER_WEIGHT 0.9
#define VISIBLE_WEIGHT	0.5
#define NEWENEMY_WEIGHT	0.75		// if he's a new enemy

//#define DAMAGE_WEIGHT	0.5	   // If he's done a lot of damage to you
//#define WEAPON_WEIGHT	1.5		// How much the weapon influences you

class EXPORT_FROM_DLL StateInfo : public Class
{
public:
	CLASS_PROTOTYPE( StateInfo );

	StateInfo();

	Burger::String						action;
	Burger::String						response;
	qboolean					ignore;

	virtual void         Archive( Archiver &arc );
	virtual void         Unarchive( Archiver &arc );
};

inline EXPORT_FROM_DLL void StateInfo::Archive
	(
	Archiver &arc
	)
   {
   Class::Archive( arc );

   arc.WriteCString( action.c_str() );
   arc.WriteCString( response.c_str() );
   arc.WriteBoolean( ignore );
   }

inline EXPORT_FROM_DLL void StateInfo::Unarchive(Archiver &arc)
{
	Class::Unarchive( arc );
	
	arc.ReadString( &action );
	arc.ReadString( &response );
	ignore = arc.ReadBoolean();
}

class EXPORT_FROM_DLL ActorState : public Class
{
public:
	CLASS_PROTOTYPE( ActorState );

	Burger::String							name;
	Burger::String							anim;
	Event							*animDoneEvent;
	Behavior						*behavior;
	PathPtr						path;
	int							thread;
	ThreadMarker				marker;
	Container<StateInfo *>	actionList;
	virtual void            Archive( Archiver &arc );
	virtual void            Unarchive( Archiver &arc );
};

inline EXPORT_FROM_DLL void ActorState::Archive
	(
	Archiver &arc
	)
   {
   int i, num;

   Class::Archive( arc );

   arc.WriteCString( name.c_str() );
   arc.WriteCString( anim.c_str() );
   arc.WriteEvent( *animDoneEvent );
   //FIXME
   arc.WriteBoolean( behavior != NULL );
   if ( behavior )
      {
      arc.WriteObject( behavior );
      }
   arc.WriteSafePointer( path );
   arc.WriteInteger( thread );
   arc.WriteObject( &marker );
   num = actionList.NumObjects();
   arc.WriteInteger( num );
   for ( i = 1; i <= num; i++ )
      {
      arc.WriteObject( actionList.ObjectAt( i ) );
      }
   }

inline EXPORT_FROM_DLL void ActorState::Unarchive
	(
	Archiver &arc
	)
   {
   Event * ev;
   int i, num;

   Class::Unarchive( arc );

   arc.ReadString( &name );
   arc.ReadString( &anim );
   ev = new Event;
   arc.ReadEvent( ev );
   animDoneEvent = ev;

   //FIXME
   behavior = NULL;
   if ( arc.ReadBoolean() )
      {
      behavior = ( Behavior * )arc.ReadObject();
      }

   arc.ReadSafePointer( &path );
   thread = arc.ReadInteger();
   arc.ReadObject( &marker );
   actionList.FreeObjectList();
   num = arc.ReadInteger();
   for ( i = 1; i <= num; i++ )
      {
      StateInfo * info;

      info = new StateInfo;
      arc.ReadObject( info );
      actionList.AddObject( info );
      }
   }

//
// Exported templated classes must be explicitly instantiated
//

class EXPORT_FROM_DLL Actor : public Sentient
{
public:
	Burger::String newanim;
	Event *newanimevent;
	int newanimnum;
	
	Burger::String spawngroup;
	
	int movement;
	stepmoveresult_t           lastmove;
	float forwardspeed;
	
	actortype_t                actortype;
	int                        attackmode;
	
	Vector							standsize_min;
	Vector							standsize_max;
	Vector							crouchsize_min;
	Vector							crouchsize_max;
	
	Burger::String								state;
	Burger::String								animname;
	Container<StateInfo *>		actionList;
	int								numonstack;
	Stack<ActorState *>			stateStack;
	
	BehaviorPtr						behavior;
	Burger::String								currentBehavior;
	
	PathPtr							path;
	
	Container<EntityPtr>			targetList;
	Container<EntityPtr>			nearbyList;
	Container<EntityPtr>			enemyList;
	EntityPtr						currentEnemy;
	qboolean							seenEnemy;
	range_t							enemyRange;
	EntityPtr						lastEnemy;
	
	float								fov;
	float								fovdot;
	float								vision_distance;
	
	Vector							startpos;
	
	Vector							move;
	Vector							movedir;
	float								movespeed;
	Vector							movevelocity;
	float								totallen;
	float								turnspeed;
	Vector							animdir;
	
	float								chattime;
	float								nextsoundtime;
	
	qboolean							hasalert;
	
	PathNodePtr						movegoal;
	PathNodePtr						soundnode;
	ThreadPtr						thread;
	
	ThreadPtr						actorthread;
	Burger::String								actorscript;
	Burger::String								actorstart;
	TouchFieldPtr					trig;
	
	qboolean                   has_melee;
	float                      melee_damage;
	float                      melee_range;
	float                      aim;
	float                      pain_threshold;
	
	qboolean                   checkStrafe;
	
	float					         next_drown_time;
	float					         air_finished;
	float                      attack_range;
	float                      shots_per_attack;
	qboolean                   deathgib;
	
	Burger::String                        kill_thread;
	Vector                     eyeoffset;
	float                      last_jump_time;
	qboolean                   nodeathfade;
	qboolean                   nochatter;
	
	CLASS_PROTOTYPE( Actor );
	
	// Initialization functions
					Actor();
					~Actor();
	void								Start( Event *ev );
	
	// Vision functions
	range_t							Range( Entity *ent );
	qboolean							InFOV(const Vector &pos );
	qboolean							InFOV( Entity *ent );
	qboolean							CanSeeFOV( Entity *ent );
	qboolean							CanSeeFrom(const Vector &pos, Entity *ent );
	qboolean							CanSee( Entity *ent );
	int								EnemyCanSeeMeFrom(const Vector &pos );
	qboolean							CanSeeEnemyFrom(const Vector &pos );
	
	// Weapon functions
	qboolean							WeaponReady( void );
	void								Attack( Event *ev );
	virtual Vector					GunPosition( void );
	virtual Vector             MyGunAngles( Vector muzzlepos, qboolean firing );
	virtual void					GetGunOrientation( Vector muzzlepos, Vector *forward, Vector *right, Vector *up );
	virtual qboolean				CanShootFrom( Vector pos, Entity *ent, qboolean usecurrentangles );
	virtual qboolean			   CanShoot( Entity *ent, qboolean usecurrentangles );
	
	// Actor type script commands
	void								FriendEvent( Event *ev );
	void								CivilianEvent( Event *ev );
	void								EnemyEvent( Event *ev );
	void								InanimateEvent( Event *ev );
	void								MonsterEvent( Event *ev );
	void								AnimalEvent( Event *ev );
	
	// Enemy management
	qboolean							HasEnemies( void );
	qboolean							IsEnemy( Entity *ent );
	void								MakeEnemy( Entity *ent, qboolean force = FALSE );
	qboolean							Likes( Entity *ent );
	qboolean							Hates( Entity *ent );
	
	// Targeting functions
	qboolean							GetVisibleTargets( void );
	void								TargetEnemies( Event *ev );
	Entity							*BestTarget( void );
	Sentient							*NearFriend( void );
	qboolean							CloseToEnemy(const Vector &pos, float howclose );
	float                      AttackRange( void );
	float                      MinimumAttackRange( void );
	
	// State control functions
	void								EnableState(const char *action );
	void								DisableState(const char *action );
	StateInfo						*SetResponse(const char *action,const char *response, qboolean ignore = FALSE );
	const char						*GetResponse(const char *action, qboolean force = FALSE );
	StateInfo						*GetState(const char *action );
	
	// State stack management
	void								ClearStateStack( void );
	qboolean							PopState( void );
	void								PushState( 	const char *newstate, ScriptThread *newthread = NULL, ThreadMarker *marker = NULL );
	
	// State control script commands
	void								DefineStateEvent( Event *ev );
	void								CopyStateEvent( Event *ev );
	void								IgnoreAllEvent( Event *ev );
	void								IgnoreEvent( Event *ev );
	void								RespondToAllEvent( Event *ev );
	void								RespondToEvent( Event *ev );
	void								ClearStateEvent( Event *ev );
	void								StateDoneEvent( Event *ev );
	void								SetStateEvent( Event *ev );
	
	// Thread management
	void								SetupThread( void );
	qboolean							DoAction(const char *name, qboolean force = FALSE );
	qboolean							ForceAction(const char *name );
	void								ProcessScript( ScriptThread *thread, Event *ev = NULL );
	void								StartMove( Event *ev );
	ScriptVariable             *SetVariable( const char *name, float value );
	ScriptVariable             *SetVariable( const char *name, int value	);
	ScriptVariable             *SetVariable( const char *name, const char *text );
	ScriptVariable             *SetVariable( const char *name, Entity *ent );
	ScriptVariable             *SetVariable( const char *name,const Vector &vec );
	
	// Thread based script commands
	void								SetScript( Event *ev );
	void								SetThread( Event *ev );
	
	// Behavior management
	void								EndBehavior( void );
	void								SetBehaviorEvent( Event *ev );
	void								SetBehavior( Behavior *newbehavior, Event *argevent = NULL, ScriptThread *thread = NULL );
	void								FinishedBehavior( Event *ev );
	void								NotifyBehavior( Event *ev );
	void                       ForwardBehaviorEvent( Event *ev );
	
	// Path and node management
	void								SetPath( Path *newpath );
	void								ReserveNodeEvent( Event *ev );
	void								ReleaseNodeEvent( Event *ev );
	
	// Animation control functions
	void								ChangeAnim( void );
	qboolean							SetAnim(const char * anim, Event *ev = NULL );
	qboolean							SetAnim(const char *anim, Event &ev );
	void								Anim( Event *ev );
	
	// Script commands
	void								CrouchSize( Event *ev );
	void                       SetFov( Event *ev );
	void                       SetVisionDistance( Event *ev );
	void								LookAt( Event *ev );
	void								TurnToEvent( Event *ev );
	void								IdleEvent( Event *ev );
	void								WalkTo( Event *ev );
	void								RunTo( Event *ev );
	void								AttackEntity( Event *ev );
	void                       AttackPlayer( Event *ev );
	void								JumpToEvent( Event *ev );
	void                       GotoEvent( Event *ev );
	
	// Script conditionals
	void								IfEnemyVisibleEvent( Event *ev );
	void								IfNearEvent( Event *ev );
	void								IfCanHideAtEvent( Event *ev );
	void								IfCanStrafeAttackEvent( Event *ev );
	void								IfCanMeleeAttackEvent( Event *ev );
	void                       IfCanShootEvent( Event *ev );
	void								IfEnemyWithinEvent( Event *ev );
	
	// Sound reaction functions
	void								IgnoreSoundsEvent( Event *ev );
	void								RespondToSoundsEvent( Event *ev );
	void								InvestigateWeaponSound( Event *ev );
	void								InvestigateMovementSound( Event *ev );
	void								InvestigatePainSound( Event *ev );
	void								InvestigateDeathSound( Event *ev );
	void								InvestigateBreakingSound( Event *ev );
	void								InvestigateDoorSound( Event *ev );
	void								InvestigateMutantSound( Event *ev );
	void								InvestigateVoiceSound( Event *ev );
	void								InvestigateMachineSound( Event *ev );
	void								InvestigateRadioSound( Event *ev );
	
	// Pain and death related functions
	void								Pain( Event *ev );
	void								Dead( Event *ev );
	void								Killed( Event *ev );
	void                       GibEvent( Event *ev );
	void                       RemoveUselessBody( Event *ev );
	
	// Movement functions
	void								ForwardSpeedEvent( Event *ev );
	void								SwimEvent( Event *ev );
	void								FlyEvent( Event *ev );
	void								NotLandEvent( Event *ev );
	qboolean							CanMoveTo( Vector pos );
	void								Accelerate( Vector steering );
	void								CalcMove( void );
	void								setAngles( Vector ang );
	stepmoveresult_t				WaterMove( void );
	stepmoveresult_t				AirMove( void );
	stepmoveresult_t				TryMove( void );
	void                       CheckWater( void );
	virtual void setSize(const Vector &min,const Vector &max);
	
	// Debug functions
	void                       ShowInfo( void );
	
	// General functions
	virtual void					Chatter( const char *sound, float chance = 10, float volume = 1.0f, int channel = CHAN_VOICE );
	void								ActivateEvent( Event *ev );
	void								UseEvent( Event *ev );
	void								Prethink( void );
	
	virtual void               Archive( Archiver &arc );
	virtual void               Unarchive( Archiver &arc );
	
	void                       SetAim( Event *ev );
	void                       SetMeleeRange( Event *ev );
	void                       SetMeleeDamage( Event *ev );
	void                       MeleeEvent( Event *ev );
	void                       AttackFinishedEvent( Event *ev );
	
	void                       CanStrafeEvent( Event *ev );
	void                       NoStrafeEvent( Event *ev );
	void                       SetPainThresholdEvent( Event *ev );
	void                       SetKillThreadEvent( Event *ev );
	void                       AttackRangeEvent( Event *ev );
	void                       AttackModeEvent( Event *ev );
	void                       ShotsPerAttackEvent( Event *ev );
	void                       ClearEnemyEvent( Event *ev );
	void					         SetHealth( Event *ev );
	void                       ClearEnemies( void );
	void                       EyeOffset( Event *ev );
	void                       NoDeathFadeEvent( Event *ev );
	void                       NoChatterEvent( Event *ev );
	void                       TurnSpeedEvent( Event *ev );
	
	qboolean                   HasWeapon( void );
	float								JumpTo( PathNode * goal, float speed );
	float                      JumpTo( Entity * goal, float speed );
	float                      JumpTo( Vector targ, float speed );
	};

inline EXPORT_FROM_DLL void Actor::Archive
	(
	Archiver &arc
	)
   {
   int i, num;
   Stack<ActorState *> tempStack;
   ActorState *s;

   Sentient::Archive( arc );

   arc.WriteCString( newanim.c_str() );
   arc.WriteEvent( *newanimevent );
   arc.WriteInteger( newanimnum );

   arc.WriteCString( spawngroup.c_str() );

   arc.WriteInteger( movement );
   // cast as stepmoveresult_t on read
   arc.WriteInteger( ( int )lastmove );
   arc.WriteFloat( forwardspeed );

   // cast as actortype_t on read
   arc.WriteInteger( ( int )actortype );
   arc.WriteInteger( attackmode );

   arc.WriteVector( standsize_min );
   arc.WriteVector( standsize_max );
   arc.WriteVector( crouchsize_min );
   arc.WriteVector( crouchsize_max );

   arc.WriteCString( state.c_str() );
   arc.WriteCString( animname.c_str() );

   num = actionList.NumObjects();
   arc.WriteInteger( num );
   for ( i = 1; i <= num; i++ )
      {
      arc.WriteObject( actionList.ObjectAt( i ) );
      }

   arc.WriteInteger( numonstack );

   // We have to push all the states onto another stack in order to
   // save them out in the proper order and to restore them after
   // saving the game.
   for( i = 1; i <= numonstack; i++ )
      {
      tempStack.Push( stateStack.Pop() );
      }

   // put them back on the stateStack as we write them out.
   for( i = 1; i <= numonstack; i++ )
      {
      s = tempStack.Pop();
      stateStack.Push( s );
      arc.WriteObject( s );
      }

   // FIXME
   arc.WriteBoolean( behavior != NULL );
   if ( behavior )
      {
      arc.WriteObject( behavior );
      }

   arc.WriteCString( currentBehavior.c_str() );

   arc.WriteSafePointer( path );

   num = targetList.NumObjects();
   arc.WriteInteger( num );
   for ( i = 1; i <= num; i++ )
      {
      arc.WriteSafePointer( targetList.ObjectAt( i ) );
      }

   num = nearbyList.NumObjects();
   arc.WriteInteger( num );
   for ( i = 1; i <= num; i++ )
      {
      arc.WriteSafePointer( nearbyList.ObjectAt( i ) );
      }

   num = enemyList.NumObjects();
   arc.WriteInteger( num );
   for ( i = 1; i <= num; i++ )
      {
      arc.WriteSafePointer( enemyList.ObjectAt( i ) );
      }

   arc.WriteSafePointer( currentEnemy );
   arc.WriteBoolean( seenEnemy );
   // cast as range_t on read
   arc.WriteInteger( ( int )enemyRange );
   arc.WriteSafePointer( lastEnemy );

   arc.WriteFloat( fov );
   arc.WriteFloat( fovdot );
   arc.WriteFloat( vision_distance );

   arc.WriteVector( startpos );

   arc.WriteVector( move );
   arc.WriteVector( movedir );
   arc.WriteFloat( movespeed );
   arc.WriteVector( movevelocity );
   arc.WriteFloat( totallen );
   arc.WriteFloat( turnspeed );
   arc.WriteVector( animdir );

   arc.WriteFloat( chattime );
   arc.WriteFloat( nextsoundtime );

   arc.WriteBoolean( hasalert );

   arc.WriteSafePointer( movegoal );
   arc.WriteSafePointer( soundnode );
   arc.WriteSafePointer( thread );

   arc.WriteSafePointer( actorthread );
   arc.WriteCString( actorscript.c_str() );
   arc.WriteCString( actorstart.c_str() );
   arc.WriteSafePointer( trig );

   arc.WriteBoolean( has_melee );
   arc.WriteFloat( melee_damage );
   arc.WriteFloat( melee_range );
   arc.WriteFloat( aim );
   arc.WriteFloat( pain_threshold );

   arc.WriteBoolean( checkStrafe );

   arc.WriteFloat( next_drown_time );
   arc.WriteFloat( air_finished );
   arc.WriteFloat( attack_range );
   arc.WriteFloat( shots_per_attack );
   arc.WriteBoolean( deathgib );

   arc.WriteCString( kill_thread.c_str() );
   arc.WriteVector( eyeoffset );
   arc.WriteFloat( last_jump_time );
   arc.WriteBoolean( nodeathfade );
   arc.WriteBoolean( nochatter );
   }

inline EXPORT_FROM_DLL void Actor::Unarchive(Archiver &arc)
{
	Event * ev;
	int i, num, temp;
	
	Sentient::Unarchive( arc );
	
	arc.ReadString( &newanim );
	ev = new Event;
	arc.ReadEvent( ev );
	newanimevent = ev;
	newanimnum = arc.ReadInteger();
	
	arc.ReadString( &spawngroup );
	
	movement = arc.ReadInteger();
	// cast as stepmoveresult_t on read
	temp = arc.ReadInteger();
	lastmove = ( stepmoveresult_t )temp;
	forwardspeed = arc.ReadFloat();
	
	// cast as actortype_t on read
	temp = arc.ReadInteger();
	actortype = ( actortype_t )temp;
	attackmode = arc.ReadInteger();
	
	arc.ReadVector( &standsize_min );
	arc.ReadVector( &standsize_max );
	arc.ReadVector( &crouchsize_min );
	arc.ReadVector( &crouchsize_max );
	
	arc.ReadString( &state );
	arc.ReadString( &animname );
	
	num = arc.ReadInteger();
	for( i = 1; i <= num; i++ )	{
		StateInfo * info;
	
		info = new StateInfo;
		arc.ReadObject( info );
		actionList.AddObject( info );
	}
	
	numonstack = arc.ReadInteger();
	
	for( i = 1; i <= numonstack; i++ ) {
		ActorState * state;
	
		state = new ActorState;
		arc.ReadObject( state );
		stateStack.Push( state );
	}
	
	behavior = NULL;
	if ( arc.ReadBoolean() ) {
		behavior = ( Behavior * )arc.ReadObject();
	}
	
	arc.ReadString( &currentBehavior );
	
	arc.ReadSafePointer( &path );
	
	num = arc.ReadInteger();
	targetList.FreeObjectList();
	targetList.Resize( num );
	for( i = 1; i <= num; i++ ) {
		EntityPtr tmp, *ptr;
	
		targetList.AddObject( tmp );
		ptr = targetList.AddressOfObjectAt( i );
		arc.ReadSafePointer( ptr );
	}
	
	num = arc.ReadInteger();
	nearbyList.FreeObjectList();
	nearbyList.Resize( num );
	for( i = 1; i <= num; i++ ) {
		EntityPtr tmp, *ptr;
	
		nearbyList.AddObject( tmp );
		ptr = nearbyList.AddressOfObjectAt( i );
		arc.ReadSafePointer( ptr );
	}
	
	num = arc.ReadInteger();
	enemyList.FreeObjectList();
	enemyList.Resize( num );
	for( i = 1; i <= num; i++ ) {
		EntityPtr tmp, *ptr;
	
		enemyList.AddObject( tmp );
		ptr = enemyList.AddressOfObjectAt( i );
		arc.ReadSafePointer( ptr );
	}
	
	arc.ReadSafePointer( &currentEnemy );
	seenEnemy = arc.ReadBoolean();
	// cast as range_t on read
	temp = arc.ReadInteger();
	enemyRange = ( range_t )temp;
	arc.ReadSafePointer( &lastEnemy );
	
	fov = arc.ReadFloat();
	fovdot = arc.ReadFloat();
	vision_distance = arc.ReadFloat();
	
	arc.ReadVector( &startpos );
	
	arc.ReadVector( &move );
	arc.ReadVector( &movedir );
	movespeed = arc.ReadFloat();
	arc.ReadVector( &movevelocity );
	totallen = arc.ReadFloat();
	turnspeed = arc.ReadFloat();
	arc.ReadVector( &animdir );
	
	chattime = arc.ReadFloat();
	nextsoundtime = arc.ReadFloat();
	
	hasalert = arc.ReadBoolean();
	
	arc.ReadSafePointer( &movegoal );
	arc.ReadSafePointer( &soundnode );
	arc.ReadSafePointer( &thread );
	
	arc.ReadSafePointer( &actorthread );
	arc.ReadString( &actorscript );
	arc.ReadString( &actorstart );
	arc.ReadSafePointer( &trig );
	
	has_melee = arc.ReadBoolean();
	melee_damage = arc.ReadFloat();
	melee_range = arc.ReadFloat();
	aim = arc.ReadFloat();
	pain_threshold = arc.ReadFloat();
	
	checkStrafe = arc.ReadBoolean();
	
	next_drown_time = arc.ReadFloat();
	air_finished = arc.ReadFloat();
	attack_range = arc.ReadFloat();
	shots_per_attack = arc.ReadFloat();
	deathgib = arc.ReadBoolean();
	
	arc.ReadString( &kill_thread );
	arc.ReadVector( &eyeoffset );
	last_jump_time = arc.ReadFloat();
	nodeathfade = arc.ReadBoolean();
	nochatter = arc.ReadBoolean();
}


class EXPORT_FROM_DLL MonsterStart : public PathNode
{
private:
	Burger::String								spawngroup;
	MonsterStart					*groupchain;
	MonsterStart					*leader;
	static MonsterStart			*GetGroupLeader(const char *groupname );

public:
	CLASS_PROTOTYPE( MonsterStart );

	MonsterStart();
	static MonsterStart			*GetRandomSpot( Burger::String& groupname );
	virtual void               Archive( Archiver &arc );
	virtual void               Unarchive( Archiver &arc );
};

inline EXPORT_FROM_DLL void MonsterStart::Archive
	(
	Archiver &arc
	)
   {
   PathNode::Archive( arc );

   arc.WriteCString( spawngroup.c_str() );
   arc.WriteObjectPointer( groupchain );
   arc.WriteObjectPointer( leader );
   }

inline EXPORT_FROM_DLL void MonsterStart::Unarchive
	(
	Archiver &arc
	)
   {
   PathNode::Unarchive( arc );

   arc.ReadString( &spawngroup );
   arc.ReadObjectPointer( ( Class ** )&groupchain );
   arc.ReadObjectPointer( ( Class ** )&leader );
   }

// Set destination to node with duck or cover set.  Class will find a path to that node, or a closer one.
class EXPORT_FROM_DLL FindCoverMovement : public StandardMovement
	{
	public:
		Actor			*self;

	inline qboolean validpath
		(
		PathNode *node,
		int i
		)

		{
		pathway_t *path;
		PathNode *n;

		path = &node->Child[ i ];

		if ( !StandardMovement::validpath( node, i ) )
			{
			return FALSE;
			}

		n = AI_GetNode( path->node );
		if ( !n || self->CloseToEnemy( n->worldorigin, 128 ) )
			{
			return FALSE;
			}

		return TRUE;
		}

	inline qboolean done
		(
		PathNode *node,
		PathNode *end
		)

		{
		if ( node == end )
			{
			return TRUE;
			}
		
		//if ( node->reject )
		if ( node->reject || !( node->nodeflags & ( AI_DUCK | AI_COVER ) ) )
			{
			return FALSE;
			}

		if ( self )
			{
			node->reject = self->CanSeeEnemyFrom( node->worldorigin );
			return !node->reject;
			}

		return FALSE;
		}
	};

// Set destination to node with flee set.  Class will find a path to that node, or a closer one.
class EXPORT_FROM_DLL FindFleeMovement : public StandardMovement
	{
	public:
		Actor			*self;

	inline qboolean validpath
		(
		PathNode *node,
		int i
		)

		{
		pathway_t *path;
		PathNode *n;

		path = &node->Child[ i ];

		if ( !StandardMovement::validpath( node, i ) )
			{
			return FALSE;
			}

		n = AI_GetNode( path->node );
		if ( !n || self->CloseToEnemy( n->worldorigin, 128 ) )
			{
			return FALSE;
			}

		return TRUE;
		}

	inline qboolean done
		(
		PathNode *node,
		PathNode *end
		)

		{
		if ( node == end )
			{
			return TRUE;
			}
		
		//if ( node->reject )
		if ( node->reject || !( node->nodeflags & AI_FLEE ) )
			{
			return FALSE;
			}

		if ( self )
			{
			node->reject = self->CanSeeEnemyFrom( node->worldorigin );
			return !node->reject;
			}

		return FALSE;
		}
	};

class EXPORT_FROM_DLL FindEnemyMovement : public StandardMovement
	{
	public:
		Actor			*self;

	inline qboolean done
		(
		PathNode *node,
		PathNode *end
		)

		{
		if ( node == end )
			{
			return TRUE;
			}
		
		if ( node->reject )
			{
			return FALSE;
			}

		if ( self )
			{
			if ( self->currentEnemy )
   			{
				node->reject = !self->CanShootFrom( node->worldorigin, self->currentEnemy, FALSE );
			   }
			else
			   {
				node->reject = FALSE;
			   }

         return !node->reject;
			}

		return FALSE;
		}
	};


typedef PathFinder<FindCoverMovement> FindCoverPath;
typedef PathFinder<FindFleeMovement> FindFleePath;
typedef PathFinder<FindEnemyMovement> FindEnemyPath;

#endif /* actor.h */
