#include "g_local.h"
#include "vector.h"
#include "mac_import.h"

Vector vec_zero(0,0,0);

/**********************************

	Parse a text string to get the amount

**********************************/

Vector::Vector(const char *text)
{
	x = 0;
	y = 0;
	z = 0;
	if ( text ) {
		Get3Floats(text,&x,&y,&z);
	}
}

/**********************************

	Return the vector's yaw in degrees (0-359 as int)

**********************************/

float Vector::toYaw(void)
{
	float yaw;

	if ( ( y == 0 ) && ( x == 0 ) ) {
		yaw = 0;
	} else {
		yaw = ( float )( ( int )( atan2( y, x ) * 180 / M_PI ) );
		if ( yaw < 0 ) {
			yaw += 360;
		}
	}
	return yaw;
}

/**********************************

	Return the vector's pitch in degrees (0-359 as int)

**********************************/

float Vector::toPitch(void)
{
	float	forward;
	float	pitch;
	
	if ( ( x == 0 ) && ( y == 0 ) ) {
		if ( z > 0 ) {
			pitch = 90;
		} else {
			pitch = 270;
		}
	} else {
		forward = ( float )Burger::FPMath::Sqrt( x * x + y * y );
		pitch = ( float )( ( int )( atan2( z, forward ) * 180 / M_PI ) );
		if ( pitch < 0 ) {
			pitch += 360;
		}
	}	
	return pitch;
}

/**********************************

	Set the angles from an arbitrary vector
	Note : I can be modifing myself, so don't touch
	the "this" record until I am done with the Input
	record

**********************************/

void Vector::toAngles(const Vector &Input)
{
	float forward;
	float yaw,pitch;
	
	if ( ( Input.x == 0 ) && ( Input.y == 0 ) ) {
		yaw = 0;
		if ( Input.z > 0 ) {
			pitch = 90;
		} else {
			pitch = 270;
		}
	} else {
		yaw = atan2( Input.y, Input.x ) * 180 / M_PI;
		if ( yaw < 0 ) {
			yaw += 360;
		}
	
		forward = ( float )Burger::FPMath::Sqrt( Input.x * Input.x + Input.y * Input.y );
		pitch = atan2( Input.z, forward ) * 180 / M_PI;
		if ( pitch < 0 ) {
			pitch += 360;
		}
	}
	x = pitch;
	y = yaw;
	z = 0;
}

/**********************************

	Create a matrix based on euler angles in a vector

**********************************/

void Vector::AngleVectors(Vector *forward,Vector *right,Vector *up) const
{
	float angle;
#if defined(__INTEL__)
	static float	sr, sp, sy, cr, cp, cy; // static to help MS compiler fp bugs
#else
	float sr, sp, sy, cr, cp, cy; // static to help MS compiler fp bugs
#endif	
	angle = yaw() * ( M_PI * 2 / 360 );
	sy = sin( angle );
	cy = cos( angle );
	
	angle = pitch() * ( M_PI * 2 / 360 );
	sp = sin( angle );
	cp = cos( angle );
	
	angle = roll() * ( M_PI * 2 / 360 );
	sr = sin( angle );
	cr = cos( angle );
	
	if ( forward ) {
		forward->setXYZ( cp * cy, cp * sy, -sp );
	}
	
	if ( right ) {
		right->setXYZ((-sr * sp * cy) + (cr * sy), (-sr * sp * sy) - (cr * cy),	(-sr * cp) );
	}
	
	if ( up ) {
		up->setXYZ( (cr * sp * cy) + (sr * sy), (cr * sp * sy) - (sr * cy), cr * cp );
	}
}

/**********************************

	Interpolate between two vectors

**********************************/

#define LERP_DELTA 1e-6
void Vector::LerpVector(const Vector &w1,const Vector &w2,float t)
{
	float omega, cosom, sinom, scale0, scale1;
	Vector v2;
	
	this->normalize(w1);
	v2.normalize(w2);	
	cosom = this[0] * v2;		/* Dot product */
	if ( ( 1.0f - cosom ) > LERP_DELTA ) {
		omega = (float)acos( cosom );
		sinom = 1.0f/(float)sin( omega );		/* Get reciprocal */
		scale0 = (float)sin( ( 1.0f - t ) * omega ) * sinom;	/* Actually divide by sinom */
		scale1 = (float)sin( t * omega ) * sinom;
	} else {
		scale0 = 1.0f - t;
		scale1 = t;
	}
	this->x = this->x*scale0 + v2.x*scale1;
	this->y = this->y*scale0 + v2.y*scale1;
	this->z = this->z*scale0 + v2.z*scale1;
}



