//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
// 
// DESCRIPTION:
// Directshow player
//

#include <burger.h>
#include <Movies.h>
#include <Errors.h>
#include <assert.h>
#include "client.h"

static int vid_modified;
static int ref_modified;
static int fullscreen_modified;

static Movie fopen_cin(const char *filen);

static Movie movie;			/* Current movie reference */
static CGrafPtr movieWorld;	/* Movie grafport */

qboolean cinematic_active;
qboolean cinematic_paused;

/*
==================
SCR_StopCinematic
==================
*/
void SCR_StopCinematic (void)
{
	if ( cl.cinematictime ) {
		cl.cinematictime = 0;	// done
		if (movie) {
			StopMovie(movie);
			DisposeMovie(movie);
			movie = 0;
		}
		if (movieWorld) {
			DisposeGWorld(movieWorld);
			movieWorld = 0;
		}
		re.EndCinematic();
		SCR_FinishCinematic();
	}

	if ( !sound_started ) {
		CL_Snd_Restart_f ();
	}
	
	if ( cl_forcecinematic->value ) {
		//
		// restore our settings
		//
		if ( ref_modified ) {
			ref_modified = 0;
			Cvar_Set( "vid_ref", "gl" );
			vid_ref->modified = TRUE;
		}
		if ( vid_modified ) {
			Cvar_SetValue( "gl_ext_palettedtexture", vid_modified );
			vid_modified = 0;
			vid_ref->modified = TRUE;
		}
		if ( swmode_modified ) {
			Cvar_SetValue( "sw_mode", swmode_modified );
			swmode_modified = 0;
			vid_ref->modified = TRUE;
		}
		if ( fullscreen_modified ) {
			Cvar_SetValue( "vid_fullscreen", 0 );
			fullscreen_modified = 0;
			vid_ref->modified = TRUE;
		}
	}
}

/*
====================
SCR_FinishCinematic

Called when either the cinematic completes, or it is aborted
====================
*/
void SCR_FinishCinematic (void)
{
	//
	// make sure we stop it
	//
	SCR_StopCinematic();
	// tell the server to advance to the next map / cinematic
	MSG_WriteByte (&cls.netchan.message, clc_stringcmd);
	SZ_Print (&cls.netchan.message, va("nextserver %i\n", cl.servercount));
}

/*
==================
SCR_RunCinematic
decode next frame
==================
*/
void SCR_RunCinematic (void) 
{
	if (cl.cinematictime <= 0) {
		return;
	}

	if ( cl.cinematictime ) {
		//
		// we render until their are no more updates
		// not TRUE right now
		//
	
		if (movie) {
			MoviesTask(movie, 0);
		}
		if (!movie || IsMovieDone(movie)) {
			SCR_StopCinematic ();
			return;
		}
	}
}
   
/*
==================
SCR_DrawCinematic

Returns TRUE if a cinematic is active, meaning the view rendering
should be skipped
==================
*/
qboolean SCR_DrawCinematic (void)
{
	PixMapHandle	pm;
	
	if (cl.cinematictime <= 0) {
		return FALSE;
	}	
	if (movieWorld) {
		pm = GetGWorldPixMap(movieWorld);
		LockPixels(pm);
	
		re.DrawCinematicFrame(0, 0, viddef.width, viddef.height,
			pm[0]->bounds.right, pm[0]->bounds.bottom,
			GetPixRowBytes(pm), (Word16 *)GetPixBaseAddr(pm));
		UnlockPixels(pm);
	}
	return TRUE;
}

/*
==================
SCR_PlayCinematic

==================
*/
void SCR_PlayCinematic ( const char *arg )
{
	char name[ MAX_OSPATH ];
	FILE * fp;
	
	// 
	// try the cd first
	//
	if ( !strstr( arg, ":" ) ) {
		Com_sprintf( name, sizeof( name ), "%s/movies/%s", FS_CDdir(), arg );
	} else {
		strncpy( name, arg, sizeof( name ) - 1 );
	}
	fp = QuakeOpenAFile( name, "rb" );
	if ( !fp ) {
		// 
		// try the hardrive next
		//
		Com_sprintf( name, sizeof( name ), "%s/movies/%s", FS_Gamedir(), arg );
		fp = QuakeOpenAFile( name, "rb" );
		if ( !fp ) {
			// we didn't find it so lets return and not do the silly video switches
			SCR_FinishCinematic();
			return;
		} else
			fclose( fp );
	} else {
		fclose( fp );
	}
	if ( cl.cinematictime > 0 ) {
		SCR_StopCinematic();
	}

	cl.cinematictime = 0;
	vid_modified = 0;
	ref_modified = 0;

	if ( cl_forcecinematic->value ) {
		//
		// make sure we are in 16-bit software period
		//
		if ( !Burger::StringCaseCompare( Cvar_VariableString( "vid_ref" ), "soft" ) ) {
			//
			// a value of 0 is 16-bit any thing else is an 8-bit variant 
			//
			vid_modified = Cvar_VariableValue( "gl_ext_palettedtexture" );
			swmode_modified = Cvar_VariableValue( "sw_mode" );
			fullscreen_modified = Cvar_VariableValue( "vid_fullscreen" );
			if ( (vid_modified) || (swmode_modified != 0) || (fullscreen_modified != 1) ) {
				if ( !fullscreen_modified )
					fullscreen_modified = 1;
				else
					fullscreen_modified = 0;

				Cvar_SetValue( "gl_ext_palettedtexture", 0 );
				Cvar_SetValue( "sw_mode", 0 );
//				Cvar_SetValue( "vid_fullscreen", 1 );
				vid_ref->modified = TRUE;
				VID_CheckChanges();
			} else
				fullscreen_modified = 0;
		} else {	// else we are in hardware
			ref_modified = 1;
			vid_modified = Cvar_VariableValue( "gl_ext_palettedtexture" );
			swmode_modified = Cvar_VariableValue( "sw_mode" );
			fullscreen_modified = Cvar_VariableValue( "vid_fullscreen" );
			if ( (vid_modified) || (swmode_modified != 0) || (fullscreen_modified != 1) ) {
				if ( !fullscreen_modified )
					fullscreen_modified = 1;
				else
					fullscreen_modified = 0;
				Cvar_Set( "vid_ref", "soft" );
				Cvar_SetValue( "gl_ext_palettedtexture", 0 );
				Cvar_SetValue( "sw_mode", 0 );
//				Cvar_SetValue( "vid_fullscreen", 1 );
				vid_ref->modified = TRUE;
				VID_CheckChanges();
			} else
				fullscreen_modified = 0;
		}
	}

	// make sure CD isn't playing music
	CDAudio_Stop();

	S_Shutdown();

	movie = fopen_cin( name );
	
	if (movie) {
		Rect	movieSize;
		OSErr	err;
		re.BeginCinematic();
		
		GetMovieBox(movie, &movieSize);
		
		err = NewGWorld(&movieWorld, 16, &movieSize, NULL, /*VideoDevice*/NULL, 0);
		
		if (err == noErr) {
			cl.cinematictime = Sys_Milliseconds();
			cl.cinematicframe = 0;
			
			SetMovieGWorld(movie,movieWorld, GetGWorldDevice(movieWorld));
			
			StartMovie(movie);
			MoviesTask(movie, 0);
		} else {
			DisposeMovie(movie);
			movie = NULL;
		}
	}
	
	if ( !movie ) {
		Com_Printf("SCR_PlayCinematic Error: %s\n", "(Unknown)" );
	}
	if ( cl.cinematictime <= 0 ) {
		SCR_FinishCinematic();
		Com_Printf ("Unable to start cinematic: %s\n", arg);
	}
}

static Movie fopen_cin(const char *filen)
{
	char temppath[MAX_OSPATH];
	unsigned char PFileName[MAX_OSPATH];		/* New filename from ExpandAPathToBufferNative */
	OSErr err;
	FSSpec spec;
	short movieFile, movieResID = 0;
	Movie theMovie;
	Str255	movieName;
	Bool movieChanged;

	int strip0;
	temppath[0] = '9';	// do burgerlib app prefix
	temppath[1] = ':';
//NCT_MAC strip ./ off of paths
	if ( (filen[0] == '.') && (filen[1] == '/'))
	{
		Burger::StringCopy(&temppath[2],&filen[2]);
		strip0 = 2;
		while (temppath[strip0] != 0)
		{
			if ( (temppath[strip0] == '/') || (temppath[strip0] == '\\') )
				temppath[strip0] = ':';
			else
				temppath[strip0] = temppath[strip0];
			++strip0;
		}
		temppath[strip0] = 0;
	} else {
		if (filen[0] == '/')
			Burger::StringCopy(&temppath[2],&filen[1]);
		else
			Burger::StringCopy(&temppath[2],filen);

		strip0 = 0;
		while (temppath[strip0] != 0)
		{
			if ( (temppath[strip0] == '/') || (temppath[strip0] == '\\') )
				temppath[strip0] = ':';
			++strip0;
		}
	}

	Burger::Filename MyFilename;
	MyFilename.ToNative(temppath);
	Burger::CStringToPString(PFileName,MyFilename.GetPtr());

	if ((err = FSMakeFSSpec(MyFilename.GetVRefNum(),MyFilename.GetDirID(),PFileName,&spec)) != noErr) {
		temppath[0] = ':';			/* full path CD style */
		if (filen[0] == '/')
			Burger::StringCopy(&temppath[1],&filen[1]);
		else
			Burger::StringCopy(&temppath[1],filen);
		strip0 = 0;
		while (temppath[strip0] != 0)
		{
			if ( (temppath[strip0] == '/') || (temppath[strip0] == '\\') )
				temppath[strip0] = ':';
			++strip0;
		}

		MyFilename.ToNative(temppath);
		Burger::CStringToPString(PFileName,MyFilename.GetPtr());
		if ((err = FSMakeFSSpec( 0, 0,PFileName, &spec)) != noErr)
			return 0;
	}
	
	if ((err = OpenMovieFile(&spec, &movieFile, fsRdPerm)) != noErr)
		return nil;

	err = NewMovieFromFile(&theMovie, movieFile, &movieResID, movieName,
			newMovieActive, &movieChanged);
	
	CloseMovieFile(movieFile);

	if (err != noErr)
		return nil;
	
	return theMovie;
}
