#include "gl_local.h"
#include "mac_drawsp.h"
#include <QDOffscreen.h>
#include <agl.h>
#include <glm.h>
#include <MacWindows.h>

// Global that defines whether or not we're a paletted software renderer
// or a direct GL one.  Used by the drawsprocket module to determine
// how to setup the context
extern int gIsOpenGL;

static int gIsFullscreen = FALSE;

// NCT_FLAG, whether or not I just throw up a window (debug) or use DrawSprocket to go 'fullscreen'
#define MACGL_DEBUG 1

qboolean GLimp_InitGraphics( qboolean fullscreen );


static int gFullscreen;

/*
 * RendererID: Official.
 */

#if 0
#define AGL_RENDERER_MESA_3DFX_ID 0x00040000

// Our globals

static GLint Vattrib[] = { AGL_RGBA,AGL_NONE };

int MacGLGet_VRAM(void)
{
	AGLPixelFormat pix;
	AGLRendererInfo info, head_info;
	GLint tot_mem;
	AGLDevice *device;
	GLint ndevs, n;

	if (!gIsOpenGL) {
		return 0;
	}
	/* Never unload a code module */

	aglConfigure(AGL_RETAIN_RENDERERS, GL_TRUE);
	
	/* Choose an rgb pixel format */

	pix = aglChoosePixelFormat(0, 0, Vattrib);
	if (pix) {
		device = aglDevicesOfPixelFormat(pix, &ndevs);
		aglDestroyPixelFormat(pix);
		if (device && ndevs>=1) {
			n = 0;
			do {
				if (!VideoDevice || (ndevs==1) || device[n]==VideoDevice) {
					head_info = aglQueryRendererInfo(&device[n],1);
					if (head_info) {
						info = head_info;
						do {
							aglDescribeRenderer(info, AGL_VIDEO_MEMORY, &tot_mem);

						// NCT 1.0.3, make sure we get the RIGHT GDevice too!
							if ( (tot_mem > 0) && (!VideoDevice || (device[n] == VideoDevice)) ) {
								aglDestroyRendererInfo(head_info);
								return (int)(tot_mem/1024);
							}
							info = aglNextRendererInfo(info);
						} while (info);
						aglDestroyRendererInfo(head_info);
					}
				}
			} while (++n<ndevs);
		}
	}
	return 0;
}
#endif

//=============================================================================
// GL STUFF

/*
** OpenGL Setup
*/

// GLOBAL
static int IsMacGLActive = FALSE;

void MacGLWindow_Activate(void)
{
	if (!gIsOpenGL) {
		return;
	}
	if (VideoWindow) {
		AGLContext gctx;
		gctx = aglGetCurrentContext();
		if (gctx) {
			aglSetDrawable(gctx,GetWindowPort(VideoWindow)); 
		}
		ShowWindow(VideoWindow);
		BringToFront(VideoWindow);
	}
}

void DestroyMacGLWindow(void)
{
	if (IsMacGLActive) {
		/* Cleanup the OpenGL context */
	
		IsMacGLActive = FALSE;
		glFlush();
		OpenGLRelease();
	}
}

void GLimp_BeginFrame( float camera_separation )
{
}

void		GLimp_EndFrame( void )
{
	glFlush();
	OpenGLSwapBuffers();
}

int GLimp_Init( void *hinstance, void *hWnd )
{
#if 0				/* If enabled, the Mac will use app memory for textures */
	glmSetMode(GLM_APPLICATION_HEAP_MODE);
#endif
	SaveGammas();
	return(TRUE);
}

void		GLimp_Shutdown( void )
{
	OpenGLSwapBuffers();
	DestroyMacGLWindow();
	
	if (gIsFullscreen == TRUE) {
		macds_Shutdown();
	}
	RestoreGammas();
}

int GLimp_SetMode( int *pwidth, int *pheight, int mode, qboolean fullscreen )
{
	int width, height;

	gIsFullscreen = fullscreen;

	ri.Con_Printf( PRINT_ALL, "Initializing OpenGL display\n");

	ri.Con_Printf (PRINT_ALL, "...setting mode %d:", mode );

	if ( !ri.Vid_GetModeInfo( &width, &height, mode ) )
	{
		ri.Con_Printf( PRINT_ALL, " invalid mode\n" );
		return rserr_invalid_mode;
	}

	ri.Con_Printf( PRINT_ALL, " %d %d\n", width, height );

	// destroy the existing window
	GLimp_Shutdown ();

	*pwidth = width;
	*pheight = height;
	
	if ( !GLimp_InitGraphics( fullscreen ) ) {
		// failed to set a valid mode in windowed mode
		return rserr_invalid_mode;
	}

	// let the sound and input subsystems know about the new window
	ri.Vid_NewWindow (width, height);
	return rserr_ok;

}

void		GLimp_AppActivate( qboolean active )
{
	if (gIsFullscreen == FALSE)
		return;

	if (active == TRUE)
	{
		macds_ActivateContext();
	}
	else
	{
		macds_PauseContext();
	}

}


/*
** GLimp_InitGraphics
**
** This initializes the GL implementation specific
** graphics subsystem.
**
** The necessary width and height parameters are grabbed from
** vid.width and vid.height.
*/
qboolean GLimp_InitGraphics( qboolean fullscreen )
{
	// if fullscreen, use DrawSprocket to switch resolutions for us
	if (!macds_InitContext(gl_vid.width,gl_vid.height,fullscreen)) {
		// if it fails, nuke the fullscreen setting
		return FALSE;
//		fullscreen = FALSE;
//		gIsFullscreen = FALSE;
//		ri.Cvar_SetValue( "vid_fullscreen", 0 );
	} else {
		macds_ActivateContext();		
	}

	// let the sound and input subsystems know about the new window
	ri.Vid_NewWindow (gl_vid.width, gl_vid.height);

	gFullscreen = fullscreen;
	IsMacGLActive = TRUE;
	return TRUE;
}