/* includes */
#include "rw_mac.h"
#include "client.h"
#include "keys.h"
#include "mac_isp.h"

// joystick defines and variables
// where should defines be moved?
#define JOY_ABSOLUTE_AXIS	0x00000000		// control like a joystick
#define JOY_RELATIVE_AXIS	0x00000010		// control like a mouse, spinner, trackball
#define	JOY_MAX_AXES		6				// X, Y, Z, R, U, V
#define JOY_AXIS_X			0
#define JOY_AXIS_Y			1
#define JOY_AXIS_Z			2
#define JOY_AXIS_R			3
#define JOY_AXIS_U			4
#define JOY_AXIS_V			5

cvar_t	*in_joystick;
cvar_t	*in_joystickabs;


// none of these cvars are saved over a session
// this means that advanced controller configuration needs to be executed
// each time.  this avoids any problems with getting back to a default usage
// or when changing from one controller to another.  this way at least something
// works.
cvar_t	*joy_name;
cvar_t	*joy_advanced;
cvar_t	*joy_advaxisx;
cvar_t	*joy_advaxisy;
cvar_t	*joy_advaxisz;
cvar_t	*joy_advaxisr;
cvar_t	*joy_advaxisu;
cvar_t	*joy_advaxisv;
cvar_t	*joy_forwardthreshold;
cvar_t	*joy_sidethreshold;
cvar_t	*joy_pitchthreshold;
cvar_t	*joy_yawthreshold;
cvar_t	*joy_forwardsensitivity;
cvar_t	*joy_sidesensitivity;
cvar_t	*joy_pitchsensitivity;
cvar_t	*joy_yawsensitivity;
cvar_t	*joy_upthreshold;
cvar_t	*joy_upsensitivity;
cvar_t *joy_invertx;
cvar_t *joy_inverty;

qboolean	joy_avail, joy_advancedinit, joy_haspov;

qboolean	in_appactive;

static int		jxv,jyv,joylookf;

/*
===========
Joy_AdvancedUpdate_f
===========
*/
static void Joy_AdvancedUpdate_f (void)
{
}

/* 
=============== 
IN_StartupJoystick 
=============== 
*/  
void IN_StartupJoystick (void) 
{ 
	cvar_t		*cv;

	jxv = 0;
	jyv = 0;
 	// assume no joystick
	joy_avail = FALSE; 

	// abort startup if user requests no joystick
	cv = Cvar_Get ("in_initjoy", "1", CVAR_NOSET);
	if ( !cv->value ) 
		return; 
 
	joy_avail = TRUE; 
	joy_advancedinit = FALSE;

	// joystick variables
	in_joystick				= Cvar_Get ("in_joystick",				"1",		CVAR_ARCHIVE);
	in_joystickabs			= Cvar_Get ("in_joystickabs",			"1",		CVAR_ARCHIVE);
	joy_invertx				= Cvar_Get ("joy_invertx",				"0",		CVAR_ARCHIVE);
	joy_inverty				= Cvar_Get ("joy_inverty",				"0",		CVAR_ARCHIVE);
	joy_name				= Cvar_Get ("joy_name",					"joystick",	0);
	joy_advanced			= Cvar_Get ("joy_advanced",				"0",		0);
	joy_advaxisx			= Cvar_Get ("joy_advaxisx",				"0",		0);
	joy_advaxisy			= Cvar_Get ("joy_advaxisy",				"0",		0);
	joy_advaxisz			= Cvar_Get ("joy_advaxisz",				"0",		0);
	joy_advaxisr			= Cvar_Get ("joy_advaxisr",				"0",		0);
	joy_advaxisu			= Cvar_Get ("joy_advaxisu",				"0",		0);
	joy_advaxisv			= Cvar_Get ("joy_advaxisv",				"0",		0);
	joy_forwardthreshold	= Cvar_Get ("joy_forwardthreshold",		"0.15",		0);
	joy_sidethreshold		= Cvar_Get ("joy_sidethreshold",		"0.15",		0);
	joy_upthreshold  		= Cvar_Get ("joy_upthreshold",			"0.15",		0);
	joy_pitchthreshold		= Cvar_Get ("joy_pitchthreshold",		"0.15",		0);
	joy_yawthreshold		= Cvar_Get ("joy_yawthreshold",			"0.15",		0);
	joy_forwardsensitivity	= Cvar_Get ("joy_forwardsensitivity",	"-1",		0);
	joy_sidesensitivity		= Cvar_Get ("joy_sidesensitivity",		"-1",		0);
	joy_upsensitivity		= Cvar_Get ("joy_upsensitivity",		"-1",		0);
	joy_pitchsensitivity	= Cvar_Get ("joy_pitchsensitivity",		"1",		0);
	joy_yawsensitivity		= Cvar_Get ("joy_yawsensitivity",		"-1",		0);
//	Cmd_AddCommand ("joy_advancedupdate", Joy_AdvancedUpdate_f);
}

/*
===========
IN_JoyMove
===========
*/
void IN_JoyMove (usercmd_t *cmd)
{
	float	speed, aspeed;
	float	fAxisValue;
//	int		i;

	// verify joystick is available and that the user wants to use it
	if (!joy_avail || !in_joystick->value)
	{
		return; 
	}
 
	// collect the joystick data, if possible

	if ( (in_speed.state & 1) ^ (int)cl_run->value)
		speed = 2;
	else
		speed = 1;
	aspeed = speed * cls.frametime;

	Input_ReadAxis(&jxv, &jyv, &joylookf);
	
	// X (turn) axis --------------------------------------------
	fAxisValue = (float) jxv;
	// convert range from -32768..32767 to -1..1 
	fAxisValue /= 32768.0;

	if ((in_strafe.state & 1) || (lookstrafe->value && mlooking))
	{
		// user wants turn control to become side control
		if (Burger::FPMath::Abs(fAxisValue) > joy_sidethreshold->value)
		{
			cmd->sidemove -= (fAxisValue * joy_sidesensitivity->value) * speed * cl_sidespeed->value;
		}
	}
	else
	{
		// user wants turn control to be turn control
		if (Burger::FPMath::Abs(fAxisValue) > joy_yawthreshold->value)
		{
			if (joy_invertx->value)
			{
				if(in_joystickabs->value)
				{
					cl.viewangles[YAW] += (((fAxisValue * joy_yawsensitivity->value * cls.scaleSensitivity) * aspeed * cl_yawspeed->value) * -1);
				}
				else
				{
					cl.viewangles[YAW] += (((fAxisValue * joy_yawsensitivity->value * cls.scaleSensitivity) * speed * 180.0) * -1);
				}
			}
			else
			{
				if(in_joystickabs->value)
				{
					cl.viewangles[YAW] += (fAxisValue * joy_yawsensitivity->value * cls.scaleSensitivity) * aspeed * cl_yawspeed->value;
				}
				else
				{
					cl.viewangles[YAW] += (fAxisValue * joy_yawsensitivity->value * cls.scaleSensitivity) * speed * 180.0;
				}
			}
		}
	}
	
	// Y (look) axis ---------------------------------------------

	fAxisValue = (float) jyv;
	// convert range from -32768..32767 to -1..1 
	fAxisValue /= 32768.0;
//	if (mlooking)
//	{
		if (Burger::FPMath::Abs(fAxisValue) > joy_pitchthreshold->value)
		{
			// pitch movement detected and pitch movement desired by user
			if (joy_inverty->value)
			{
				if(in_joystickabs->value)
				{
					cl.viewangles[PITCH] += (((fAxisValue * joy_pitchsensitivity->value * cls.scaleSensitivity) * aspeed * cl_pitchspeed->value) * -1);
				}
				else
				{
					cl.viewangles[PITCH] += (((fAxisValue * joy_pitchsensitivity->value * cls.scaleSensitivity) * speed * 180.0) * -1);
				}
			}
			else
			{
				if(in_joystickabs->value)
				{
					cl.viewangles[PITCH] += (fAxisValue * joy_pitchsensitivity->value * cls.scaleSensitivity) * aspeed * cl_pitchspeed->value;
				}
				else
				{
					cl.viewangles[PITCH] += (fAxisValue * joy_pitchsensitivity->value * cls.scaleSensitivity) * speed * 180.0;
				}
			}
		}
//	}
	
}

