/**********************************

	Our Mac implementation of the player input routines for Q2

**********************************/

#include <burger.h>
#include "client.h"
#include "keys.h"
#include "rw_mac.h"
#include "mac_isp.h"
#include "gl_local.h"

qboolean mlooking;

static in_state_t	*in_state;
static cvar_t 	*sensitivity;
static cvar_t 	*lookstrafe;
static cvar_t 	*m_side;
static cvar_t 	*m_yaw;
static cvar_t 	*m_pitch;
static cvar_t 	*m_forward;
static cvar_t 	*freelook;
static cvar_t	*m_filter;
static cvar_t	*in_mouse;

static qboolean mouse_avail;
static int		mx, my;
static int     	mouse_buttonstate;
static Word mouse_oldbuttonstate;
static int   	mouse_x, mouse_y;
static int		old_mouse_x, old_mouse_y;
static float 	old_windowed_mouse;
static int 		p_mouse_x, p_mouse_y;

/* keyboard event queue */

typedef struct KeyQueue_t {
	int key;			/* Key that the event is for */
	int down;			/* TRUE if pressed */
} KeyQueue_t;

static KeyQueue_t keyq[64];	/* Pending keyboard events */
static int keyq_head;		/* Head of the list */
static int keyq_tail;		/* End of the list */

// Var that keeps track of the last time we polled the mouse, added a refresh timer
// to ensure smooth mouse sampling of the delta so mouselook doesn't suck

static int gLastMouseSample = 0;


/* this is a table of three 128 entry sets, one for normal, one for capslock down
   and one for shiftkey pressed, translates the keycode to a ASCII or Quake char code */

static int MacScanTrans[] = {
// normal
'a',	// 0x00
's',	// 0x01
'd',	// 0x02
'f',	// 0x03
'h',	// 0x04
'g',	// 0x05
'z',	// 0x06
'x',	// 0x07
'c',	// 0x08
'v',	// 0x09
0xFF,	// 0x0A
'b',	// 0x0B
'q',	// 0x0C
'w',	// 0x0D
'e',	// 0x0E
'r',	// 0x0F
'y',	// 0x10
't',	// 0x11
'1',	// 0x12
'2',	// 0x13
'3',	// 0x14
'4',	// 0x15
'6',	// 0x16
'5',	// 0x17
'=',	// 0x18
'9',	// 0x19
'7',	// 0x1A
'-',	// 0x1B
'8',	// 0x1C
'0',	// 0x1D
']',	// 0x1E
'o',	// 0x1F
'u',	// 0x20
'[',	// 0x21
'i',	// 0x22
'p',	// 0x23
K_ENTER,	// 0x24
'l',	// 0x25
'j',	// 0x26
'\'',	// 0x27
'k',	// 0x28
';',	// 0x29
'\\',	// 0x2A
',',	// 0x2B
'/',	// 0x2C
'n',	// 0x2D
'm',	// 0x2E
'.',	// 0x2F
K_TAB,	// 0x30
' ',	// 0x31
'~',	// 0x32
K_BACKSPACE,	// 0x33
K_ENTER,	// 0x34
K_ESCAPE,	// 0x35
0xFF,	// 0x36
K_ALT,	// 0x37
K_SHIFT,	// 0x38
0xFF,	// 0x39
K_AUX32,		//option
K_CTRL,	// 0x3B
K_SHIFT,	// 0x3C
K_ALT,	// 0x3D
K_CTRL,	// 0x3E
0xff,	// 0x3f
0xff,	// 0x40
K_KP_DEL,	// 0x41
0xff,	// 0x42
'*',	// 0x43
0xff,	// 0x44
K_KP_PLUS,	// 0x45
0xFF,	// 0x46
0xFF,	// 0x47
0xFF,	// 0x48
0xFF,	// 0x49
0xFF,	// 0x4A
K_KP_SLASH,	// 0x4B
K_KP_ENTER,	// 0x4C
0xff,	// 0x4D
K_KP_MINUS,	// 0x4E
0xff,	// 0x4F
0xff,	// 0x50
'=',	// 0x51
K_KP_INS,	// 0x52
K_KP_END,	// 0x53
K_KP_DOWNARROW,	// 0x54
K_KP_PGDN,	// 0x55
K_KP_LEFTARROW,	// 0x56
K_KP_5,	// 0x57
K_KP_RIGHTARROW,	// 0x58
K_KP_HOME,	// 0x59
0xff,	// 0x5A
K_KP_UPARROW,	// 0x5B
K_KP_PGUP,	// 0x5C
0xff,	// 0x5D
0xff,	// 0x5E
0xff,	// 0x5F
K_F5,	// 0x60
K_F6,	// 0x61
K_F7,	// 0x62
K_F3,	// 0x63
K_F8,	// 0x64
K_F9,	// 0x65
0xff,	// 0x66
K_F11,	// 0x67
0xff,	// 0x68
0xff,	// 0x69
0xff,	// 0x6A
0xff,	// 0x6B
0xff,	// 0x6C
K_F10,	// 0x6D
0xff,	// 0x6E
K_F12,	// 0x6F
0xff,	// 0x70
K_PAUSE,	// 0x71
K_INS,	// 0x72
K_HOME,	// 0x73
K_PGUP,	// 0x74
K_DEL,	// 0x75
K_F4,	// 0x76
K_END,	// 0x77
K_F2,	// 0x78
K_PGDN,	// 0x79
K_F1,	// 0x7A
K_LEFTARROW,	// 0x7B
K_RIGHTARROW,	// 0x7C
K_DOWNARROW,	// 0x7D
K_UPARROW,	// 0x7E
0xff,	// 0x7F

// capslock
'A',	// 0x00
'S',	// 0x01
'D',	// 0x02
'F',	// 0x03
'H',	// 0x04
'G',	// 0x05
'Z',	// 0x06
'X',	// 0x07
'C',	// 0x08
'V',	// 0x09
0xFF,	// 0x0A
'B',	// 0x0B
'Q',	// 0x0C
'W',	// 0x0D
'E',	// 0x0E
'R',	// 0x0F
'Y',	// 0x10
'T',	// 0x11
'1',	// 0x12
'2',	// 0x13
'3',	// 0x14
'4',	// 0x15
'6',	// 0x16
'5',	// 0x17
'=',	// 0x18
'9',	// 0x19
'7',	// 0x1A
'-',	// 0x1B
'8',	// 0x1C
'0',	// 0x1D
']',	// 0x1E
'O',	// 0x1F
'U',	// 0x20
'[',	// 0x21
'I',	// 0x22
'P',	// 0x23
K_ENTER,	// 0x24
'L',	// 0x25
'J',	// 0x26
'\'',	// 0x27
'K',	// 0x28
';',	// 0x29
'\\',	// 0x2A
',',	// 0x2B
'/',	// 0x2C
'N',	// 0x2D
'M',	// 0x2E
'.',	// 0x2F
K_TAB,	// 0x30
' ',	// 0x31
'`',	// 0x32
K_BACKSPACE,	// 0x33
K_ENTER,	// 0x34
K_ESCAPE,	// 0x35
0xFF,	// 0x36
K_ALT,	// 0x37
K_SHIFT,	// 0x38
0xFF,	// 0x39
K_AUX32,		//option
K_CTRL,	// 0x3B
K_SHIFT,	// 0x3C
K_ALT,	// 0x3D
K_CTRL,	// 0x3E
0xff,	// 0x3f
0xff,	// 0x40
K_KP_DEL,	// 0x41
0xff,	// 0x42
'*',	// 0x43
0xff,	// 0x44
K_KP_PLUS,	// 0x45
0xFF,	// 0x46
0xFF,	// 0x47
0xFF,	// 0x48
0xFF,	// 0x49
0xFF,	// 0x4A
K_KP_SLASH,	// 0x4B
K_KP_ENTER,	// 0x4C
0xff,	// 0x4D
K_KP_MINUS,	// 0x4E
0xff,	// 0x4F
0xff,	// 0x50
'=',	// 0x51
K_KP_INS,	// 0x52
K_KP_END,	// 0x53
K_KP_DOWNARROW,	// 0x54
K_KP_PGDN,	// 0x55
K_KP_LEFTARROW,	// 0x56
K_KP_5,	// 0x57
K_KP_RIGHTARROW,	// 0x58
K_KP_HOME,	// 0x59
0xff,	// 0x5A
K_KP_UPARROW,	// 0x5B
K_KP_PGUP,	// 0x5C
0xff,	// 0x5D
0xff,	// 0x5E
0xff,	// 0x5F
K_F5,	// 0x60
K_F6,	// 0x61
K_F7,	// 0x62
K_F3,	// 0x63
K_F8,	// 0x64
K_F9,	// 0x65
0xff,	// 0x66
K_F11,	// 0x67
0xff,	// 0x68
0xff,	// 0x69
0xff,	// 0x6A
0xff,	// 0x6B
0xff,	// 0x6C
K_F10,	// 0x6D
0xff,	// 0x6E
K_F12,	// 0x6F
0xff,	// 0x70
K_PAUSE,	// 0x71
K_INS,	// 0x72
K_HOME,	// 0x73
K_PGUP,	// 0x74
K_DEL,	// 0x75
K_F4,	// 0x76
K_END,	// 0x77
K_F2,	// 0x78
K_PGDN,	// 0x79
K_F1,	// 0x7A
K_LEFTARROW,	// 0x7B
K_RIGHTARROW,	// 0x7C
K_DOWNARROW,	// 0x7D
K_UPARROW,	// 0x7E
0xff,	// 0x7F

//shift
'A',	// 0x00
'S',	// 0x01
'D',	// 0x02
'F',	// 0x03
'H',	// 0x04
'G',	// 0x05
'Z',	// 0x06
'X',	// 0x07
'C',	// 0x08
'V',	// 0x09
0xFF,	// 0x0A
'B',	// 0x0B
'Q',	// 0x0C
'W',	// 0x0D
'E',	// 0x0E
'R',	// 0x0F
'Y',	// 0x10
'T',	// 0x11
'!',	// 0x12
'@',	// 0x13
'#',	// 0x14
'$',	// 0x15
'^',	// 0x16
'%',	// 0x17
'=',	// 0x18
'(',	// 0x19
'&',	// 0x1A
'-',	// 0x1B
'*',	// 0x1C
')',	// 0x1D
'}',	// 0x1E
'O',	// 0x1F
'U',	// 0x20
'{',	// 0x21
'I',	// 0x22
'P',	// 0x23
K_ENTER,	// 0x24
'L',	// 0x25
'J',	// 0x26
'"',	// 0x27
'K',	// 0x28
':',	// 0x29
'|',	// 0x2A
'<',	// 0x2B
'?',	// 0x2C
'N',	// 0x2D
'M',	// 0x2E
'>',	// 0x2F
K_TAB,	// 0x30
' ',	// 0x31
'~',	// 0x32
K_BACKSPACE,	// 0x33
K_ENTER,	// 0x34
K_ESCAPE,	// 0x35
0xFF,	// 0x36
K_ALT,	// 0x37
K_SHIFT,	// 0x38
0xFF,	// 0x39
K_AUX32,		//option
K_CTRL,	// 0x3B
K_SHIFT,	// 0x3C
K_ALT,	// 0x3D
K_CTRL,	// 0x3E
0xff,	// 0x3f
0xff,	// 0x40
K_KP_DEL,	// 0x41
0xff,	// 0x42
'*',	// 0x43
0xff,	// 0x44
K_KP_PLUS,	// 0x45
0xFF,	// 0x46
0xFF,	// 0x47
0xFF,	// 0x48
0xFF,	// 0x49
0xFF,	// 0x4A
K_KP_SLASH,	// 0x4B
K_KP_ENTER,	// 0x4C
0xff,	// 0x4D
K_KP_MINUS,	// 0x4E
0xff,	// 0x4F
0xff,	// 0x50
'=',	// 0x51
K_KP_INS,	// 0x52
K_KP_END,	// 0x53
K_KP_DOWNARROW,	// 0x54
K_KP_PGDN,	// 0x55
K_KP_LEFTARROW,	// 0x56
K_KP_5,	// 0x57
K_KP_RIGHTARROW,	// 0x58
K_KP_HOME,	// 0x59
0xff,	// 0x5A
K_KP_UPARROW,	// 0x5B
K_KP_PGUP,	// 0x5C
0xff,	// 0x5D
0xff,	// 0x5E
0xff,	// 0x5F
K_F5,	// 0x60
K_F6,	// 0x61
K_F7,	// 0x62
K_F3,	// 0x63
K_F8,	// 0x64
K_F9,	// 0x65
0xff,	// 0x66
K_F11,	// 0x67
0xff,	// 0x68
0xff,	// 0x69
0xff,	// 0x6A
0xff,	// 0x6B
0xff,	// 0x6C
K_F10,	// 0x6D
0xff,	// 0x6E
K_F12,	// 0x6F
0xff,	// 0x70
K_PAUSE,	// 0x71
K_INS,	// 0x72
K_HOME,	// 0x73
K_PGUP,	// 0x74
K_DEL,	// 0x75
K_F4,	// 0x76
K_END,	// 0x77
K_F2,	// 0x78
K_PGDN,	// 0x79
K_F1,	// 0x7A
K_LEFTARROW,	// 0x7B
K_RIGHTARROW,	// 0x7C
K_DOWNARROW,	// 0x7D
K_UPARROW,	// 0x7E
0xff	// 0x7F
};

/**********************************

	Translates non-ASCII scancodes to the 'Quake' key scancodes so the
	engine knows what the heck we're pressing

**********************************/

static int XLateKey(int theKey)
{
	int scancode;
	int ret;
	int isAlpha;

	if (theKey <= 127) {	/* must be 0-127 */
		scancode = theKey;	/* set the base scancode */

		isAlpha = MacScanTrans[scancode];
		if ( (isAlpha >= 'a') && (isAlpha <= 'z') ) {
			return(isAlpha);	/* don't translate a-z */
		}
		if ( (( KeyArray[SC_LEFTSHIFT] | KeyArray[SC_RIGHTSHIFT] ) & 1)) {
			scancode += 256;	/* if shift key pressed, access the third set */
		} else if (KeyArray[SC_CAPSLOCK] & 1) {
			scancode += 128;	/* if capslock down, access the second set */
		}
		ret = MacScanTrans[scancode];	/* convert to ASCII or a Quake code */
		if (ret != 0xFF) {				/* filter the not used to null */
			return(ret);
		}
	}
	return 0;
}

/*****************************************************************************/
/* KEYBOARD                                                                  */
/*****************************************************************************/

Key_Event_fp_t Key_Event_fp;

void KBD_Init(Key_Event_fp_t fp)
{
	Key_Event_fp = fp;
}

void KBD_Update(void)
{
	int eventtype;
	int thekey;
	
	if (gIn_Active == FALSE) {	/* if backgrounded, just bail and don't update events */
		return;
	}

	eventtype = KeyboardGet2();
	if (eventtype) {
		do {
			thekey = ((eventtype >> 8)&0x7f);
			if (eventtype&0x8000) {
				eventtype = FALSE;
			} else {
				eventtype = TRUE;
			}
			if (Key_Event_fp) {		/* Failsafe */
				Key_Event_fp(XLateKey(thekey),eventtype);
			}
			eventtype = KeyboardGet2();				/* Any more keys? */
		} while (eventtype);
	}
}

void KBD_Flush(void)
{
	KeyboardFlush();
}

void KBD_Close(void)
{
	KeyboardFlush();
}


static void Force_CenterView_f (void)
{
	in_state->viewangles[PITCH] = 0;
}

static void RW_IN_MLookDown (void) 
{ 
	mlooking = TRUE; 
}

static void RW_IN_MLookUp (void) 
{
	mlooking = FALSE;
	if (lookspring->value) {
		in_state->IN_CenterView_fp ();
	}
}

void RW_IN_Init(in_state_t *in_state_p)
{
//	int mtype;
//	int i;

	//fprintf(stderr, "GL RW_IN_Init\n");

	in_state = in_state_p;

	// mouse variables
// NCT_CHANGE, removed ri. since this isn't linked in with the renderer	
	
	m_filter = Cvar_Get ("m_filter", "0", CVAR_ARCHIVE);
    in_mouse = Cvar_Get ("in_mouse", "0", 0); //NCT_CHANGE no more CVAR_ARCHIVE
	in_mouse->value = 0; // NCT_ADD
	freelook = Cvar_Get( "freelook", "0", 0 );
	lookstrafe = Cvar_Get ("lookstrafe", "0", 0);
	sensitivity = Cvar_Get ("sensitivity", "11", CVAR_ARCHIVE);	// 1.0.3 fix!
	m_pitch = Cvar_Get ("m_pitch", "0.022", 0);
	m_yaw = Cvar_Get ("m_yaw", "0.022", 0);
	m_forward = Cvar_Get ("m_forward", "1", 0);
	m_side = Cvar_Get ("m_side", "0.8", 0);

	Cmd_AddCommand ("+mlook", RW_IN_MLookDown);
	Cmd_AddCommand ("-mlook", RW_IN_MLookUp);

	Cmd_AddCommand ("force_centerview", Force_CenterView_f);

	Cmd_AddCommand ("desktop", Mac_Desktop_f);
	Cmd_AddCommand ("resume", Mac_Resume_f);
	Cmd_AddCommand ("input", Input_ShowConfigureDialog);
	
	mouse_x = mouse_y = 0.0;
	mouse_avail = TRUE;
	
	IN_StartupJoystick();	//NCT_AXIS
}

void RW_IN_Shutdown(void)
{
	mouse_avail = FALSE;

	Cmd_RemoveCommand ("force_centerview");
	Cmd_RemoveCommand ("+mlook");
	Cmd_RemoveCommand ("-mlook");
	
	Cmd_RemoveCommand ("desktop");
	Cmd_RemoveCommand ("resume");
	Cmd_RemoveCommand ("input");
}

/*
===========
IN_Commands
===========
*/
void RW_IN_Commands (void)
{
	Word i;
	Word NewMouse;
	Word MouseDown;
	Word MouseUp;
	Word Mask;
	int Wheel;
	
	if (mouse_avail) {

		/* translate mouse button state vars to keypress events using mousebutton codes */

		NewMouse = MouseReadButtons();					/* Read the mouse */
		MouseUp = NewMouse^mouse_oldbuttonstate;		/* Which changed? */
		MouseDown = MouseUp&NewMouse;					/* Get the mouse down events */
		MouseUp = MouseUp&(~MouseDown);					/* Get the mouse up events */
		mouse_oldbuttonstate = NewMouse;				/* Save the current mouse state */

		Mask = MouseClicked&(~MouseDown);				/* Get the extra mouse down events */
		MouseDown |= Mask;								/* Send a down and an up */
		MouseUp |= Mask;
		i = 0;
		Mask = 1;
		do {
			Word Index;
			if (i<3) {
				Index = K_MOUSE1+i;
			} else {
				Index = (K_AUX1-3)+i;
			}
			if (MouseDown&Mask) {
				in_state->Key_Event_fp(Index, TRUE);
			}
			if (MouseUp&Mask) {
				in_state->Key_Event_fp(Index, FALSE);
			}
			Mask<<=1;
		} while (++i<32);
		MouseClicked = 0;		/* Ack the mouse down's */
		Wheel = MouseReadWheel();
		if (Wheel) {
			if (Wheel>0) {
				in_state->Key_Event_fp(K_MWHEELUP, TRUE);
				in_state->Key_Event_fp(K_MWHEELUP, FALSE);
			} else {
				in_state->Key_Event_fp(K_MWHEELDOWN, TRUE);
				in_state->Key_Event_fp(K_MWHEELDOWN, FALSE);
			}
		}
	}
}

/*
===========
IN_Move
===========
*/
void RW_IN_Move (usercmd_t *cmd)
{
	int delx,dely,lookflag;

	IN_JoyMove(cmd);	// NCT_AXIS

	if (!mouse_avail) {
		return;
	}
 
// NCT_MOVE 7/9/99

	Input_ReadMouseDeltas(&delx,&dely,&lookflag);
	if (freelook->value != lookflag) {
		Cvar_SetValue( "freelook", lookflag );
	}
	mx = delx;
	my = dely;
  
   	/* translate mouse deltas to some sort of movement action */
   
	if (m_filter->value) {
		mouse_x = (mx + old_mouse_x) * 0.5f;
		mouse_y = (my + old_mouse_y) * 0.5f;
	} else {
		mouse_x = mx;
		mouse_y = my;
	}
	old_mouse_x = mx;
	old_mouse_y = my;

	if (!mouse_x && !mouse_y)
		return;

	mouse_x *= sensitivity->value;  
	mouse_y *= sensitivity->value;	

// add mouse X/Y movement to cmd
	if ( (*in_state->in_strafe_state & 1) || 
		(lookstrafe->value && mlooking ))
		cmd->sidemove += m_side->value * mouse_x;
	else
		in_state->viewangles[YAW] -= m_yaw->value * mouse_x;

	if ( (mlooking || freelook->value) && 
		!(*in_state->in_strafe_state & 1)) {
		in_state->viewangles[PITCH] += m_pitch->value * mouse_y;
	} else {
		cmd->forwardmove -= m_forward->value * mouse_y;
	}
	mx = my = 0;
}

void RW_IN_Frame (void)
{
}

void RW_IN_Activate(qboolean active)
{
}


