#include "macfuncs.h"
#include "rw_mac.h"
#include "vid.h"
#include <ToolUtils.h>
#include <DrawSprocket.h>
#include <DiskInit.h>

#ifdef __cplusplus
extern "C" {
#endif
extern DSpContextReference MacContext;	/* Reference to a reserved DrawSprocket reference */
#ifdef __cplusplus
}
#endif

static int gBackgroundSetup = FALSE;

int gResumeSelected = FALSE;
int gQuitSelected = FALSE;
static int gIsSuspended = FALSE;

enum {
	kMBAR_Main			=	128,
	kMENU_Apple			=	128,
	kMENU_File,
	kMENU_Edit
};

static void DoDiskEvent(EventRecord *theEvent)
{
#if !TARGET_API_MAC_CARBON
	Point	thePoint;
	SInt16	value;

	if (theEvent->message & 0xFFFF0000) {		//	Error mounting disk
		SetPt(&thePoint, 100, 100);
		value = DIBadMount(thePoint, theEvent->message);
	}
#endif
}

static void DoDefaultKeyDown(EventRecord *inEvent)
{
	SInt8 theKey;
	SInt8 theCode;
	SInt32 menuResult;

	theKey = inEvent->message & charCodeMask;
	theCode = (inEvent->message & keyCodeMask) >> 8;
	if ((inEvent->modifiers & cmdKey) != 0) {
		Mac_MenuPrepForSelect();
		menuResult = MenuKey(theKey);
		if (HiWord(menuResult) != 0) {
			if (Mac_MenuBarDispatch(menuResult, inEvent->modifiers)) {
				return;
			}
		}
	}
}

static void DoDefaultMouseDown(EventRecord *inEvent)
{
	SInt16 whatPart;
	SInt32 menuResult;
	WindowPtr whichWindow;

	whatPart = FindWindow(inEvent->where, &whichWindow);

	switch (whatPart) {
	case inMenuBar:
		Mac_MenuPrepForSelect();
		menuResult = MenuSelect(inEvent->where);

		if (HiWord(menuResult) != 0) {
			Mac_MenuBarDispatch(menuResult, inEvent->modifiers);
		}
		break;

	case inSysWindow:
#if !TARGET_API_MAC_CARBON
		SystemClick(inEvent, whichWindow);
#endif
		break;
	case inDrag:
		{
#if !TARGET_API_MAC_CARBON
			Rect Bounds;
			Bounds = GetGrayRgn()[0]->rgnBBox;
#else
			Rect Bounds;
			GetRegionBounds(GetGrayRgn(),&Bounds);
#endif
			DragWindow(whichWindow, inEvent->where, &Bounds);
			break;
		}
	}
	return;
}

static void DoResume(void)
{
	gIsSuspended = FALSE;
}

static void DoSuspend(void)
{
	gIsSuspended = TRUE;
}


static int MacOSBackgroundEvent(void)
{
	EventRecord theEvent;
	Bool wasProcessed;

	WaitNextEvent(everyEvent, &theEvent, 32767, NULL);		

// This is a bit of a hack, since we're assuming all the rendering libraries use and
// initialize DrawSprocket (which they do), we'll at least check to make sure it is
// installed on the system
// Added check to see if vid_fullscreen is set, if it ain't then we don't check for DrawSprocket

	if (VID_IsFullscreen() == TRUE) {
		if (MacContext) {
			// See if DrawSprocket will munch the event
			DSpProcessEvent(&theEvent, &wasProcessed);
			if (wasProcessed == TRUE) {
				return(gResumeSelected);
			}
		}
	}

	switch (theEvent.what) {

	case mouseDown:
		DoDefaultMouseDown(&theEvent);
		break;
	
	case keyDown:
		 autoKey:
		DoDefaultKeyDown(&theEvent);
		break;

	case osEvt:
		if (theEvent.message & 0x01000000) {		//	Suspend/resume event
			if (theEvent.message & 0x00000001) {	//	Resume
				DoResume();
			} else {
				DoSuspend();					//	Suspend
			}
		}
		break;

	case diskEvt:
		DoDiskEvent(&theEvent);
		break;	
	case kHighLevelEvent:
		AEProcessAppleEvent(&theEvent);
		break;
	}
	return(gResumeSelected);		
}

// Need to implement IN_Activate and check for if active in IN_Frame (see in_win.c)


int MacOSBackground(void)
{
	int doResume;
	MenuHandle myMH;

	if (gBackgroundSetup == FALSE) {	//do we need to setup background environment?
		myMH = GetMenuHandle(kMENU_File);
		if (myMH) {
			EnableMenuItem(myMH, 1);
		}
		// Do things like pause video context, suspend InputSprocket if active, etc.
		InitBackground();

		gBackgroundSetup = TRUE;
		gResumeSelected = FALSE;
		gQuitSelected = FALSE;
	}
	ShowCursor();

	doResume = MacOSBackgroundEvent();

	if (doResume == TRUE) {
		myMH = GetMenuHandle(kMENU_File);
		if (myMH) {
			DisableMenuItem(myMH, 1);
		}
		gBackgroundSetup = FALSE;
		HideCursor();
		KillBackground();
	}
	return(doResume);		// maintain status quo..or not
}




