//		Includes

#define USE_OLD_INPUT_SPROCKET_LABELS 0
#define USE_OLD_ISPNEED_STRUCT 0

#include <burger.h>
#include <InputSprocket.h>
#include <CodeFragments.h>
#include <FixMath.h>
#include <TextUtils.h>

#include "mac_ispdata.h"
#include "mac_isp.h"
#include "vid.h"
#include "snd_sound.h"

/*
pretty stock bindlist

bind TAB "inven"
bind ENTER "invuse"
bind ESCAPE "togglemenu"
bind SPACE "+moveup"
bind ' "invdrop"
bind + "sizeup"
bind , "+moveleft"
bind - "sizedown"
bind . "+moveright"
bind / "weapnext"
bind 0 "use BFG10K"
bind 1 "use Blaster"
bind 2 "use Shotgun"
bind 3 "use Super Shotgun"
bind 4 "use Machinegun"
bind 5 "use Chaingun"
bind 6 "use Grenade Launcher"
bind 7 "use Rocket Launcher"
bind 8 "use HyperBlaster"
bind 9 "use Railgun"
bind = "sizeup"
bind [ "invprev"
bind \ "+mlook"
bind ] "invnext"
bind ` "toggleconsole"
bind a "+moveleft"
bind b "use rebreather"
bind c "+movedown"
bind d "+moveright"
bind e "use environment suit"
bind g "use grenades"
bind h "wave 0"
bind i "use invulnerability"
bind j "wave 1"
bind k "wave 2"
bind l "wave 3"
bind p "use power shield"
bind q "use quad damage"
bind s "+back"
bind t "messagemode"
bind u "wave 4"
bind w "+forward"
bind x "score"
bind z "+lookdown"
bind ~ "toggleconsole"
bind BACKSPACE "invdrop"
bind UPARROW "+forward"
bind DOWNARROW "+back"
bind LEFTARROW "+left"
bind RIGHTARROW "+right"
bind ALT "+strafe"
bind CTRL "+attack"
bind SHIFT "+speed"
bind F1 "cmd help"
bind F2 "menu_savegame"
bind F3 "menu_loadgame"
bind F4 "menu_keys"
bind F5 "menu_startserver"
bind F6 "echo Quick Saving...; wait; save quick"
bind F9 "echo Quick Loading...; wait; load quick"
bind F10 "menu_quit"
bind F12 "screenshot"
bind INS "+klook"
bind DEL "+lookdown"
bind PGDN "+lookup"
bind END "centerview"
bind MOUSE1 "+attack"
bind MOUSE2 "+moveup"
bind MOUSE3 "+forward"
bind PAUSE "pause"

*/

static char *cmdstrlist[] = {
	"+attack",
	"weapnext",
	"+forward",
	"+back",
	"+left",
	"+right",
	"+speed",
	"+moveleft",
	"+moveright",
	"+strafe",
	"+lookup",
	"+lookdown",
	"centerview",
	"+mlook",
	"+klook",
	"+moveup",
	"+movedown",
	"inven",
	"invuse",
	"invdrop",
	"invprev",
	"invnext",
	"cmd help",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL, // NCT_AXIS
	NULL, // NCT_AXIS
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
//	NULL,
	NULL	//"menu_quit"	
};

//		Private Definitions
//		Private Types
//		Private Variables
static	Bool gISPInitialized = FALSE;
static	Bool gISPActive = FALSE;
Bool	gKeyboardTypingEnabled = TRUE;
Bool	gMouseAsMacOSCursorEnabled = TRUE;

ISpElementReference			gInputElements[kNeed_NeedCount];
ISpElementListReference 	gEventsElementList = NULL;

//		ISpAxisToSampleAxis
static SInt32	ISpAxisToSampleAxis (ISpAxisData axisValue, SInt32 min, SInt32 max)
{
	UInt32 divisor;

	divisor = kISpAxisMaximum / (max - min);
	return ((axisValue / divisor) + min);
}

//		Private Functions

Bool	Input_ISpConfigureEventProc (EventRecord * inEvent);
void	Input_DisableMouseForCursor (void);

//		Public Variables

//		Input_Available
int Input_Available (void)
{
	if (Burger::Mac::GetInputSprocketVersion()>=0x140) {		/* Require 1.4 or later */
		return TRUE;
	}
	return FALSE;
}

//		Input_Initialize
int Input_Initialize (void)
{
	OSStatus	theError;

	//	This structure defines the input needs that we'll be requesting from InputSprocket
	ISpNeed		tempNeed;
	ISpNeed		myNeeds[kNeed_NeedCount];

	if (gISPInitialized) {
		return(-1);
	}
	
	FastMemSet(myNeeds,0,sizeof(myNeeds));
	FastMemSet(&tempNeed,0,sizeof(tempNeed));
	tempNeed.reserved1 = 0;
	tempNeed.reserved2 = 0;
	tempNeed.reserved3 = 0;

	// we'll init all the player 1 items now (everything but quit unless we add a second player)
	tempNeed.playerNum = 0;
	
	// First we init (player 1) items that are not part of a group (group 0)
	tempNeed.group = 0;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Attack + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Attack;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_Fire;
	tempNeed.flags = kISpNeedFlag_Button_ActiveWhenDown | kISpNeedFlag_EventsOnly;
	myNeeds[kNeed_Attack] = tempNeed;

//	tempNeed.flags = kISpNeedFlag_EventsOnly;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_NextW + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_NextW;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_Next;
	myNeeds[kNeed_NextW] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_WalkF + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_WalkF;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_MoveForward;
	myNeeds[kNeed_WalkF] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_BackP + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_BackP;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_MoveBackward;
	myNeeds[kNeed_BackP] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_TurnLeft + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_TurnLeft;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_TurnLeft;
	myNeeds[kNeed_TurnLeft] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_TurnRight + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_TurnRight;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_TurnRight;
	myNeeds[kNeed_TurnRight] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Run + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Run;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_Run;
	myNeeds[kNeed_Run] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_StepLeft + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_StepLeft;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_SlideLeft;
	myNeeds[kNeed_StepLeft] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_StepRight + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_StepRight;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_SlideRight;
	myNeeds[kNeed_StepRight] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Sidestep + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Sidestep;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_SideStep;
	myNeeds[kNeed_Sidestep] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_LookUp + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_LookUp;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_LookUp;
	myNeeds[kNeed_LookUp] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_LookDown + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_LookDown;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_LookDown;
	myNeeds[kNeed_LookDown] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_CenterView + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_CenterView;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_CenterView] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Mouselook + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Mouselook;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_Look;
	myNeeds[kNeed_Mouselook] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Keylook + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Keylook;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Keylook] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_UpJump + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_UpJump;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_Btn_Jump;
	myNeeds[kNeed_UpJump] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_DownCrouch + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_DownCrouch;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_DownCrouch] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Inventory + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Inventory;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Inventory] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_UseItem + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_UseItem;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_UseItem] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_DropItem + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_DropItem;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_DropItem] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_PrevItem + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_PrevItem;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_PrevItem] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_NextItem + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_NextItem;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_NextItem] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Help + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Help;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Help] = tempNeed;

	// Now group 1, which is for changing Move/Look
	tempNeed.group = 1;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Walk + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Walk;
	tempNeed.theKind = kISpElementKind_Delta;
	tempNeed.theLabel = kISpElementLabel_Delta_Z;
	tempNeed.flags = 0;
	myNeeds[kNeed_Walk] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Look + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Look;
	tempNeed.theKind = kISpElementKind_Delta;
	tempNeed.theLabel = kISpElementLabel_Delta_Y;
	tempNeed.flags = kISpNeedFlag_Delta_AlreadyAxis; // NCT_AXIS
	myNeeds[kNeed_Look] = tempNeed;

// NCT_AXIS
	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_LookAxis + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_LookAxis;
	tempNeed.theKind = kISpElementKind_Axis;
	tempNeed.theLabel = kISpElementLabel_Axis_Pitch;
	tempNeed.flags = kISpNeedFlag_Axis_AlreadyDelta;
	myNeeds[kNeed_LookAxis] = tempNeed;
	
	// Now group 2, which is for changing turn
	tempNeed.group = 2;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Turn + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Turn;
	tempNeed.theKind = kISpElementKind_Delta;
	tempNeed.theLabel = kISpElementLabel_Delta_X;
	tempNeed.flags = kISpNeedFlag_Delta_AlreadyAxis;	// NCT_AXIS
	myNeeds[kNeed_Turn] = tempNeed;

// NCT_AXIS
	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_TurnAxis + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_TurnAxis;
	tempNeed.theKind = kISpElementKind_Axis;
	tempNeed.theLabel = kISpElementLabel_Axis_Roll;
	tempNeed.flags = kISpNeedFlag_Axis_AlreadyDelta;
	myNeeds[kNeed_TurnAxis] = tempNeed;
	
	// Now group 3, which is aux button crap
	tempNeed.group = 3;
	tempNeed.flags = kISpNeedFlag_EventsOnly;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Aux25 + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Aux25;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Aux25] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Aux26 + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Aux26;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Aux26] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Aux27 + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Aux27;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Aux27] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Aux28 + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Aux28;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Aux28] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Aux29 + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Aux29;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Aux29] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Aux30 + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Aux30;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Aux30] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Aux31 + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Aux31;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Aux31] = tempNeed;

	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Aux32 + 1);
	tempNeed.iconSuiteResourceId = 1000 + kNeed_Aux32;
	tempNeed.theKind = kISpElementKind_Button;
	tempNeed.theLabel = kISpElementLabel_None;
	myNeeds[kNeed_Aux32] = tempNeed;
	
	// now we init utility needs that affect all players (just have quit for now)
//	tempNeed.playerNum = 0;
//	tempNeed.group = 0;

//	GetIndString (tempNeed.name, kSTRn_NeedNames, kNeed_Quit + 1);
//	tempNeed.iconSuiteResourceId = 1000 + kNeed_Quit;
//	tempNeed.theKind = kISpElementKind_Button;
//	tempNeed.theLabel = kISpElementLabel_Btn_Quit;
//	tempNeed.flags = kISpNeedFlag_Utility | kISpNeedFlag_EventsOnly;
//	myNeeds[kNeed_Quit] = tempNeed;
	
	// Alright, now that the array is set up, we can call ISp to init stuff

#if 0			/* Burgerlib already starts this... */
	theError = ISpStartup ();
	if (theError) {
		DebugString("Could not initialize InputSprocket\n");
		return(theError);
	}
#endif

	//	Setup the input sprocket elements
	theError = ISpElement_NewVirtualFromNeeds(kNeed_NeedCount, myNeeds, gInputElements, 0);
	if (theError) {
		DebugString("Could not create ISp virtual controls from needs\n");
		return(theError);
	}

	//	Init InputSprocket and tell it our needs
	theError = ISpInit (kNeed_NeedCount, myNeeds, gInputElements, 'SiN!', 'DEFS', 0, 128, 0); 
	if (theError)
	{
		DebugString("Could not initialize high-level ISp\n");
		return(theError);
	}

	// Create a element list containg all the 'normal' buttons (we get events on these)
	theError = ISpElementList_New(0, NULL, &gEventsElementList, 0);
	if (theError)
	{
		DebugString("Could not create button element list\n");
		return(theError);
	}
	
	// we set the refcon to the need enum value, so we can use it later
	// doing some shortcut error checking for readability
	theError  = ISpElementList_AddElements (gEventsElementList, 
					kNeed_Attack, 			1, &gInputElements[kNeed_Attack]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_NextW, 			1, &gInputElements[kNeed_NextW]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_WalkF, 			1, &gInputElements[kNeed_WalkF]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_BackP, 			1, &gInputElements[kNeed_BackP]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_TurnLeft, 		1, &gInputElements[kNeed_TurnLeft]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_TurnRight, 		1, &gInputElements[kNeed_TurnRight]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Run, 				1, &gInputElements[kNeed_Run]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_StepLeft, 		1, &gInputElements[kNeed_StepLeft]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_StepRight, 		1, &gInputElements[kNeed_StepRight]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Sidestep, 		1, &gInputElements[kNeed_Sidestep]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_LookUp, 			1, &gInputElements[kNeed_LookUp]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_LookDown, 		1, &gInputElements[kNeed_LookDown]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_CenterView, 		1, &gInputElements[kNeed_CenterView]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Mouselook, 		1, &gInputElements[kNeed_Mouselook]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Keylook, 			1, &gInputElements[kNeed_Keylook]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_UpJump, 			1, &gInputElements[kNeed_UpJump]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_DownCrouch, 		1, &gInputElements[kNeed_DownCrouch]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Inventory, 		1, &gInputElements[kNeed_Inventory]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_UseItem, 			1, &gInputElements[kNeed_UseItem]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_DropItem, 		1, &gInputElements[kNeed_DropItem]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_PrevItem, 		1, &gInputElements[kNeed_PrevItem]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_NextItem, 		1, &gInputElements[kNeed_NextItem]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Help,	 			1, &gInputElements[kNeed_Help]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Aux25, 			1, &gInputElements[kNeed_Aux25]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Aux26, 			1, &gInputElements[kNeed_Aux26]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Aux27, 			1, &gInputElements[kNeed_Aux27]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Aux28, 			1, &gInputElements[kNeed_Aux28]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Aux29, 			1, &gInputElements[kNeed_Aux29]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Aux30, 			1, &gInputElements[kNeed_Aux30]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Aux31, 			1, &gInputElements[kNeed_Aux31]);
	theError  |= ISpElementList_AddElements (gEventsElementList, 
					kNeed_Aux32, 			1, &gInputElements[kNeed_Aux32]);
//	theError  |= ISpElementList_AddElements (gEventsElementList, 
//					kNeed_Quit, 			1, &gInputElements[kNeed_Quit]);					

	if (theError) {
		DebugString("Could not fill button element list. Error number may be inaccurate\n");
		return(theError);
	}
	
	
	// we leave kayboard and mouse disabled. We will enable them when the 'game' is active
	
	gISPInitialized = TRUE;

	//	Enable speech input (we commit to calling ISpTickle at WNE time)
 	//ISpDevices_ActivateClass (kISpDeviceClass_SpeechRecognition);
	
	Input_DisableMouseForCursor();

	return noErr;
}
	
//		Input_Suspend
int Input_Suspend (void)
{
	ShowCursor();
	if (!gISPInitialized) {
		return(-1);
	}
	InputSetState(FALSE);
	return FALSE;
}

//		Input_Resume
int Input_Resume (void)
{
	HideCursor();
	if (!gISPInitialized) {
		return(-1);
	}
	InputSetState(TRUE);
	return 0;
}

//		Input_ShowConfigureDialog
void	Input_ShowConfigureDialog (void)
{

	if (!gISPInitialized)
		return;
	// we want to configure the mouse and keyboard, so they must be on before configure

	S_StopAllSounds ();

	ShowCursor();
	ToggleVoodoo(0);
	ISpConfigure (&Input_ISpConfigureEventProc);
	ToggleVoodoo(1);  // 1.0.3, was 0, doh!
	HideCursor();
	
	// NCT_ADD update the screen
#if !TARGET_API_MAC_CARBON
	{
		GrafPtr savePort,wPort;
		GetPort(&savePort);
		GetWMgrPort(&wPort);
		SetPort(wPort);
		PenPat(&qd.black);
		PaintRgn(GetGrayRgn());
		SetPort(savePort);	
	}
#endif
	// restore the mouse and keyboard, so we can use for MacOS things
}

//		Input_ISpConfigureEventProc (private)
Bool  Input_ISpConfigureEventProc (EventRecord * inEvent)
{
#pragma unused(inEvent)
	// maintain network connections, etc here
	
	return FALSE;
}

void Input_ReadAxis(int *axisx, int *axisy, int *lookflag)
{
	// axis
	ISpAxisData axisValue;
	SInt32 Xvalue,Yvalue;
	//int axisLive;
	//float temp;
	OSStatus error;
	ISpElementEvent event;
	Bool wasEvent;

	Xvalue = *axisx;
	Yvalue = *axisy;

// axis
	if (gISPInitialized) {
		// we check the axis, to see if _it_ was moved, if so, we use that value
		error = ISpElement_GetNextEvent (gInputElements[kNeed_TurnAxis], sizeof (event), &event, &wasEvent);
		if (!error && wasEvent) {
			error = ISpElement_GetSimpleState(gInputElements[kNeed_TurnAxis], &axisValue);
			if (!error) {
				Xvalue = ISpAxisToSampleAxis (axisValue, -32768, 32767);
			}
			ISpElement_Flush(gInputElements[kNeed_TurnAxis]);
		}

		// we check the axis, to see if _it_ was moved, if so, we use that value
		error = ISpElement_GetNextEvent (gInputElements[kNeed_LookAxis], sizeof (event), &event, &wasEvent);
		if (!error && wasEvent) {
			error = ISpElement_GetSimpleState(gInputElements[kNeed_LookAxis], &axisValue);
			if (!error) {
				Yvalue = ISpAxisToSampleAxis (axisValue, -32768, 32767);
			}
			ISpElement_Flush(gInputElements[kNeed_LookAxis]);
		}
	}
	*axisx = Xvalue;
	*axisy = Yvalue;
	*lookflag = TRUE;
}

#define PIXPERINCH 600.0f	//600 dpi on conversion of mouse motion
// ------------------------------------------------------------
int Input_ReadMouseDeltas(int *delx, int *dely, int *lookflag)
{
	// delta
	Fixed32 moveDelta,turnDelta,lookDelta;
	float temp;
	OSStatus error;
	ISpElementEvent event;
	Bool wasEvent;

	if (!gISPInitialized) {
		int Burgerx,Burgery;
		MouseReadDelta(&Burgerx,&Burgery);
		moveDelta = 0;
		turnDelta = Burgerx*163;
		lookDelta = Burgery*-163;
		
	} else {
	
// delta, yeah baby....
	
		moveDelta = 0;
		turnDelta = 0;
		lookDelta = 0;
		do {
			error = ISpElement_GetNextEvent(gInputElements[kNeed_Walk], sizeof(event),&event,&wasEvent);
			if (wasEvent && !error) {
				moveDelta += (Fixed32) event.data;
			}
		} while (wasEvent && !error);

		do {
			error = ISpElement_GetNextEvent(gInputElements[kNeed_Turn], sizeof(event),&event,&wasEvent);
			if (wasEvent && !error) {
				turnDelta += (Fixed32) event.data;
			}
		} while (wasEvent && !error);

		do {
			error = ISpElement_GetNextEvent(gInputElements[kNeed_Look], sizeof(event),&event,&wasEvent);
			if (wasEvent && !error) {
				lookDelta += (Fixed32) event.data;
			}
		} while (wasEvent && !error);
	}	
	/* turn delta is easy, no choice there.. */

	temp = FIXEDTOFLOAT(turnDelta);
	temp *= PIXPERINCH;			/* convert to pixels */
	*delx = (int)temp;			/* grab int part */
	if ( (lookDelta == 0) && (moveDelta != 0) ) {
		temp = FIXEDTOFLOAT(moveDelta);
		temp = -temp;				/* it's inverse of what we expect */
		temp *= PIXPERINCH;			/* convert to pixels */
		*dely = (int)temp;			/* grab int part */
		*lookflag = 0;
	} else {
		temp = FIXEDTOFLOAT(lookDelta);
		temp = -temp;				/* it's inverse of what we expect */
		temp *= PIXPERINCH;			/* convert to pixels */
		*dely = (int)temp;			/* grab int part */
		*lookflag = 1;
	}
	return 0;
}


//		Input_GetButtonEvents
void Input_GetButtonEvents (unsigned time)
{
	OSStatus			error = noErr;
	ISpElementEvent		event;
	Bool 			wasEvent;
	char 				cmd[1024];
	char				*kb;
	int 				offset;

	if (!gISPInitialized) {
		return;
	}
	
	// give time to some non-interrupt driven input drivers (like speech recognition)
	ISpTickle ();
	
	// get all pending events
	do {
		error = ISpElementList_GetNextEvent (gEventsElementList, sizeof (event), &event, &wasEvent);
		
		if (wasEvent && !error) {
			/* get the pointer to the string signifying the command to give Quake */
			if ( (event.refCon >= kNeed_Attack) && (event.refCon < kNeed_NeedCount) ) {
				kb = cmdstrlist[event.refCon];

				if (event.data == kISpButtonDown) {
					if (kb) {
						if (kb[0] == '+') {	// button commands add keynum and time as a parm
							// using event.Refcon for key
							Com_sprintf (cmd, sizeof(cmd), "%s %i %i\n", kb, event.refCon+1, time);
							Cbuf_AddText (cmd);
						} else {
							Cbuf_AddText (kb);
							Cbuf_AddText ("\n");
						}					
					} else {
						// Check if aux range and we fake a keypress
						if ( (event.refCon >= kNeed_Aux25) && (event.refCon <= kNeed_Aux32) ) {							
							offset = event.refCon - kNeed_Aux25;
							
							/* Actually adding from Aux24 since I goofed and forgot I
							   reassigned Aux32 to the mac 'option' key */
							offset += K_AUX24;
							Do_Key_Event_Down(offset);
						}
					}
				} else if (event.data == kISpButtonUp) {
					if (kb) {
						if (kb[0] == '+') {
							Com_sprintf (cmd, sizeof(cmd), "-%s %i %i\n", kb+1, event.refCon+1, time);
							Cbuf_AddText (cmd);
						}
					} else {
						// Check if aux range and we fake a keypress
						if ( (event.refCon >= kNeed_Aux25) && (event.refCon <= kNeed_Aux32) ) {							
							offset = event.refCon - kNeed_Aux25;
							
							/* Actually adding from Aux24 since I goofed and forgot I
							   reassigned Aux32 to the mac 'option' key */
							offset += K_AUX24;
							Do_Key_Event_Up(offset);
						}
					}
				}
				
			}
		}
	} while (wasEvent && !error);
}

//		Input_DisableMouseForCursor
void	Input_DisableMouseForCursor (void)
{
	// Note, activating inside ISp means unavailable for normal MacOS
	if (!gISPInitialized) {
		return;
	}
	if (gMouseAsMacOSCursorEnabled) {
		ISpDevices_ActivateClass (kISpDeviceClass_Mouse);
	}
	gMouseAsMacOSCursorEnabled = FALSE;
}

