//#include <MacHeaders.h>
#include <CodeFragments.h>
#include <Resources.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <burger.h>
#include "macfuncs.h"
#include "qcommon.h"

#define DUMPLIBSYMBOLS 0
#define	GAMELIBPATH 	1
#define MAX_OPENLIBS	8
#define MAX_SYMCACHE	256

// Our globals
static int						gGameLibIndex = 0;
static CFragConnectionID		gConIDArray[MAX_OPENLIBS];
static void						*gLibPtrArray[MAX_OPENLIBS];
static int						gLibLoadCount = 0;

static char sdumpname[255];

static void *gLoad_Ptr[MAX_SYMCACHE];
static char *gLoad_Sym[MAX_SYMCACHE];
static int	gLoad_Cache = -1;

static void *gSave_Ptr[MAX_SYMCACHE];
static char *gSave_Sym[MAX_SYMCACHE];
static int	gSave_Cache = -1;

// Our function prototypes
void *RealLoadGameLib(const char *dir, char *name, int count);

void FlushLoadCache(void);
void FlushSaveCache(void);
char *CheckLoadCache(void *addr);
void *CheckSaveCache(char *symb);
void AddLoadCache(void *addr, char *symb);
void AddSaveCache(void *addr, char *symb);

// -- Symbol cache routines to speed load/save times
void FlushLoadCache(void)
{
int count;

	if (gLoad_Cache <= 0)
		return;

	for (count = 0; count < gLoad_Cache; count++)
	{
		gLoad_Ptr[count] = 0;
		if (gLoad_Sym[count] != 0) {
			Free(gLoad_Sym[count]);
			gLoad_Sym[count] = 0;
		}
	}	
	gLoad_Cache = -1;
}
void FlushSaveCache(void)
{
int count;

	if (gSave_Cache <= 0)
		return;

	for (count = 0; count < gSave_Cache; count++)
	{
		gSave_Ptr[count] = 0;
		if (gSave_Sym[count] != 0) {
			Free(gSave_Sym[count]);
			gSave_Sym[count] = 0;
		}
	}	
	gSave_Cache = -1;
}
char *CheckLoadCache(void *addr)
{
int count;

	if (gLoad_Cache <= 0)
		return(0);

	for (count = 0; count < gLoad_Cache; count++)
	{
		if (addr == gLoad_Ptr[count])
			return(gLoad_Sym[count]);
	}
	return(0);
}
void *CheckSaveCache(char *symb)
{
int count;

	if (gSave_Cache <= 0)
		return(0);

	for (count = 0; count < gSave_Cache; count++)
	{
		if (!strcmp(symb,gSave_Sym[count]))
			return(gSave_Ptr[count]);
	}
	return(0);
}
void AddLoadCache(void *addr, char *symb)
{
	if (gLoad_Cache == -1)
		gLoad_Cache = 0;

	if (gLoad_Cache >= MAX_SYMCACHE)
		return;
		
	gLoad_Ptr[gLoad_Cache] = addr;
	gLoad_Sym[gLoad_Cache] = Burger::StringDuplicate(symb);
	++gLoad_Cache;
	
}
void AddSaveCache(void *addr, char *symb)
{
	
	if (gSave_Cache == -1)
		gSave_Cache = 0;
		
	if (gSave_Cache >= MAX_SYMCACHE)
		return;
		
	gSave_Ptr[gSave_Cache] = addr;
	gSave_Sym[gSave_Cache] = Burger::StringDuplicate(symb);
	++gSave_Cache;
	
}


// Take an address (that presumably is a pointer to a function or variable
// and find what symbol it is
char *Mac_AddressToSymbol(void *addr)
{
long symcount;
int dumpcount;
void *symAddr;
CFragSymbolClass symClass;
//char classname[255];
//char dbugstr[255];
Str255 psymname;
char *cachename;

	if (addr == NULL)
		return(0);
		
	// Flush the opposite cache when we get called
	FlushSaveCache();
	cachename = CheckLoadCache(addr);
	if (cachename != 0)
		return(cachename);

	CountSymbols(gConIDArray[gGameLibIndex], &symcount);
		
	for (dumpcount = 1; dumpcount <= symcount; dumpcount++)
	{
		GetIndSymbol(gConIDArray[gGameLibIndex], dumpcount, psymname, (char **)&symAddr, &symClass);
		Burger::PStringToCString(sdumpname,psymname);
		if (symAddr == addr)
		{
			// match!
//			sprintf(dbugstr, " %08x match %s\n",(int)addr, sdumpname);
//			Debug::String(dbugstr);
			AddLoadCache(addr,sdumpname);
			return(sdumpname);
		}
	}
	return(0);	
}


// Do the reverse lookup
void *Mac_SymbolToAddress(char *symb)
{
long symcount;
int dumpcount;
void *symAddr;
CFragSymbolClass symClass;
//char classname[255];
//char dbugstr[255];
Str255 psymname;
void *cacheaddr;

	if (symb == NULL)
		return(0);

	// Flush the opposite cache when we get called
	FlushLoadCache();
	cacheaddr = CheckSaveCache(symb);
	if (cacheaddr != 0)
		return(cacheaddr);
		
	CountSymbols(gConIDArray[gGameLibIndex], &symcount);
		
	for (dumpcount = 1; dumpcount <= symcount; dumpcount++)
	{
		GetIndSymbol(gConIDArray[gGameLibIndex], dumpcount, psymname, (char **)&symAddr, &symClass);
		Burger::PStringToCString(sdumpname,psymname);
		if (!strcmp(sdumpname,symb))
		{
			// match!
			//sprintf(dbugstr, "%08x Loadmatch %s\n",(int)symAddr,sdumpname);
			//DebugString(dbugstr);
			AddSaveCache(symAddr,symb);
			return(symAddr);
		}
	}
//	sprintf(dbugstr,"No match for %s !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n",symb);
//	Debug::String(dbugstr);
	return(0);	
}


// -----------------------------------------

/* 
	This is a special case routine for loading gameppc.lib
	since it exists in a subfolder that is determined by the
	value of the cvar "gamelib", so we have to use GetDiskFragment
	instead of GetSharedLibrary
*/

static char defaultDir[] = "base";

static void *LoadGameLib(char *name, int count)
{
	const char *rdir;
	void *resu;
	// first get the value of gamedir
	rdir = Cvar_VariableString("gamedir");
	/* if default, use baseq2 */
	if (*rdir == 0)
		rdir = defaultDir;
	resu = RealLoadGameLib(rdir,name,count);
	if (resu == 0)
	{
		// If the 'gamedir' isn't valid (like if it get reset by playing
		// a demo movie), force it to baseq2 (1.0.2)
		rdir = defaultDir;
		resu = RealLoadGameLib(rdir,name,count);
	}
	return(resu);
}
void *RealLoadGameLib(const char *dir, char *name, int count)
{
	FSSpec libSpec;
	Handle cfragH;
	char temppath[256];
	char *mangledpath;
	char Ppath[4096];
	OSStatus theErr;
	short oldResFile;
	short ourResFile;
	CFragResourcePtr cfrgptr;
	CFragResourceMemberPtr cmemptr;
	UInt32 loffset;
	UInt32 llength;
	Burger::Filename MyFilename;
	Str255 errname;


	/* setup path to BurgerLib expand on */
	sprintf(temppath,"9:%s:%s",dir,name);
	MyFilename.Expand(temppath);
	
	mangledpath = MyFilename.GetPtr();
	mangledpath[((Burger::StringLength(mangledpath)-1))] = 0;		/* strip trailing : */
	++mangledpath;								/* strip leading : */
	

	Burger::CStringToPString((Word8*)Ppath,mangledpath);

	/* Make the FSSpec from our full path */
	theErr = FSMakeFSSpec(0,0,(Word8 *)Ppath,&libSpec);
	if (theErr != noErr)
		return(0);
	
	/* Now use our FSSpec to first load the cfrg resource and parse it */
	
	/* save old current res file */
	oldResFile = CurResFile();
	
	/* open gameppc.lib's resource fork */
	ourResFile = FSpOpenResFile(&libSpec,fsRdPerm);
	if (ourResFile == -1)
	{
		DebugString("FSpOpenResFile failed on gameppc.lib\n");
		return(0);
	}

	/* load the cfrag resource */	
	cfragH = GetResource(kCFragResourceType,kCFragResourceID);
	cfrgptr = (CFragResourcePtr) *cfragH;
	cmemptr = (CFragResourceMemberPtr) &(cfrgptr->firstMember);

	/* get the offset and length of the first code fragment */
	loffset = cmemptr->offset;
	llength = cmemptr->length;
		
	ReleaseResource(cfragH);		/* release the cfrag resource */
	
	/* restore old cur res file */
	UseResFile(oldResFile);
	
	/* Now call GetDiskFragment with all our spiffy info */
	theErr = GetDiskFragment(&libSpec, loffset, llength, NULL, kReferenceCFrag,
							  &gConIDArray[count], (char **)&gLibPtrArray[count], errname);
	if (theErr != noErr) {
		DebugMessage("\n%s\n %d\n",errname,theErr);
		return(0);
	}

	++gLibLoadCount;

	return(gConIDArray[count]);
	
}

// -----------------------------------------

void *LoadMacLibrary(char *name)
{
int count;
char libname[255];
Str255 errname;
OSErr err;


#ifdef DUMPLIBSYMBOLS
// temp dump crap
long symcount;
int dumpcount;
void *symAddr;
CFragSymbolClass symClass;
char dumpname[255];
char classname[255];
#endif

// Just a paranoid safety thing here
	FlushLoadCache();
	FlushSaveCache();

	if (gLibLoadCount == 0)
	{
		//reset arrays
		for (count = 0; count < MAX_OPENLIBS; count++)
		{
			gConIDArray[count] = 0;
		}		
	}	
	count = 0;
	while (count < MAX_OPENLIBS)
	{
		if (gConIDArray[count] == 0)
		{
			/*
				This is a little hack to make sure we know which lib
				to search for Mac_AddressToSymbol and SymbolToAddress
			*/
			if (strstr(name,"gameppc.lib"))
			{
				gGameLibIndex = count;
#ifdef GAMELIBPATH
				/* Parse the path for loading gameppc.lib from "gamedir" cvar */
				return(LoadGameLib(name,count));
#endif			
			}		
			Burger::CStringToPString((Word8*)libname,name);
			err = GetSharedLibrary((Word8 *)libname, kPowerPCCFragArch, kLoadCFrag, &gConIDArray[count],
					  			   (char **)&gLibPtrArray[count], errname);
			if (err != noErr)
			{
				sprintf(libname,"LoacMacLibrary error %d \n",err);
				DebugString(libname);
				Burger::PStringToCString(libname,errname);
				DebugString(libname);
				Com_Printf(libname);
				Com_Printf("LoadMacLibrary error %d \n",err);
				return(0);
			}

#ifdef DUMPLIBSYMBOLS			
			CountSymbols(gConIDArray[count], &symcount);
			DebugMessage("%s has %d symbols \n",name,symcount);
			
			for (dumpcount = 1; dumpcount <= symcount; dumpcount++)
			{
				GetIndSymbol(gConIDArray[count], dumpcount, errname, (char **)&symAddr, &symClass);
				Burger::PStringToCString(dumpname,errname);
				switch (symClass) {
					case kCodeCFragSymbol:		sprintf(classname,"CODE");
												break;
					case kDataCFragSymbol:		sprintf(classname,"DATA");
												break;
					case kTVectorCFragSymbol:	sprintf(classname,"VECT");
												break;
					case kTOCCFragSymbol:		sprintf(classname,"TOCC");
												break;
					case kGlueCFragSymbol:		sprintf(classname,"GLUE");
												break;												
				}
				DebugMessage("SYMBOL %d, NAME %s, TYPE %s \n",dumpcount,dumpname,classname);
			}
#endif
			
			++gLibLoadCount;
			return(gConIDArray[count]);
		}
		++count;
	}
	return(0);	//failed
	
}
void FreeMacLibrary(void *ptrref)
{
int count;

	for (count = 0; count < MAX_OPENLIBS; count++)
	{
		if (gConIDArray[count] == (CFragConnectionID) ptrref)
		{
			CloseConnection(&gConIDArray[count]);
			gConIDArray[count] = 0;
			gLibPtrArray[count] = 0;
			--gLibLoadCount;
			return;
		}
	}
}
void *GetSymbol(void *ptrref, char *symname)
{
char psymname[256];
int count;
void *funcptr;
CFragSymbolClass theClass;
OSErr err;

	for (count = 0; count < MAX_OPENLIBS; count++)
	{
		if (gConIDArray[count] == (CFragConnectionID)ptrref)
		{
			Burger::CStringToPString((Word8*)psymname,symname);
			err = FindSymbol(gConIDArray[count], (Word8 *)psymname, (char **)&funcptr, &theClass);
			if (err != noErr)
			{
//				sprintf(psymname,"GetSymbol error %d\n",err);
//				DebugString(psymname);
				Com_Printf("GetSymbol err %d\n",err);
				
				return(0);
			}
			return(funcptr);
		}
	}
	return(0);
}
