#include <Quickdraw.h>
#include <Dialogs.h>
#include <ToolUtils.h>
#include <Devices.h>
#include "macfuncs.h"
#include "mac_isp.h"
#include "mac_version.h"

// Resource IDs

#define kAboutDLOG		1000
#define kFatalDLOG		1001
#define kCmdLineDLOG	1500

enum {
	kMBAR_Main			=	128,
	kMENU_Apple			=	128,
	kMENU_File,
	kMENU_Edit
};

enum {
	kMenuCMD_AboutBox	=	'abot'
};

enum {
	kMenuCMD_ResumeGame		=	'resm',
	kMenuCMD_ConfigureInput	=	'cnfg',
	kMenuCMD_Quit			=	'quit'
};

enum {
	kMenuCMD_Undo		=	'undo',
	kMenuCMD_Cut		=	'cut ',
	kMenuCMD_Copy		=	'copy',
	kMenuCMD_Paste		=	'past',
	kMenuCMD_Clear		=	'cler'
};

static Bool gMenuHandlerInited = FALSE;
static Handle theMenuBar;

static void DrawDefaultButton (DialogPtr theDialog)
{
	Rect		itemRect;
	Handle		itemHandle;
	short		itemType;
										// Get at the item's bounds.
	GetDialogItem(theDialog, 1, &itemType, &itemHandle, &itemRect);
	InsetRect(&itemRect, -4, -4);		// Inset (outset?) bounds by -4 pixels.
	PenSize(3, 3);						// Set the pen 3 pixels thick.
	FrameRoundRect(&itemRect, 16, 16);	// Draw the button outline.
	PenNormal();						// And restore pen to 1 pixel thick.
}

static void GetDialogString (DialogPtr theDialog, short item, StringPtr theString)
{
	Rect		itemRect;
	Handle		itemHandle;
	short		itemType;
										// Get handle to dialog item.
	GetDialogItem(theDialog, item, &itemType, &itemHandle, &itemRect);
	GetDialogItemText(itemHandle, theString);	// Extract text from item handle.
}

static void SetDialogString (DialogPtr theDialog, short item, StringPtr theString)
{
	Rect itemRect;
	Handle itemHandle;
	short itemType;
										// Get handle to dialog item.
	GetDialogItem(theDialog, item, &itemType, &itemHandle, &itemRect);
	SetDialogItemText(itemHandle, theString);	// Extract text from item handle.
}

void DoCommandLineDialog(void)
{
	DialogPtr	thedlog;
	int done;
	short item;
	char rawpstr[1024];
	char rawcstr[1024];

	thedlog = GetNewDialog(kCmdLineDLOG, 0, (WindowRef)-1);
	SetPortDialogPort(thedlog);
	ShowWindow(GetDialogWindow(thedlog));
	DrawDefaultButton(thedlog);
	FlushEvents(everyEvent, 0);
	
	done = FALSE;
	
	do {
		ModalDialog(0,&item);
		if (item == 1) {
			done = TRUE;
		}
	} while (!done);
	GetDialogString(thedlog, 2, (StringPtr)rawpstr);
	Burger::PStringToCString(rawcstr,(Word8*)rawpstr);
	BypassPlug = TRUE;			/* Don't show the intro */
	Mac_SetCommandLine(rawcstr);
	DisposeDialog(thedlog);
}

void DoAboutDialog(void)
{
	DialogPtr	thedlog;
	int done;
	short item;

	thedlog = GetNewDialog(kAboutDLOG, 0, (WindowRef)-1);
	SetPortDialogPort(thedlog);
	ShowWindow(GetDialogWindow(thedlog));
	DrawDefaultButton(thedlog);
	FlushEvents(everyEvent, 0);

	done = FALSE;
	do {
		ModalDialog(0,&item);
		if (item == 1) {
			done = TRUE;
		}
	} while (done == FALSE);
	DisposeDialog(thedlog);
}

void DoMacFatalDialog(const char *fatalstring)
{
	DialogPtr	thedlog;
	int done;
	short item;
	Word8 FooStr[256];

	Burger::CStringToPString(FooStr,fatalstring);
	InputSetState(FALSE);	/* Make sure InputSprocket is DEAD! */
	thedlog = GetNewDialog(kFatalDLOG, 0, (WindowRef)-1);
	SetPortDialogPort(thedlog);
	SetDialogString(thedlog,3,FooStr);
	ShowWindow(GetDialogWindow(thedlog));
	DrawDefaultButton(thedlog);
	FlushEvents(everyEvent, 0);

	InitCursor();
	ShowCursor();
#if defined(_DEBUG)
	DumpHandles();
#endif
	done = FALSE;
	do {
		ModalDialog(0,&item);
		if (item == 1) {
			done = TRUE;
		}
	} while (done == FALSE);
	DisposeDialog(thedlog);
}

//		MenuInit

void Mac_MenuBarInit(void)
{
MenuHandle myMH;
OSErr err;
	
	//	Load the menu bar
	theMenuBar = GetNewMBar(kMBAR_Main);
	if (theMenuBar == nil)
		ExitToShell();

	//	Install the menu bar
	SetMenuBar(theMenuBar);

	//	Add Apple Menu items
	AppendResMenu(GetMenuHandle(kMENU_Apple), 'DRVR');
	
	DisposeHandle(theMenuBar);
	
	err = MemError();
	if (err != noErr)
		DebugStr("\pAck, not noerr");
	
	myMH = GetMenuHandle(kMENU_File);
	if (myMH)
	{
		DisableMenuItem(myMH, 1);
		DisableMenuItem(myMH, 2);
		DisableMenuItem(myMH, 4);
	}

	DrawMenuBar();

	gMenuHandlerInited = TRUE;
}

//		MenuPrepForSelect

void Mac_MenuPrepForSelect(void)
{
	DrawMenuBar();
}

//		MenuDoChoice

int Mac_MenuBarDispatch(int inMenu, const short inModifiers)
{
SInt16		theMenu;
SInt16		theItem;
//UInt32		menuCommand;
MenuHandle	theMenuHandle;
Bool		handled = TRUE;

	if (FALSE == gMenuHandlerInited)
		return (FALSE);
		
	theMenu = HiWord(inMenu);
	theItem = LoWord(inMenu);
	theMenuHandle = GetMenuHandle(theMenu);

	//	Small hackish thing because the menu command stuff doesn't
	//	bother mentioning that you got a desk accessory.
	if ((kMENU_Apple == theMenu) && (theItem > 2))
	{
	Str255	daName;
	GrafPtr	oldPort;
	
		GetPort(&oldPort);
		GetMenuItemText(theMenuHandle, theItem, daName);
//		OpenDeskAcc(daName);
		SetPort(oldPort);
	}
	else
	{
		if ( (theMenu == kMENU_Apple) && (theItem == 1) )
		{
			DoAboutDialog();
		}
		if (theMenu == kMENU_File)
		{
			switch(theItem)
			{
				case 1:	gResumeSelected = TRUE;
						break;
				case 4:	gQuitSelected = TRUE;
						break;
			}
			
		}
	}

	HiliteMenu(0);
	DrawMenuBar();

	return (handled);
}
