#include "Burger.h"
#include "qcommon.h"
#include "rw_mac.h"
#include "errno.h"
#include "mac_isp.h"
#include "client.h"

void VID_Activate(qboolean isActive);

byte *membase;
int hunkmaxsize;
int cursize;

void InitBackground(void)
{
	IN_Activate(FALSE);
	VID_Activate(FALSE);
	Input_Suspend();	/* inputsprocket */
//	S_Activate(FALSE);
}


void KillBackground(void)
{
	IN_Activate(TRUE);
	VID_Activate(TRUE);
	Input_Resume();	/* inputsprocket */
}


void Mac_Resume_f(void)
{
	if (gMacBackground == FALSE)
		return;
		
	CL_Pause_f();			// unpause if necessary
	gMacBackground = FALSE;
	
	Com_Printf("Resuming from Mac desktop\n");
}

void Mac_Desktop_f(void)
{
cvar_t	*isPaused;

	// Already in background?
	if (gMacBackground == TRUE)
		return;

/*
	First, check if game is paused.  If it isn't, try to pause
	the game.  Since we can't pause in multiplayer, and the
	background issues would be nasty on the Mac, sorry, but no
	switching to the desktop when you're connected to a server
*/

	isPaused = Cvar_Get ("paused", "0", 0);
	
	if (isPaused == 0)
	{
		// never pause in multiplayer
		if (Cvar_VariableValue ("maxclients") > 1 || !Com_ServerState ())
		{
			Com_Printf("Can't switch to desktop in multiplayer");
			Cvar_SetValue ("paused", !isPaused->value);
			return;
		}

		Cvar_SetValue ("paused", !isPaused->value);
	}
	Com_Printf("Switching to Mac desktop\n");
	
	gMacBackground = TRUE;
}

// NCT_HACK function to convert path formats so the mac is happy

int	fprintfx(FILE * file, const char * format, ...)
{
	va_list		argptr;
	char		msg[2048];
	int			startoffset,endoffset,leng;
	
	va_start (argptr,format);
	vsprintf (msg,format,argptr);
	va_end (argptr);

	leng = Burger::StringLength(msg);
	startoffset = 0;
	while (msg[startoffset] != 0)
	{
		if (msg[startoffset] == 0xA)	//CR (converting Mac to 0x0D0A
		{
			for (endoffset = leng; endoffset > startoffset; endoffset--)
			{
				msg[endoffset+1] = msg[endoffset];
			}
			msg[startoffset] = 0xD;
			++startoffset;
			msg[startoffset] = 0xA;
		}
		++startoffset;
	}
	return(fwrite(msg,1,Burger::StringLength(msg),file));
}

#include <MacMemory.h>

void Hunk_Begin(int maxsize)
{
	cursize = 0;
	hunkmaxsize = (maxsize + 31) & ~31;	//	round to cacheline as below does

	// NCT_HACK to check for out of memory condition
	BombFlag = FALSE;  // BurgerLib

	DebugMessage("Hunk_Begin(%d)\n",hunkmaxsize);
	
	membase = (byte *)NewPtrClear(hunkmaxsize);
	if (membase == NULL) {
		Sys_Error ("Not enough free memory.  If you experience this during extended gameplay, consider increasing Sin's preferred RAM size.  You can do this in the Finder by selecting Get Info on the Sin application icon.");
	}	
#if _DEBUG
	DebugAddSourceLine((void **)membase,__FILE__,__LINE__);
#endif
	BombFlag = TRUE;
	
	if (!membase) {
		DebugString("VirtualAlloc reserve failed\n");
		Sys_Error("VirtualAlloc reserve failed");
	}
}

void *Hunk_Alloc (int size)
{
	DebugMessage("Hunk_Alloc(%d)\n",size);
	size = (size+31)&~31;	//round to cacheline
	cursize += size;
	if (cursize > hunkmaxsize) {
		DebugString("Hunk_Alloc overflow!!!!!!!!!!!!!!\n");
		Sys_Error("Hunk_Alloc overflow");
	}
	return(void *)(membase+cursize-size);
}

int Hunk_End(void **DestPtr)
{
	DebugMessage("Hunk_End() = %d\n",cursize);
	if (cursize != hunkmaxsize) {
		SetPtrSize((char *)membase,cursize);
	}
	DestPtr[0] = membase;
	return(cursize);
}

void Hunk_Delete(void *DestPtr)
{
	Free(DestPtr);
}