/*
	Modeless Dialog Utilities
	
*/

//#include <MacHeaders.h>
#include <Events.h>
#include <Windows.h>
#include <Dialogs.h>
#include <string.h>
//#include "burger.h"
#include "ModelessDLOG.h"

// Externs (for console text)
extern char MyConsoleInput[256];
extern int MyConsoleIOffset;

// Function prototypes
int MDU_InFront(void);
void MDU_UpdateControls(MDTemplate *dialogTemp);
void FindNextET(MDTemplate *dialogTemp, int step);
Bool IsDialogETNumeral(MDTemplate *dialogTemp);

// Globals...
static MDTemplate *modelessDialogs[MAX_MDU_DLOGS];
static Word32 numDialogs;

// Init the Modeless Dialog Box Utilities
void MDU_Init(void)
{
int count;

	numDialogs = 0;
	
	for (count = 0; count < MAX_MDU_DLOGS; count++)
		modelessDialogs[count] = NULL;
}



// Setup a modeless dialog box from a template
int MDU_RegisterDialog(MDTemplate *dialogTemp)
{
int dlogRef;

	if (numDialogs >= MAX_MDU_DLOGS)
		return(-1);
	
	++numDialogs;
	
	for (dlogRef = 0; dlogRef < MAX_MDU_DLOGS; dlogRef++)
	{
		if (modelessDialogs[dlogRef] == NULL)
		{
			modelessDialogs[dlogRef] = dialogTemp;
			SetWRefCon(GetDialogWindow(dialogTemp->theDialog), kModelessID);

			return(dlogRef);
		}
	}
	return(-1);
}



// Get rid of a modeless dialog box
void MDU_ReleaseDialog(int dialogRef)
{
	modelessDialogs[dialogRef] = NULL;
	--numDialogs;
}
	
	
	
	
// Show the bloody modeless dialog	
void MDU_ShowDialog(int dialogRef)
{
	MDU_UpdateControls(modelessDialogs[dialogRef]);
	ShowWindow(GetDialogWindow(modelessDialogs[dialogRef]->theDialog));
	SelectWindow(GetDialogWindow(modelessDialogs[dialogRef]->theDialog));
}



// Hide the blasted modeless dialog
void MDU_HideDialog(int dialogRef)
{
	HideWindow(GetDialogWindow(modelessDialogs[dialogRef]->theDialog));
}



// Set a modeless dialog control template up
void MDU_SetCtlTemplate(MDCtlTemplate *ctl, Word32 ctlID,
						Word32 ctlType, Word32 ctlFlags,
						Word32 ctlFamily, Word32 *ctlValue,
						void (*clickCallback)(EventRecord *clickEvent, short theItem),
						void (*itemDraw)())
{
	ctl->itemID = ctlID;
	ctl->itemType = ctlType;
	ctl->itemFlags = ctlFlags;
	ctl->itemFamily = ctlFamily;
	ctl->itemValue = ctlValue;
	ctl->clickCallback = clickCallback;
	ctl->itemDraw = itemDraw;
}




// Is one of our modeless dialogs in front, if so, which?	
int MDU_InFront(void)
{
WindowPtr front;
long ref;	
int count;
	
	front = FrontWindow();	
	ref = GetWRefCon(front);
	if (ref == kModelessID)
	{
		for (count = 0; count < MAX_MDU_DLOGS; count++)
		{
			if(GetDialogWindow(modelessDialogs[count]->theDialog) == front)
			return(count);
		}
	}
	
	return(-1);
}	
	
	
	
	
// -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// Key event handler for our modeless dialog boxes
Bool MDU_HandleKey(EventRecord *theEvent)
{
int ourDialog;
short whatItem;

MDCtlTemplate *myCtl;
Handle myHn;
Rect myRec;
short mytyp;
Str255 theText;
MDTemplate *dialogTemp;
int count;

	ourDialog = 0;
	if (!IsDialogEvent(theEvent))
		return(FALSE);
		
	// Need to check if any EditText controls
	// if so, check if tab key
	// if tab key, let DialogSelect munch it and figure out which EditText will get hit next
	// if not tab key, figure out if EditText can munch key
	// if EditText can, let DialogSelect munch it

	if ( (modelessDialogs[ourDialog]->areEditText) == TRUE)	
	{
		// tab = ASCII 09
		if ( ((theEvent->message)&charCodeMask) == 9)
		{
				FindNextET(modelessDialogs[ourDialog], 1);
		}
		else
		{ // not tab, check if we need to filter keypress....

// NCT_HACK make 'return' keypress send the command text to the console
			if ( (((theEvent->message)&charCodeMask) == 0xd)	|| (((theEvent->message)&charCodeMask) == 0xa) )
			{
				dialogTemp = modelessDialogs[0];

				myCtl = dialogTemp->firstCtl;
				for (count = 0; count < dialogTemp->numItems; count++)
				{
					if ( (myCtl->itemType) == kEditTextDialogItem)
					{
						GetDialogItem(dialogTemp->theDialog,myCtl->itemID,&mytyp,&myHn,&myRec);
						GetDialogItemText(myHn,theText);
						Burger::PStringToCString(MyConsoleInput,theText);
						MyConsoleIOffset = Burger::StringLength(MyConsoleInput);
						SetDialogItemText(myHn,NULL);
						SelectDialogItemText(dialogTemp->theDialog, myCtl->itemID, 0, 32767);
						dialogTemp->selectedEditText = myCtl->itemID; // Move inside dialogTemp!
						count = dialogTemp->numItems;
					}
					++myCtl;
				}
			}	
		
		}
	}
	DialogSelect(theEvent, &(modelessDialogs[ourDialog]->theDialog), &whatItem);
	return(TRUE);
}




// -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// Handle an idle event in our modeless dialog
Bool MDU_HandleIdle(EventRecord *theEvent)
{
int ourDialog;
short whatItem;

	ourDialog = 0;
	if (!IsDialogEvent(theEvent))
		return(FALSE);

	DialogSelect(theEvent, &(modelessDialogs[ourDialog]->theDialog), &whatItem);

	return(TRUE);
}



// Handle a click in our modeless dialog
Bool MDU_HandleClick(EventRecord *theEvent)
{
int ourDialog;
int count;
int itemCount;
MDCtlTemplate *currentControl, *matchControl;
short whatItem;
Handle itemHandle;
Rect itemRect;
short itemTyp;

	ourDialog = 0;
	if (!IsDialogEvent(theEvent))
		return(FALSE);

	if (DialogSelect(theEvent, &(modelessDialogs[ourDialog]->theDialog), &whatItem) == TRUE)
	{
		// Try and find match in item templates and see if this control that got clicked
		// has its own handling proceedure...
		currentControl = modelessDialogs[ourDialog]->firstCtl;
		for (itemCount = 0; itemCount < modelessDialogs[ourDialog]->numItems; itemCount++)
		{
			if (currentControl->itemID == whatItem)
			{
				if ( ((currentControl->itemFlags) & kMDItem_Override) != 0)
					if (currentControl->clickCallback != NULL)
					{
						currentControl->clickCallback(theEvent, whatItem);
						return(TRUE);
					}
				// Handle the click ourselves...radio/push/check control probably....
				
				if (currentControl->itemType == kCheckBoxDialogItem)
				{
					if (currentControl->itemValue != NULL)
					{
						if ( ( *(currentControl->itemValue) == 0))
							*(currentControl->itemValue) = 1;
						else
							*(currentControl->itemValue) = 0;
						
						GetDialogItem(modelessDialogs[ourDialog]->theDialog, whatItem, &itemTyp, &itemHandle, &itemRect);
						SetControlValue((ControlHandle)itemHandle, (short)(*(currentControl->itemValue)));
					}
				}
				
				if (currentControl->itemType == kRadioButtonDialogItem)
				{
					matchControl = modelessDialogs[ourDialog]->firstCtl;
					for (count = 0; count < modelessDialogs[ourDialog]->numItems; count++)
					{
						if ( matchControl->itemID != whatItem)
							if (matchControl->itemType == kRadioButtonDialogItem)
								if (matchControl->itemFamily == currentControl->itemFamily)
								{
									GetDialogItem(modelessDialogs[ourDialog]->theDialog, matchControl->itemID, &itemTyp, &itemHandle, &itemRect);
									SetControlValue((ControlHandle)itemHandle, 0);
									if (matchControl->itemValue != NULL)
										*(matchControl->itemValue) = 0;								
								}
						++matchControl;
					}
					GetDialogItem(modelessDialogs[ourDialog]->theDialog, whatItem, &itemTyp, &itemHandle, &itemRect);
					SetControlValue((ControlHandle)itemHandle, 1);
					if (currentControl->itemValue != NULL)
						*(currentControl->itemValue) = 1;													
				}			
				return(TRUE);					
			}
			++currentControl;
		}
	
	}

	return(FALSE);
}




// Is this event possible for our modeless dialog?
Bool MDU_IsDialog(EventRecord *theEvent)
{
	if (MDU_InFront() < 0)
		return(FALSE);
	else
		return(TRUE);

	theEvent=theEvent;	
	return(FALSE);
}



// Handle an update event for our modeless dialog
Bool MDU_HandleUpdate(EventRecord *theEvent)
{
int ourDialog;
//GrafPtr oldGrafPort;
//short whatItem;

	if (!IsDialogEvent(theEvent))
		return(FALSE);

	ourDialog = MDU_InFront();

	if (ourDialog < 0)
		return(FALSE);

	return(TRUE);
}



// Handle a close event for our modeless dialog
Bool MDU_HandleClose(EventRecord *theEvent)
{
int ourDialog;

	ourDialog = MDU_InFront();

	if (ourDialog < 0)
		return(FALSE);

	theEvent=theEvent;	

	MDU_HideDialog(ourDialog);

	return(TRUE);
}



// Handle a drag event for our modeless dialog
Bool MDU_HandleDrag(EventRecord *theEvent)
{
	int ourDialog;
	Rect Bounds;

	ourDialog = 0;

#if !TARGET_API_MAC_CARBON
	Bounds = GetGrayRgn()[0]->rgnBBox;
#else
	GetRegionBounds(GetGrayRgn(),&Bounds);
#endif
	DragWindow(GetDialogWindow(modelessDialogs[ourDialog]->theDialog), theEvent->where, &Bounds);

	return(TRUE);
}



// Handle updating the controls for our modeless dialog	
void MDU_UpdateControls(MDTemplate *dialogTemp)
{	
int count;
MDCtlTemplate *myCtl;
Word32 theVal;
Handle itemHandle;
Rect itemRect;
short itemType;

			// Set the control values for controls
			// in our jurisdiction...
			
	myCtl = dialogTemp->firstCtl;			
	for (count = 0; count < dialogTemp->numItems; count++)
	{
		if ( (myCtl->itemType) != kEditTextDialogItem)
		{
			if ( (myCtl->itemValue) != NULL)
			{
				theVal = (*(myCtl->itemValue));
				GetDialogItem(dialogTemp->theDialog, myCtl->itemID, &itemType, &itemHandle, &itemRect);
				SetControlValue((ControlHandle)itemHandle, theVal);
			}
		}
		++myCtl;
	}
	
	if ( (dialogTemp->areEditText) == TRUE)	
	{
		myCtl = dialogTemp->firstCtl;
		for (count = 0; count < dialogTemp->numItems; count++)
		{
			if ( (myCtl->itemType) == kEditTextDialogItem)
			{
				SelectDialogItemText(dialogTemp->theDialog, myCtl->itemID, 0, 32767);
				dialogTemp->selectedEditText = myCtl->itemID; // Move inside dialogTemp!
				count = dialogTemp->numItems;
			}
			++myCtl;
		}
	}
}




// Step through the EditText controls in a modeless dialog
void FindNextET(MDTemplate *dialogTemp, int step)
{
int count, offset;
MDCtlTemplate *myCtl, *baseCtl;

	if (dialogTemp->areEditText == FALSE)
		return;
		
	myCtl = dialogTemp->firstCtl;
	for (count = 0; count < dialogTemp->numItems; count++)
	{
		if ( (myCtl->itemType) == kEditTextDialogItem)
			if ( (myCtl->itemID) == (dialogTemp->selectedEditText) )
			{
				offset = count;
				baseCtl = myCtl;
				
				count += step;
				myCtl += step;
				
				if (count < 0)
				{
					count = (dialogTemp->numItems)-1;
					myCtl = (dialogTemp->firstCtl)+count;
				}
				if (count > dialogTemp->numItems)
				{
					count = 0;
					myCtl = dialogTemp->firstCtl;
				}
				
				while (count != offset)
				{				// ++
					if (myCtl->itemType == kEditTextDialogItem)
					{
						dialogTemp->selectedEditText = myCtl->itemID;
						return;
					}

				count += step;
				myCtl += step;
				
					if (count < 0)
					{
						count = (dialogTemp->numItems)-1;
						myCtl = (dialogTemp->firstCtl)+count;
					}
					if (count > dialogTemp->numItems)
					{
						count = 0;
						myCtl = dialogTemp->firstCtl;
					}
				}				// ++
			}
		++myCtl;
	}
}




// Is a EditText control supposed to be numbers only?
Bool IsDialogETNumeral(MDTemplate *dialogTemp)
{
int count;
MDCtlTemplate *myCtl;

	if (dialogTemp->areEditText == FALSE)
		return(FALSE);
		
	myCtl = dialogTemp->firstCtl;
	for (count = 0; count < dialogTemp->numItems; count++)
	{
		if ( (myCtl->itemType) == kEditTextDialogItem)
			if ( (myCtl->itemID) == (dialogTemp->selectedEditText) )
				if ( (myCtl->itemFlags & kMDItem_EditTextNumber) != 0)
					return(TRUE);
				else
					return(FALSE);
		++myCtl;
	}
	return(FALSE);
}

