#include "Burger.h"
#include "r_local.h"
#include "mac_drawsp.h"

int gIsOpenGL = FALSE;

static int gRenderActive = TRUE;

qboolean SWimp_InitGraphics(qboolean fullscreen);

void SWimp_BeginFrame( float camera_separation )
{
	LockVideoPage();
	vid.buffer = VideoPointer;
	
}

void SWimp_EndFrame (void)
{
	if (gRenderActive == FALSE)
		return;
	macds_SwapBuffer();
	LockVideoPage();
	vid.buffer = VideoPointer;
}

int SWimp_Init( void *hInstance, void *wndProc )
{
	return(TRUE);
}

void SWimp_SetPalette( const unsigned char *palette)
{
	unsigned char tpal[256*3];
	int count;

	// uses padded xRGB 8 bit per color format
	for (count = 0; count < 256; count++) {
		tpal[(count*3)+0] = palette[(count*4)+0];
		tpal[(count*3)+1] = palette[(count*4)+1];
		tpal[(count*3)+2] = palette[(count*4)+2];
	}
}


void SWimp_Shutdown( void )
{
	macds_Shutdown();
}

rserr_t SWimp_SetMode( int *pwidth, int *pheight, int mode, qboolean fullscreen )
{
int width,height;

	ri.Con_Printf(PRINT_ALL, "Initializing software display\n");
	ri.Con_Printf(PRINT_ALL, "...setting mode %d:",mode);
	if ( !ri.Vid_GetModeInfo(&width,&height,mode) )
	{
		ri.Con_Printf(PRINT_ALL, " invalid mode\n");
		return(rserr_invalid_mode);
	}
	
	ri.Con_Printf(PRINT_ALL, " %d %d\n", width, height);
	
	// release existing video stuph
	SWimp_Shutdown();

	*pwidth = width;
	*pheight = height;

	if (!SWimp_InitGraphics(fullscreen))
	{
		return(rserr_invalid_mode);
	}
	

	// let sound and input subsystems know about the new window
	ri.Vid_NewWindow(width,height);
	
	return(rserr_ok);
}

void SWimp_AppActivate( qboolean active )
{
	if (active == TRUE)
	{
		gRenderActive = TRUE;
		macds_ActivateContext();
	} else {
		gRenderActive = FALSE;
		macds_PauseContext();
	}
}

qboolean SWimp_InitGraphics(qboolean fullscreen)
{
	if ( r_colorlighting->value > 0 ) {
		vid.colorlighting = TRUE;
	} else {
		vid.colorlighting = FALSE;
	}
	if (macds_InitContext(vid.width,vid.height,fullscreen) == FALSE) {
		return(FALSE);
	}

	macds_ActivateContext();

	LockVideoPage();
	vid.rowbytes = VideoWidth;
	vid.buffer = VideoPointer;
	vid.bitdepth = macds_GetDepth();

	if ( vid.bitdepth > 8 ) {
		vid.colorlighting = TRUE;
		vid.rowbytes /= 2;
	}
	ri.Vid_NewWindow(vid.width,vid.height);
	return(TRUE);
}

