//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/cl_bspline.cpp                                         $
// $Revision:: 3                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 2/5/02 7:36 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// Charlie Brown's uniform non-rational bspline class.
// 

#include "cl_bspline.h"
#include "client.h"

#ifndef REF_HARD_LINKED
void BSpline::Set(Vector *control_points_,int num_control_points_,splinetype_t type)
{
	int i;

	SetType( type );
	if ( control_points ) {
		delete [] control_points;
		control_points = NULL;
	}
	num_control_points = num_control_points_;
	if ( num_control_points ) {
		control_points = new Vector[ num_control_points ];
		assert( control_points );

		for( i = 0; i < num_control_points; i++ ) {
			control_points[ i ] = control_points_[ i ];
		}
	}
}

void BSpline::Clear(void)
{
	if( control_points ) {
		delete [] control_points;
		control_points = NULL;
	}
	num_control_points = 0;
}

inline void BSpline::EvalNormal(Vector *Output,float u)
{
	int		segment_id;
	float		B[ 4 ];
	float		tmp;
	float		u_2;
	float		u_3;

	segment_id = ( int )u;
	if ( segment_id < 0 ) {
		segment_id = 0;
	}
	if ( segment_id > num_control_points - 4) {
		segment_id = num_control_points - 4;
	}
	u -= ( float )segment_id;

	u_2 = u * u;
	u_3 = u * u_2;

	tmp = 1 - u;
	B[ 0 ] = ( tmp * tmp * tmp ) * ( 1.0f / 6.0f );
	B[ 1 ] = ( 3.0f * u_3 - 6.0f * u_2 + 4.0f ) * ( 1.0f / 6.0f );
	B[ 2 ] = ( -3.0f * u_3 + 3.0f * u_2 + 3.0f * u + 1 ) * ( 1.0f / 6.0f );
	B[ 3 ] = u_3 * ( 1.0f / 6.0f );

	Output->x = 
		control_points[ 0 + segment_id ].x * B[ 0 ] +
		control_points[ 1 + segment_id ].x * B[ 1 ] +
		control_points[ 2 + segment_id ].x * B[ 2 ] +
		control_points[ 3 + segment_id ].x * B[ 3 ];
	Output->y = 
		control_points[ 0 + segment_id ].y * B[ 0 ] +
		control_points[ 1 + segment_id ].y * B[ 1 ] +
		control_points[ 2 + segment_id ].y * B[ 2 ] +
		control_points[ 3 + segment_id ].y * B[ 3 ];
	Output->z = 
		control_points[ 0 + segment_id ].z * B[ 0 ] +
		control_points[ 1 + segment_id ].z * B[ 1 ] +
		control_points[ 2 + segment_id ].z * B[ 2 ] +
		control_points[ 3 + segment_id ].z * B[ 3 ];
}

inline void BSpline::EvalLoop(Vector *Output,float t)
{
	int segment_id;
	float B[ 4 ];
	float tmp;
	float u;
	float u_2;
	float u_3;
	int i;
	int j;

	segment_id = ( int )floor( t );
	u = t - floor( t );

	segment_id %= num_control_points;
	if ( segment_id < 0 ) {
		segment_id += num_control_points;
	}

	u_2 = u * u;
	u_3 = u * u_2;

	tmp = 1 - u;
	B[ 0 ] = ( tmp * tmp * tmp ) * ( 1.0f / 6.0f );
	B[ 1 ] = ( 3.0f * u_3 - 6.0f * u_2 + 4.0f ) * ( 1.0f / 6.0f );
	B[ 2 ] = ( -3.0f * u_3 + 3.0f * u_2 + 3.0f * u + 1 ) * ( 1.0f / 6.0f );
	B[ 3 ] = u_3 * ( 1.0f / 6.0f );

	Output->x = 0;
	Output->y = 0;
	Output->z = 0;
	for( i = 0, j = segment_id; i < 4; i++, j++ ) {
		if ( j >= num_control_points ) {
			j -= num_control_points;
		}
		Output->x = Output->x+(control_points[j].x * B[i]);
		Output->y = Output->y+(control_points[j].y * B[i]);
		Output->z = Output->z+(control_points[j].z * B[i]);
	}
}

inline void BSpline::EvalClamp(Vector *Output,float t)
{
	int		segment_id;
	float		B[ 4 ];
	float		tmp;
	float		u;
	float		u_2;
	float		u_3;
	int		i;
	int		j;

	segment_id = ( int )floor( t );
	u = t - floor( t );

	u_2 = u * u;
	u_3 = u * u_2;

	tmp = 1 - u;
	B[ 0 ] = ( tmp * tmp * tmp ) * ( 1.0f / 6.0f );
	B[ 1 ] = ( 3.0f * u_3 - 6.0f * u_2 + 4.0f ) * ( 1.0f / 6.0f );
	B[ 2 ] = ( -3.0f * u_3 + 3.0f * u_2 + 3.0f * u + 1 ) * ( 1.0f / 6.0f );
	B[ 3 ] = u_3 * ( 1.0f / 6.0f );

	for( i = 0; i < 4; i++, segment_id++ ) {
		j = segment_id;
		if ( j < 0 ) {
			j = 0;
		} else if ( j >= num_control_points ) {
			j = num_control_points - 1;
		}
		Output->x = Output->x+(control_points[j].x * B[i]);
		Output->y = Output->y+(control_points[j].y * B[i]);
		Output->z = Output->z+(control_points[j].z * B[i]);
	}
}

Vector BSpline::Eval(float u)
{
	Vector Result;
	switch( curvetype ) {
	default:
	case SPLINE_NORMAL:
		EvalNormal(&Result,u);
		break;
	case SPLINE_CLAMP:
		EvalClamp(&Result,u);
		break;
	case SPLINE_LOOP:
		EvalLoop(&Result,u);
		break;
	}
	return Result;
}

void BSpline::AppendControlPoint
	(
	const Vector& new_control_point
	)

	{
	Vector *old_control_points;
	int i;

	old_control_points = control_points;
	num_control_points++;

	control_points = new Vector[num_control_points];
	assert( control_points );

	if ( old_control_points )
		{
		for( i = 0; i < num_control_points - 1; i++ )
			{
			control_points[ i ] = old_control_points[ i ];
			}
		delete [] old_control_points;
		}

	control_points[ num_control_points - 1 ] = new_control_point;
	}
#endif

#define MAX_BSPLINES 64
BSpline *bsplineList[MAX_BSPLINES];

void InitializeBSplines( void )
   {
   int i;
   for (i=0;i<MAX_BSPLINES;i++)
      {
      bsplineList[i] = 0;
      }
   }

qboolean CreateBSpline(int num, int numpoints, float *points)
{
	Vector *veclist;
	int i,j;
	
	if (num > MAX_BSPLINES)
		return FALSE;
	
	veclist = new Vector[numpoints];
	
	for (i=0,j=0;i<numpoints;i++,j+=2) {
		veclist[i].setXYZ(points[j], points[j+1], 0);
	}
	
	bsplineList[num] = new BSpline(veclist, numpoints, SPLINE_LOOP);
	delete [] veclist;
	return TRUE;
}

qboolean BSplineExists(int num)
{
	return (bsplineList[num] != 0);
}

void GetBSplinePosition(int num, float u, float *x, float *y)
{
	Vector vec;
	vec = bsplineList[num]->Eval(u);
	*x = vec.x;
	*y = vec.y;
}
