//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/gl_rsurf.c                                             $
// $Revision:: 5                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 8/8/01 4:46 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// surface-related refresh code
//

#include <assert.h>
#include <alloca.h>
#include <ctype.h>
#include "gl_local.h"

#ifndef GL_UNPACK_CLIENT_STORAGE_APPLE
#define GL_UNPACK_CLIENT_STORAGE_APPLE 0x85B2
#endif

#ifndef GL_BGRA
#define GL_BGRA 0x80E1
#endif

#if defined(BURGER_MAC) && TARGET_API_MAC_CARBON && 0
#define qglTexSubImage2D2(a,b,c,d,e,f,g,h,i) qglTexSubImage2D(a,b,c,d,e,f,GL_BGRA,GL_UNSIGNED_INT_8_8_8_8_REV,i)
#else
#define qglTexSubImage2D2(a,b,c,d,e,f,g,h,i) qglTexSubImage2D(a,b,c,d,e,f,g,h,i)

#endif


#ifdef SIN
int c_randval;
vec3_t zero={0,0,0};
#endif

#ifdef SIN
extern vec3_t modelorg;
#else
static vec3_t modelorg;		// relative to viewpoint
#endif

extern msurface_t *r_alpha_surfaces;

#define DYNAMIC_LIGHT_WIDTH  128
#define DYNAMIC_LIGHT_HEIGHT 128

#define LIGHTMAP_BYTES 4

#define	BLOCK_WIDTH		128
#define	BLOCK_HEIGHT	128

#define	MAX_LIGHTMAPS	128

int		c_visible_lightmaps;
int		c_visible_textures;

#define GL_LIGHTMAP_FORMAT GL_RGBA

typedef struct {
	int internal_format;
	int	current_lightmap_texture;

	msurface_t	*lightmap_surfaces[MAX_LIGHTMAPS];

	int			allocated[BLOCK_WIDTH];

	// the lightmap texture data needs to be kept in
	// main memory so texsubimage can update properly
	byte lightmap_buffer[4*BLOCK_WIDTH*BLOCK_HEIGHT];
} gllightmapstate_t;

static gllightmapstate_t gl_lms;


static void		LM_InitBlock( void );
static void		LM_UploadBlock( qboolean dynamic );
static qboolean	LM_AllocBlock (int w, int h, int *x, int *y);

extern void R_SetCacheState( msurface_t *surf );
extern void GL_R_BuildLightMap (msurface_t *surf, byte *dest, int stride);

#ifdef SIN
#define MAX_RENDERED_SPLATTERS 256
static  int num_splatter_surfaces;
static  msurface_t * splatter_surfaces[ MAX_RENDERED_SPLATTERS ];
#endif

/*
=============================================================

	BRUSH MODELS

=============================================================
*/

/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
static image_t *R_TextureAnimation (mtexinfo_t *tex)
    {
#ifdef SIN
    image_t * image;
    int c;

    image = tex->image;

    if ( tex->surf )
       {
       c = tex->surf->animation_frame;

       while ( c )
          {
          tex = tex->next;
          c--;
          }
       }
    else
       {
       assert( 0 );
       }

    if ( tex )
       image = tex->image;
    else
       assert( 0 );
    
    return image;
#else
    int c;
    c = currententity->frame % tex->numframes;

	while (c)
	{
		tex = tex->next;
		c--;
	}
	return tex->image;
#endif
}

#if 0
/*
=================
WaterWarpPolyVerts

Mangles the x and y coordinates in a copy of the poly
so that any drawing routine can be water warped
=================
*/
glpoly_t *WaterWarpPolyVerts (glpoly_t *p)
{
	int		i;
	float	*v, *nv;
	static byte	buffer[1024];
	glpoly_t *out;

	out = (glpoly_t *)buffer;

	out->numverts = p->numverts;
	v = p->verts[0];
	nv = out->verts[0];
	for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE, nv+=VERTEXSIZE)
	{
		nv[0] = v[0] + 4*sin(v[1]*0.05+r_newrefdef.time)*sin(v[2]*0.05+r_newrefdef.time);
		nv[1] = v[1] + 4*sin(v[0]*0.05+r_newrefdef.time)*sin(v[2]*0.05+r_newrefdef.time);

		nv[2] = v[2];
		nv[3] = v[3];
		nv[4] = v[4];
		nv[5] = v[5];
		nv[6] = v[6];
	}

	return out;
}

/*
================
DrawGLWaterPoly

Warp the vertex coordinates
================
*/
void DrawGLWaterPoly (glpoly_t *p)
{
	int		i;
	float	*v;

	p = WaterWarpPolyVerts (p);
	qglBegin (GL_TRIANGLE_FAN);
	v = p->verts[0];
	for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
	{
		qglTexCoord2f (v[3], v[4]);
		qglVertex3fv (v);
	}
	qglEnd ();
}
void DrawGLWaterPolyLightmap (glpoly_t *p)
{
	int		i;
	float	*v;

	p = WaterWarpPolyVerts (p);
	qglBegin (GL_TRIANGLE_FAN);
	v = p->verts[0];
	for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
	{
		qglTexCoord2f (v[5], v[6]);
		qglVertex3fv (v);
	}
	qglEnd ();
}
#endif

#ifdef SIN
float r_solidcolors[256*3];
void R_DrawTriangleOutlines (void);
#endif

/*
================
DrawGLPoly
================
*/

#ifndef AMD3D

#ifdef SIN
extern cvar_t *gl_drawflat;
static void DrawGLPoly (msurface_t *s)
#else
static void DrawGLPoly (glpoly_t *p)
#endif
   {
	int		i;
	float  	*v;
#ifdef SIN
   glpoly_t *p = s->polys;

   if (gl_drawflat->value)
     {
     int    index = (((int)s>>3)&0xff);
     float  *c = &r_solidcolors[index*3];
     //
     // solid poly
     //
     qglDisable (GL_TEXTURE_2D);

     qglBegin (GL_POLYGON);
     v = p->verts[0];
     for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
        {
        qglColor3f ( c[0],c[1],c[2] );
        qglVertex3fv (v);
        }
     qglEnd ();
     qglColor3f ( 1.0,1.0,1.0 );
     qglEnable (GL_TEXTURE_2D);
      }
   else if (s->texinfo->flags & SURF_ENVMAPPED)
      {
      vec3_t eye_normal;
      vec3_t eye_vert;
      vec3_t normal;
      vec3_t transform[3];
      float two_nu, m, fx, fy, fz;

      if ( s->flags & SURF_PLANEBACK )
         VectorNegate(s->plane->normal, normal);
      else
         VectorCopy(s->plane->normal, normal);

	   VectorCopy (vright, transform[0]);
	   VectorNegate (vup, transform[1]);
	   VectorCopy (vpn, transform[2]);

		eye_normal[0] = DotProduct(normal, transform[0]);
		eye_normal[1] = DotProduct(normal, transform[1]);
		eye_normal[2] = DotProduct(normal, transform[2]);

      qglBegin (GL_POLYGON);
      v = p->verts[0];
      for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
         {
		   eye_vert[0] = DotProduct(v, transform[0]) - r_origin[0];
		   eye_vert[1] = DotProduct(v, transform[1]) - r_origin[1];
		   eye_vert[2] = DotProduct(v, transform[2]) - r_origin[2];
         VectorNormalize( eye_vert );
         two_nu = 2.0f * DotProduct( eye_normal, eye_vert );
         fx = eye_vert[0] - eye_normal[0] * two_nu;
         fy = eye_vert[1] - eye_normal[1] * two_nu;
         fz = eye_vert[2] - eye_normal[2] * two_nu;
         m = 2.0F * Burger::FPMath::Sqrt( fx*fx + fy*fy + (fz+1.0F)*(fz+1.0F) );
         if (m==0.0F)
            {
      	   qglTexCoord2f ( 0.5f, 0.5f );
            }
         else
            {
            float mInv = 1.0F / m;
      	   qglTexCoord2f ( fx * mInv + 0.5f, fy * mInv + 0.5F );
            }
         qglVertex3fv (v);
         }
      qglEnd ();
      }
   else
      {
      qglBegin (GL_POLYGON);
      v = p->verts[0];
      for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
         {
         qglTexCoord2f (v[3] + s->texinfo->surf->delta_s, v[4] + s->texinfo->surf->delta_t);
         qglVertex3fv (v);
         }
      qglEnd ();
      }
#else
  	qglBegin (GL_POLYGON);
	v = p->verts[0];
	for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
	{
		qglTexCoord2f (v[3], v[4]);
		qglVertex3fv (v);
	}
	qglEnd ();
#endif
   }

#endif

/*
** R_DrawTriangleOutlines
*/
void R_DrawTriangleOutlines (void)
{
	int			i, j;
	glpoly_t	*p;

	if (!gl_showtris->value)
		return;

	qglDisable (GL_TEXTURE_2D);
	qglDisable (GL_DEPTH_TEST);
	qglColor4f (1,1,1,1);

	for (i=0 ; i<MAX_LIGHTMAPS ; i++)
	{
		msurface_t *surf;

		for ( surf = gl_lms.lightmap_surfaces[i]; surf != 0; surf = surf->lightmapchain )
		{
			p = surf->polys;
			for ( ; p ; p=p->chain)
			{
				for (j=2 ; j<p->numverts ; j++ )
				{
					qglBegin (GL_LINE_STRIP);
					qglVertex3fv (p->verts[0]);
					qglVertex3fv (p->verts[j-1]);
					qglVertex3fv (p->verts[j]);
					qglVertex3fv (p->verts[0]);
					qglEnd ();
				}
			}
		}
	}

	qglEnable (GL_DEPTH_TEST);
	qglEnable (GL_TEXTURE_2D);
}

/*
** DrawGLPolyChain
*/
#ifndef AMD3D
static void DrawGLPolyChain( glpoly_t *p, float soffset, float toffset )
{
	if ( soffset == 0 && toffset == 0 )
	{
		for ( ; p != 0; p = p->chain )
		{
			float *v;
			int j;

			qglBegin (GL_POLYGON);
			v = p->verts[0];
			for (j=0 ; j<p->numverts ; j++, v+= VERTEXSIZE)
			{
				qglTexCoord2f (v[5], v[6] );
				qglVertex3fv (v);
			}
			qglEnd ();
		}
	}
	else
	{
		for ( ; p != 0; p = p->chain )
		{
			float *v;
			int j;

			qglBegin (GL_POLYGON);
			v = p->verts[0];
			for (j=0 ; j<p->numverts ; j++, v+= VERTEXSIZE)
			{
				qglTexCoord2f (v[5] - soffset, v[6] - toffset );
				qglVertex3fv (v);
			}
			qglEnd ();
		}
	}
}
#endif

/*
** R_BlendLightMaps
**
** This routine takes all the given light mapped surfaces in the world and
** blends them into the framebuffer.
*/
static void R_BlendLightmaps (void)
{
	int			i;
	msurface_t	*surf, *newdrawsurf = 0;

	// don't bother if we're set to fullbright
	if (r_fullbright->value)
		return;
	if (!r_worldmodel->lightdata)
		return;

	// don't bother writing Z
	qglDepthMask( 0 );

#ifdef SIN
	qglDepthFunc (GL_EQUAL);
#endif

	/*
	** set the appropriate blending mode unless we're only looking at the
	** lightmaps.
	*/
	if (!gl_lightmap->value)
	{
		qglEnable (GL_BLEND);

		if ( gl_saturatelighting->value )
		{
			qglBlendFunc( GL_ONE, GL_ONE );
		}
		else
		{
			if ( gl_monolightmap->CVarString[0] != '0' )
			{
				switch ( toupper( gl_monolightmap->CVarString[0] ) )
				{
				case 'I':
					qglBlendFunc (GL_ZERO, GL_SRC_COLOR );
					break;
				case 'L':
					qglBlendFunc (GL_ZERO, GL_SRC_COLOR );
					break;
				case 'A':
				default:
					qglBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
					break;
				}
			}
			else
			{
				qglBlendFunc (GL_ZERO, GL_SRC_COLOR );
			}
		}
	}

	if ( currentmodel == r_worldmodel )
		c_visible_lightmaps = 0;

#ifdef SIN
   num_splatter_surfaces = 0;
#endif
	/*
	** render static lightmaps first
	*/
	for ( i = 1; i < MAX_LIGHTMAPS; i++ )
	{
		if ( gl_lms.lightmap_surfaces[i] )
		{
			if (currentmodel == r_worldmodel)
				c_visible_lightmaps++;
			GL_Bind( gl_state.lightmap_textures + i);

			for ( surf = gl_lms.lightmap_surfaces[i]; surf != 0; surf = surf->lightmapchain )
			{
				if ( surf->polys )
#ifdef SIN
               {
               if ( surf->splatter_surface && num_splatter_surfaces < MAX_RENDERED_SPLATTERS )
                  {
                  splatter_surfaces[ num_splatter_surfaces ] = surf;
                  }
					DrawGLPolyChain( surf->polys, 0, 0 );
               }
#else
					DrawGLPolyChain( surf->polys, 0, 0 );
#endif
			}
		}
	}
#ifdef SIN
   //
   // render the splatter maps
   //
/*
	qglBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
   for ( i = 0; i < num_splatter_surfaces; i++ )
      {
      splatter_t * splatter;
      glpoly_t * p;

      splatter = splatter_surfaces[ i ]->splatter_surface;
		GL_Bind( gl_state.splatter_textures + splatter->texture_num );
      p = surf->polys;
		for ( ; p != 0; p = p->chain )
		   {
			float *v;
			int j;

			qglBegin (GL_POLYGON);
			v = p->verts[0];
			for (j=0 ; j<p->numverts ; j++, v+= VERTEXSIZE)
			   {
				qglTexCoord2f( v[7] * splatter->s_mult, v[8] * splatter->t_mult );
				qglVertex3fv (v);
			   }
			qglEnd ();
         }
      }
	if (!gl_lightmap->value)
   	{
		if ( gl_saturatelighting->value )
   		{
			qglBlendFunc( GL_ONE, GL_ONE );
	   	}
		else
		   {
			if ( gl_monolightmap->string[0] != '0' )
			   {
				switch ( toupper( gl_monolightmap->string[0] ) )
				   {
				case 'I':
					qglBlendFunc (GL_ZERO, GL_SRC_COLOR );
					break;
				case 'L':
					qglBlendFunc (GL_ZERO, GL_SRC_COLOR );
					break;
				case 'A':
				default:
					qglBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
					break;
				   }
			   }
			else
			   {
				qglBlendFunc (GL_ZERO, GL_SRC_COLOR );
   			}
	   	}
	   }
*/
#endif

	/*
	** render dynamic lightmaps
	*/
	if ( gl_dynamic->value )
	{
		LM_InitBlock();

		GL_Bind( gl_state.lightmap_textures+0 );

		if (currentmodel == r_worldmodel)
			c_visible_lightmaps++;

		newdrawsurf = gl_lms.lightmap_surfaces[0];

		for ( surf = gl_lms.lightmap_surfaces[0]; surf != 0; surf = surf->lightmapchain )
		{
			int		smax, tmax;
			byte	*base;

			smax = (surf->extents[0]>>4)+1;
			tmax = (surf->extents[1]>>4)+1;

			if ( LM_AllocBlock( smax, tmax, &surf->dlight_s, &surf->dlight_t ) )
			{
				base = gl_lms.lightmap_buffer;
				base += ( surf->dlight_t * BLOCK_WIDTH + surf->dlight_s ) * LIGHTMAP_BYTES;

				GL_R_BuildLightMap (surf, base, BLOCK_WIDTH*LIGHTMAP_BYTES);
			}
			else
			{
				msurface_t *drawsurf;

				// upload what we have so far
				LM_UploadBlock( TRUE );

				// draw all surfaces that use this lightmap
				for ( drawsurf = newdrawsurf; drawsurf != surf; drawsurf = drawsurf->lightmapchain )
				{
					if ( drawsurf->polys )
						DrawGLPolyChain( drawsurf->polys,
							              ( drawsurf->light_s - drawsurf->dlight_s ) * ( 1.0 / 128.0 ),
										( drawsurf->light_t - drawsurf->dlight_t ) * ( 1.0 / 128.0 ) );
				}

				newdrawsurf = drawsurf;

				// clear the block
				LM_InitBlock();

				// try uploading the block now
				if ( !LM_AllocBlock( smax, tmax, &surf->dlight_s, &surf->dlight_t ) )
				{
					ri.Sys_Error( ERR_FATAL, "Consecutive calls to LM_AllocBlock(%d,%d) failed (dynamic)\n", smax, tmax );
				}

				base = gl_lms.lightmap_buffer;
				base += ( surf->dlight_t * BLOCK_WIDTH + surf->dlight_s ) * LIGHTMAP_BYTES;

				GL_R_BuildLightMap (surf, base, BLOCK_WIDTH*LIGHTMAP_BYTES);
			}
		}

		/*
		** draw remainder of dynamic lightmaps that haven't been uploaded yet
		*/
		if ( newdrawsurf )
			LM_UploadBlock( TRUE );

		for ( surf = newdrawsurf; surf != 0; surf = surf->lightmapchain )
		{
			if ( surf->polys )
				DrawGLPolyChain( surf->polys, ( surf->light_s - surf->dlight_s ) * ( 1.0 / 128.0 ), ( surf->light_t - surf->dlight_t ) * ( 1.0 / 128.0 ) );
		}
	}

	/*
	** restore state
	*/
	qglDisable (GL_BLEND);
	qglBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
#ifdef SIN
	if (gl_ztrick->value)
   	{
		if (ztrickframe & 1)
   		{
			qglDepthFunc (GL_LEQUAL);
   		}
		else
	   	{
			qglDepthFunc (GL_GEQUAL);
		   }
      }
   else
      qglDepthFunc (GL_LEQUAL);
#endif
	qglDepthMask( 1 );
}


/*
================
R_RenderBrushPoly
================
*/
void R_RenderBrushPoly (msurface_t *fa)
   {
	int			maps;
	image_t		*image;
	qboolean is_dynamic = FALSE;

#ifdef SIN
   if (fa->texinfo->flags & SURF_NODRAW)
      {
      return;
      }
  	GL_EnableMultitexture( FALSE );
#endif

	c_brush_polys++;
	image = R_TextureAnimation (fa->texinfo);

#ifdef SIN
	GL_BindImage( image );

   if ( fa->texinfo->flags & SURF_ADD )
      {
      GL_TexEnv( GL_MODULATE );
      qglEnable( GL_BLEND );
      qglColor4f(fa->texinfo->surf->color[0],fa->texinfo->surf->color[1],fa->texinfo->surf->color[2],1.0F );
      qglBlendFunc( GL_ONE, GL_ONE );
      qglDepthMask( 0 );
      DrawGLPoly( fa );
      qglDepthMask( 1 );
      qglDisable( GL_BLEND );
      return;
      }
   else if (fa->flags & SURF_DRAWTURB)
      {
      if (fa->texinfo->flags & SURF_WAVY)
         {
         // warp texture, no lightmaps
         GL_TexEnv( GL_MODULATE );
         qglColor4f( gl_state.inverse_intensity, gl_state.inverse_intensity, gl_state.inverse_intensity,1.0F );
         if (gl_drawflat->value)
            EmitWavySolidPolys(fa);
         else
            EmitWavyPolys (fa);
         return;
         }
      else
         {
         // warp texture, no lightmaps
         GL_TexEnv( GL_MODULATE );
         qglColor4f( gl_state.inverse_intensity,
         gl_state.inverse_intensity,
         gl_state.inverse_intensity,
         1.0F );
         if (gl_drawflat->value)
             EmitSolidPolys(fa);
         else
             EmitWaterPolys (fa);
         GL_TexEnv( GL_REPLACE );
         return;
         }
      }
   else if (fa->texinfo->flags & SURF_MASKED)
      {
      if (fa->texinfo->flags & SURF_NONLIT)
         {
         if (r_fullbright->value)
            qglColor4f( 1.0F,1.0F,1.0F,1.0F);
         else
            {
            if (fa->texinfo->flags & SURF_USECOLOR)
               {
               qglColor4f
                  (
                  fa->texinfo->surf->nonlit*fa->texinfo->surf->color[0],
                  fa->texinfo->surf->nonlit*fa->texinfo->surf->color[1],
                  fa->texinfo->surf->nonlit*fa->texinfo->surf->color[2],
			         1.0F
                  );
               }
            else
               {
               qglColor4f
                  (
                  fa->texinfo->surf->nonlit,
                  fa->texinfo->surf->nonlit,
                  fa->texinfo->surf->nonlit,
			         1.0F
                  );
               }
            }
         }
      else
         {
         qglColor3f( 1.0F, 1.0F, 1.0F);
         }
      GL_TexEnv(GL_MODULATE);
      qglEnable(GL_ALPHA_TEST);
      DrawGLPoly (fa);
      qglDisable (GL_ALPHA_TEST);
      GL_TexEnv(GL_REPLACE);
      qglColor4f( 1.0F, 1.0F, 1.0F, 1.0F);
      if (fa->texinfo->flags & SURF_NONLIT)
         {
         return;
         }
      }
   else if (fa->texinfo->flags & SURF_NONLIT)
      {
      GL_TexEnv( GL_MODULATE );
      if (r_fullbright->value)
         qglColor4f( 1.0F,1.0F,1.0F,1.0F);
      else
         {
         if (fa->texinfo->flags & SURF_USECOLOR)
            {
            qglColor4f
               (
               fa->texinfo->surf->nonlit*fa->texinfo->surf->color[0],
               fa->texinfo->surf->nonlit*fa->texinfo->surf->color[1],
               fa->texinfo->surf->nonlit*fa->texinfo->surf->color[2],
			      1.0F
               );
            }
         else
            {
            qglColor4f
               (
               fa->texinfo->surf->nonlit,
               fa->texinfo->surf->nonlit,
               fa->texinfo->surf->nonlit,
			      1.0F
               );
            }
         }
      DrawGLPoly (fa);
      GL_TexEnv( GL_REPLACE );
      qglColor4f( 1.0F, 1.0F, 1.0F, 1.0F);
      return;
      }
   else
      {
      GL_TexEnv( GL_REPLACE );
      DrawGLPoly (fa);
      }
#else
	GL_Bind( image->texnum );
    if (fa->flags & SURF_DRAWTURB)

        {
		// warp texture, no lightmaps
		GL_TexEnv( GL_MODULATE );
		qglColor4f( gl_state.inverse_intensity,
			        gl_state.inverse_intensity,
					gl_state.inverse_intensity,
					1.0F );
		EmitWaterPolys (fa);
		GL_TexEnv( GL_REPLACE );
		return;
    }
    else
	{
		GL_TexEnv( GL_REPLACE );
	}
    DrawGLPoly (fa->polys);
#endif

	/*
	** check for lightmap modification
	*/
	for ( maps = 0; maps < MAXLIGHTMAPS && fa->styles[maps] != 255; maps++ )
	{
#ifdef SIN
		if ( r_newrefdef.lightstyles[fa->styles[maps]].rgbvalue != fa->cached_light[maps] )
			goto dynamic;
#else
		if ( r_newrefdef.lightstyles[fa->styles[maps]].white != fa->cached_light[maps] )
			goto dynamic;
#endif
	}

	// dynamic this frame or dynamic previously
	if ( ( fa->dlightframe == r_framecount ) )
	{
dynamic:
		if ( gl_dynamic->value )
		{
#ifdef SIN
          	if (!( fa->texinfo->flags & (SURF_SKY|SURF_WARP|SURF_WAVY)))
#else
	    	if (!( fa->texinfo->flags & (SURF_SKY|SURF_TRANS33|SURF_TRANS66|SURF_WARP ) ) )
#endif
			{
#ifdef SIN
            if (fa->texinfo->surf->translucence == 0)
#endif
				is_dynamic = TRUE;
			}
		}
	}
	if ( is_dynamic )
	{
		if ( ( fa->styles[maps] >= 32 || fa->styles[maps] == 0 ) && ( fa->dlightframe != r_framecount ) )
		{
			Word32 temp[34*34];
			int			smax, tmax;

			smax = (fa->extents[0]>>4)+1;
			tmax = (fa->extents[1]>>4)+1;

			GL_R_BuildLightMap( fa, (Word8 *)temp, smax*4 );
			R_SetCacheState( fa );

			GL_Bind( gl_state.lightmap_textures + fa->lightmaptexturenum );

#if defined(BURGER_MAC) && TARGET_API_MAC_CARBON
			qglPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE,1);
#endif
			qglTexSubImage2D2( GL_TEXTURE_2D, 0,
							  fa->light_s, fa->light_t,
							  smax, tmax,
							  GL_LIGHTMAP_FORMAT,
							  GL_UNSIGNED_BYTE, temp );
			fa->lightmapchain = gl_lms.lightmap_surfaces[fa->lightmaptexturenum];
			gl_lms.lightmap_surfaces[fa->lightmaptexturenum] = fa;
		}
		else
		{
			fa->lightmapchain = gl_lms.lightmap_surfaces[0];
			gl_lms.lightmap_surfaces[0] = fa;
		}
	}
	else
	{
		fa->lightmapchain = gl_lms.lightmap_surfaces[fa->lightmaptexturenum];
		gl_lms.lightmap_surfaces[fa->lightmaptexturenum] = fa;
	}
}


/*
================
GL_R_DrawAlphaSurfaces

Draw water surfaces and windows.
The BSP tree is waled front to back, so unwinding the chain
of alpha_surfaces will draw back to front, giving proper ordering.
================
*/
void GL_R_DrawAlphaSurfaces (void)
{
	msurface_t	*s;
	float		intens;

	//
	// go back to the world matrix
	//
   qglLoadMatrixf (r_world_matrix);

	qglEnable (GL_BLEND);
	GL_TexEnv( GL_MODULATE );

	// the textures are prescaled up for a better lighting range,
	// so scale it back down
	intens = gl_state.inverse_intensity;

	for (s=r_alpha_surfaces ; s ; s=s->texturechain)
      {
#ifdef SIN

      qglEnable( GL_BLEND );
      GL_TexEnv( GL_MODULATE );

      if (s->texinfo->flags & SURF_NODRAW) continue;

      if ( s->owner )
         {
         qglPushMatrix ();
         // If this surface has an origin, translate to it.  This is
         // for alpha entity surfaces.
         qglTranslatef( s->owner->origin[0], s->owner->origin[1], s->owner->origin[2] );
         qglRotatef (s->owner->angles[1],  0, 0, 1);
         qglRotatef (s->owner->angles[0],  0, 1, 0);
         qglRotatef (s->owner->angles[2],  1, 0, 0);
         }
      GL_BindImage( s->texinfo->image );
#else
		GL_Bind( s->texinfo->image->texnum );
#endif
		c_brush_polys++;
#ifdef SIN
      if ( s->owner && s->owner->flags & RF_TRANSLUCENT )
         qglColor4f (intens,intens,intens,s->owner->alpha);
      else
         qglColor4f (intens,intens,intens,1 - s->texinfo->surf->translucence);
#else
      if (s->texinfo->flags & SURF_TRANS33)
         qglColor4f (intens,intens,intens,0.33);
		else if (s->texinfo->flags & SURF_TRANS66)
			qglColor4f (intens,intens,intens,0.66);
		else
			qglColor4f (intens,intens,intens,1);
#endif
		if (s->flags & SURF_DRAWTURB)
#ifdef SIN
         {
         if (s->texinfo->flags & SURF_WAVY)
            {
            if (gl_drawflat->value)
               EmitWavySolidPolys (s);
            else
               EmitWavyPolys (s);
            }
         else
            {
            if (gl_drawflat->value)
               EmitSolidPolys(s);
            else
               EmitWaterPolys (s);
            }
         }
      else if ( s->texinfo->flags & SURF_ADD )
         {
         qglDepthMask( 0 );
         qglColor4f(s->texinfo->surf->color[0],s->texinfo->surf->color[1],s->texinfo->surf->color[2],1.0F );
         qglBlendFunc( GL_ONE, GL_ONE );
         DrawGLPoly( s );
         qglColor4f( 1.0F, 1.0F, 1.0F, 1.0F);
         qglDepthMask( 1 );
         qglBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
         }
#else
         EmitWaterPolys (s);
#endif
		else
#ifdef SIN
         DrawGLPoly (s);
         if ( s->owner )
            qglPopMatrix();
#else
			DrawGLPoly (s->polys);
#endif
	}

	GL_TexEnv( GL_REPLACE );
	qglColor4f (1,1,1,1);
	qglDisable (GL_BLEND);

	r_alpha_surfaces = NULL;
}

/*
================
DrawTextureChains
================
*/
static void DrawTextureChains (void)
{
	int		i;
	msurface_t	*s;
	image_t		*image;

	c_visible_textures = 0;

   for (i=0, image=gltextures ; i<numgltextures ; i++,image++)
      {
      if (!image->registration_sequence)
         continue;
      s = image->texturechain;
      if (!s)
         continue;
      c_visible_textures++;
#ifdef SIN
      for ( ; s ; s=s->texturechain)
         {
         R_RenderBrushPoly (s);
         GL_R_DrawConsole(s,1,zero);
         }
#else
      for ( ; s ; s=s->texturechain)
         R_RenderBrushPoly (s);
#endif

      image->texturechain = NULL;
      }

	GL_TexEnv( GL_REPLACE );
}

static void GL_RenderLightmappedPoly( msurface_t *surf )
{
	int		i, nv;
	int		map;
	float	*v;
   image_t *image;
	qboolean is_dynamic = FALSE;
	unsigned lmtex;
	glpoly_t *p;

   if (surf->texinfo->flags & SURF_NODRAW)
      {
      return;
      }

   image = R_TextureAnimation( surf->texinfo );
	lmtex = surf->lightmaptexturenum;
	nv = surf->polys->numverts;

	for ( map = 0; map < MAXLIGHTMAPS && surf->styles[map] != 255; map++ )
	{
		if ( r_newrefdef.lightstyles[surf->styles[map]].rgbvalue != surf->cached_light[map] )
			goto dynamic;
	}

	// dynamic this frame or dynamic previously
	if ( ( surf->dlightframe == r_framecount ) )
	{
dynamic:
		if ( gl_dynamic->value )
		{
       		if ( !(surf->texinfo->flags & (SURF_SKY|SURF_WARP|SURF_WAVY)))
			{
            if (surf->texinfo->surf->translucence == 0)
				is_dynamic = TRUE;
			}
		}
	}

	if ( is_dynamic ) {
//		Word32 *temp = alloca(sizeof(Word32) * 128*128);
		static Word32 temp[128*128];
		int smax, tmax;

		if ( ( surf->styles[map] >= 32 || surf->styles[map] == 0 ) && ( surf->dlightframe != r_framecount ) )
		{
			smax = (surf->extents[0]>>4)+1;
			tmax = (surf->extents[1]>>4)+1;

			GL_R_BuildLightMap( surf, (Word8 *)temp, smax*4 );
			R_SetCacheState( surf );

			GL_MBind( GL_TEXTURE1_SGIS, gl_state.lightmap_textures + surf->lightmaptexturenum );

			lmtex = surf->lightmaptexturenum;

#if defined(BURGER_MAC) && TARGET_API_MAC_CARBON
			qglPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE,1);
#endif
			qglTexSubImage2D2(GL_TEXTURE_2D,0,surf->light_s, surf->light_t,smax, tmax,
				GL_LIGHTMAP_FORMAT,GL_UNSIGNED_BYTE,temp);
		} else {
			smax = (surf->extents[0]>>4)+1;
			tmax = (surf->extents[1]>>4)+1;

			GL_R_BuildLightMap( surf, (Word8 *)temp, smax*4 );
			GL_MBind( GL_TEXTURE1_SGIS, gl_state.lightmap_textures + 0 );
			lmtex = 0;

#if defined(BURGER_MAC) && TARGET_API_MAC_CARBON
			qglPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE,1);
#endif
			qglTexSubImage2D2(GL_TEXTURE_2D,0,surf->light_s, surf->light_t,smax,tmax,
				GL_LIGHTMAP_FORMAT,GL_UNSIGNED_BYTE,temp);
		}
		c_brush_polys++;
		GL_MBindImage( GL_TEXTURE0_SGIS, image );
		GL_MBind( GL_TEXTURE1_SGIS, gl_state.lightmap_textures + lmtex );
	} else {
		c_brush_polys++;
		GL_MBindImage( GL_TEXTURE0_SGIS, image );
		GL_MBind( GL_TEXTURE1_SGIS, gl_state.lightmap_textures + lmtex );
	}
   if (surf->texinfo->flags & SURF_ENVMAPPED) {
      vec3_t eye_normal;
      vec3_t eye_vert;
      vec3_t normal;
      vec3_t transform[3];
      float two_nu, m, fx, fy, fz;

      if ( surf->flags & SURF_PLANEBACK )
         VectorNegate(surf->plane->normal, normal);
      else
         VectorCopy(surf->plane->normal, normal);

	   VectorCopy (vright, transform[0]);
	   VectorNegate (vup, transform[1]);
	   VectorCopy (vpn, transform[2]);

		eye_normal[0] = DotProduct(normal, transform[0]);
		eye_normal[1] = DotProduct(normal, transform[1]);
		eye_normal[2] = DotProduct(normal, transform[2]);

	   for ( p = surf->polys; p; p = p->chain )
	      {
         qglBegin (GL_POLYGON);
         v = p->verts[0];
         for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
            {
		      eye_vert[0] = DotProduct(v, transform[0]) - r_origin[0];
		      eye_vert[1] = DotProduct(v, transform[1]) - r_origin[1];
		      eye_vert[2] = DotProduct(v, transform[2]) - r_origin[2];
            VectorNormalize( eye_vert );
            two_nu = 2.0f * DotProduct( eye_normal, eye_vert );
            fx = eye_vert[0] - eye_normal[0] * two_nu;
            fy = eye_vert[1] - eye_normal[1] * two_nu;
            fz = eye_vert[2] - eye_normal[2] * two_nu;
            m = 2.0F * Burger::FPMath::Sqrt( fx*fx + fy*fy + (fz+1.0F)*(fz+1.0F) );
            if (m==0.0F)
               {
		         qglMTexCoord2fSGIS( GL_TEXTURE0_SGIS, 0.5f, 0.5f);
               qglMTexCoord2fSGIS( GL_TEXTURE1_SGIS, v[5], v[6]);
               }
            else
               {
               float mInv = 1.0F / m;
		         qglMTexCoord2fSGIS( GL_TEXTURE0_SGIS, fx * mInv + 0.5f, fy * mInv + 0.5F);
               qglMTexCoord2fSGIS( GL_TEXTURE1_SGIS, v[5], v[6]);
               }
            qglVertex3fv (v);
            }
         qglEnd ();
         }
      }
   else if (surf->texinfo->flags & SURF_MASKED)
      {
      //
      // masked but lit
      // non-lit goes to RenderBrushPoly
      //
      qglEnable(GL_ALPHA_TEST);
	   for ( p = surf->polys; p; p = p->chain )
	      {
		   v = p->verts[0];
		   qglBegin (GL_POLYGON);
		   for (i=0 ; i< nv; i++, v+= VERTEXSIZE)
		      {
		      qglMTexCoord2fSGIS( GL_TEXTURE0_SGIS, v[3] + surf->texinfo->surf->delta_s, v[4] + surf->texinfo->surf->delta_t);
            qglMTexCoord2fSGIS( GL_TEXTURE1_SGIS, v[5], v[6]);
		      qglVertex3fv (v);
		      }
         qglEnd ();
	      }
      qglDisable (GL_ALPHA_TEST);
      }
   else
      {     
      for ( p = surf->polys; p; p = p->chain )
	      {
		   v = p->verts[0];
		   qglBegin (GL_POLYGON);
		   for (i=0 ; i< nv; i++, v+= VERTEXSIZE)
		      {
            qglMTexCoord2fSGIS( GL_TEXTURE0_SGIS, v[3] + surf->texinfo->surf->delta_s, v[4] + surf->texinfo->surf->delta_t);
            qglMTexCoord2fSGIS( GL_TEXTURE1_SGIS, v[5], v[6]);
		      qglVertex3fv (v);
		      }
         qglEnd ();
	      }
      if ( surf->texinfo->flags & SURF_ADD )
         {
         qglDisable( GL_BLEND );
         }
      }
   //
   // render the splatter map
   //
}


/*
=================
R_DrawInlineBModel
=================
*/
static void R_DrawInlineBModel (void)
{
	int			i, k;
	cplane_t	*pplane;
	float		dot;
	msurface_t	*psurf;
	dlight_t	*lt;

	// calculate dynamic lighting for bmodel
	if ( !gl_flashblend->value )
	{
		lt = r_newrefdef.dlights;
		for (k=0 ; k<r_newrefdef.num_dlights ; k++, lt++)
		{
			GL_R_MarkLights (lt, 1<<k, currentmodel->nodes + currentmodel->firstnode);
		}
	}

	psurf = &currentmodel->surfaces[currentmodel->firstmodelsurface];

	//
	// draw texture
	//
	for (i=0 ; i<currentmodel->nummodelsurfaces ; i++, psurf++)
	{
	// find which side of the node we are on
		pplane = psurf->plane;

		dot = DotProduct (modelorg, pplane->normal) - pplane->dist;

	// draw the polygon
      if  ( (currententity->flags & RF_TRANSLUCENT) ||
          ( (psurf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) ||
			 (!(psurf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)))
		{
         if ( (psurf->texinfo->surf->translucence > 0) || (currententity->flags & RF_TRANSLUCENT) || ( psurf->texinfo->flags & SURF_ADD ) )
            {	// add to the translucent chain
				psurf->texturechain = r_alpha_surfaces;
				r_alpha_surfaces = psurf;
            // save off the owner of this surface
            psurf->owner = currententity;
			   }
//      	else if ( qglMTexCoord2fSGIS && !( psurf->flags & (SURF_DRAWTURB|SURF_DRAWNONLIT|SURF_DRAWMASKED) ) && !gl_drawflat->value )
      	else if ( qglMTexCoord2fSGIS && !( psurf->flags & (SURF_DRAWTURB|SURF_DRAWNONLIT) ) && !gl_drawflat->value && !r_fullbright->value &&
                   !(psurf->texinfo->flags & SURF_ADD ) ) 
			{
            GL_RenderLightmappedPoly( psurf );
            GL_R_DrawConsole(psurf,1,zero);
			}
			else
			{
            GL_EnableMultitexture( FALSE );
            R_RenderBrushPoly (psurf);
            GL_EnableMultitexture( TRUE );
         }
		}
	}

	if ( !(currententity->flags & RF_TRANSLUCENT) )
	{
		if ( !qglMTexCoord2fSGIS )
			R_BlendLightmaps ();
	}
	else
	{
		qglDisable (GL_BLEND);
		qglColor4f (1,1,1,1);
		GL_TexEnv( GL_REPLACE );
	}
}

/*
=================
R_DrawBrushModel
=================
*/
void R_DrawBrushModel (entity_t *e)
{
	vec3_t		mins, maxs;
	int			i;
	qboolean	rotated;

	if (currentmodel->nummodelsurfaces == 0)
		return;

	currententity = e;
	gl_state.currenttextures[0] = gl_state.currenttextures[1] = -1;

	if (e->angles[0] || e->angles[1] || e->angles[2])
	{
		rotated = TRUE;
		for (i=0 ; i<3 ; i++)
		{
			mins[i] = e->origin[i] - currentmodel->radius;
			maxs[i] = e->origin[i] + currentmodel->radius;
		}
	}
	else
	{
		rotated = FALSE;
		VectorAdd (e->origin, currentmodel->mins, mins);
		VectorAdd (e->origin, currentmodel->maxs, maxs);
	}

	if (R_CullBox (mins, maxs))
		return;

	qglColor3f (1,1,1);
	FastMemSet (gl_lms.lightmap_surfaces, 0, sizeof(gl_lms.lightmap_surfaces));

	VectorSubtract (r_newrefdef.vieworg, e->origin, modelorg);
#ifdef SIN
	if (rotated)
	{
		vec3_t	temp;

		VectorCopy (modelorg, temp);
		modelorg[0] = DotProduct (temp, e->mat[0]);
		modelorg[1] = DotProduct (temp, e->mat[1]);
		modelorg[2] = DotProduct (temp, e->mat[2]);
	}
#else
	if (rotated)
	{
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (modelorg, temp);
		AngleVectors (e->angles, forward, right, up);
		modelorg[0] = DotProduct (temp, forward);
		modelorg[1] = -DotProduct (temp, right);
		modelorg[2] = DotProduct (temp, up);
	}
#endif



#ifdef SIN
   qglPushMatrix ();

e->angles[0] = -e->angles[0];	// stupid quake bug
e->angles[2] = -e->angles[2];	// stupid quake bug
	R_RotateForEntity (e);
e->angles[0] = -e->angles[0];	// stupid quake bug
e->angles[2] = -e->angles[2];	// stupid quake bug

	GL_EnableMultitexture( TRUE );
	R_DrawInlineBModel ();
   GL_EnableMultitexture( FALSE );


   qglPopMatrix ();

   qglPushMatrix ();
   qglTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);
   qglRotatef (e->angles[1],  0, 0, 1);
   qglRotatef (-e->angles[0],  0, 1, 0);
   qglRotatef (e->angles[2],  1, 0, 0);
   GL_EnableMultitexture( FALSE );
   GL_R_DrawConsole( &e->model->surfaces[e->model->firstmodelsurface], e->model->nummodelsurfaces, e->origin);
   qglPopMatrix ();
#else
    qglPushMatrix ();

e->angles[0] = -e->angles[0];	// stupid quake bug
e->angles[2] = -e->angles[2];	// stupid quake bug
	R_RotateForEntity (e);
e->angles[0] = -e->angles[0];	// stupid quake bug
e->angles[2] = -e->angles[2];	// stupid quake bug

	GL_EnableMultitexture( TRUE );
	GL_SelectTexture( GL_TEXTURE0_SGIS );
	GL_TexEnv( GL_REPLACE );
	GL_SelectTexture( GL_TEXTURE1_SGIS );
	GL_TexEnv( GL_MODULATE );

	R_DrawInlineBModel ();

	GL_EnableMultitexture( FALSE );
	qglPopMatrix ();
#endif
}

/*
=============================================================

	WORLD MODEL

=============================================================
*/

/*
================
R_RecursiveWorldNode
================
*/
static void R_RecursiveWorldNode (mnode_t *node)
{
	int			c, side, sidebit;
	cplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	float		dot;
	image_t		*image;

	if (node->contents == CONTENTS_SOLID)
		return;		// solid

	if (node->visframe != r_visframecount)
		return;
	if (R_CullBox (node->minmaxs, node->minmaxs+3))
		return;

// if a leaf node, draw stuff
	if (node->contents != -1)
	{
		pleaf = (mleaf_t *)node;

		// check for door connected areas
		if (r_newrefdef.areabits)
		{
			if (! (r_newrefdef.areabits[pleaf->area>>3] & (1<<(pleaf->area&7)) ) )
				return;		// not visible
		}

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

		return;
	}

// node is just a decision point, so go down the apropriate sides

// find which side of the node we are on
	plane = node->plane;

	switch (plane->type)
	{
	case PLANE_X:
		dot = modelorg[0] - plane->dist;
		break;
	case PLANE_Y:
		dot = modelorg[1] - plane->dist;
		break;
	case PLANE_Z:
		dot = modelorg[2] - plane->dist;
		break;
	default:
		dot = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	if (dot >= 0)
	{
		side = 0;
		sidebit = 0;
	}
	else
	{
		side = 1;
		sidebit = SURF_PLANEBACK;
	}

// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	// add translucent surfaces to list in bsp order
   for ( c = node->numsurfaces, surf = r_worldmodel->surfaces + node->firstsurface; c ; c--, surf++)
	{
		if (surf->visframe != r_framecount)
			continue;

		if ( (surf->flags & SURF_PLANEBACK) != sidebit )
			continue;		// wrong side

      if (surf->texinfo->flags & SURF_SKY)
      {  // just adds to visible sky bounds
         R_AddSkySurface (surf);
      }

#ifdef SIN
       // Sin has arbitrary translucency
      else if ( ( surf->texinfo->surf->translucence > 0) || ( surf->texinfo->flags & SURF_ADD ) )
#else
      if (surf->texinfo->flags & (SURF_TRANS33|SURF_TRANS66))
#endif

         {  // add to the translucent chain
			surf->texturechain = r_alpha_surfaces;
			r_alpha_surfaces = surf;
         }
      else 
         {
         if ( qglMTexCoord2fSGIS && !( surf->flags & (SURF_DRAWTURB|SURF_DRAWNONLIT) ) && !gl_drawflat->value && !r_fullbright->value )
            {
            GL_RenderLightmappedPoly( surf );
#ifdef SIN
            GL_R_DrawConsole( surf, 1, zero );
#endif
            }
         else
            {
            // the polygon is visible, so add it to the texture
            // sorted chain
            // FIXME: this is a hack for animation
            image = R_TextureAnimation (surf->texinfo);
            surf->texturechain = image->texturechain;
            image->texturechain = surf;
            }
         }
   }

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
/*
// draw stuff
	c = node->numsurfaces;
	surf = r_worldmodel->surfaces + node->firstsurface;

	for ( ; c ; c--, surf++)
	{
		if (surf->visframe != r_framecount)
			continue;

		if ( (surf->flags & SURF_PLANEBACK) != sidebit )
			continue;		// wrong side

		if (surf->texinfo->flags & SURF_SKY)
		{	// just adds to visible sky bounds
			R_AddSkySurface (surf);
		}
#ifdef SIN
	    else if (surf->texinfo->translucence > 0)
#else
		else if (surf->texinfo->flags & (SURF_TRANS33|SURF_TRANS66))
#endif
		{	// add to the translucent chain
//			surf->texturechain = alpha_surfaces;
//			alpha_surfaces = surf;
		}
		else
		{
#ifdef SIN
//      	if ( qglMTexCoord2fSGIS && !( surf->flags & (SURF_DRAWTURB|SURF_DRAWNONLIT|SURF_DRAWMASKED) ) && !gl_drawflat->value )
      	if ( qglMTexCoord2fSGIS && !( surf->flags & (SURF_DRAWTURB|SURF_DRAWNONLIT) ) && !gl_drawflat->value && !r_fullbright->value )
#else
			if ( qglMTexCoord2fSGIS && !( surf->flags & SURF_DRAWTURB ) )
#endif
			{
				GL_RenderLightmappedPoly( surf );
			}
			else
			{
				// the polygon is visible, so add it to the texture
				// sorted chain
				// FIXME: this is a hack for animation
				image = R_TextureAnimation (surf->texinfo);
				surf->texturechain = image->texturechain;
				image->texturechain = surf;
			}
		}
	}
*/
}


/*
=============
R_DrawWorld
=============
*/
void R_DrawWorld (void)
{
	entity_t	ent;

   if (!r_drawworld->value)
		return;

	if ( r_newrefdef.rdflags & RDF_NOWORLDMODEL )
		return;

	currentmodel = r_worldmodel;

	VectorCopy (r_newrefdef.vieworg, modelorg);

	// auto cycle the world frame for texture animation
	FastMemSet (&ent, 0, sizeof(ent));
    ent.frame = (int)(r_newrefdef.time*2);
#ifdef SIN
    c_randval  = rand();
#endif
    currententity = &ent;

	gl_state.currenttextures[0] = gl_state.currenttextures[1] = -1;

	qglColor3f (1,1,1);
	FastMemSet (gl_lms.lightmap_surfaces, 0, sizeof(gl_lms.lightmap_surfaces));
	R_ClearSkyBox ();

	if ( qglMTexCoord2fSGIS ) {
		GL_EnableMultitexture( TRUE );
	
		if ( gl_lightmap->value ) {
			GL_SelectTexture( GL_TEXTURE1_SGIS );
			GL_TexEnv( GL_REPLACE );
		}
		R_RecursiveWorldNode (r_worldmodel->nodes);
		GL_EnableMultitexture( FALSE );
	} else {
		R_RecursiveWorldNode (r_worldmodel->nodes);
	}

	/*
	** theoretically nothing should happen in the next two functions
	** if multitexture is enabled
	*/
	DrawTextureChains ();
	R_BlendLightmaps ();

	R_DrawSkyBox ();

	R_DrawTriangleOutlines ();
}


/*
===============
GL_R_MarkLeaves

Mark the leaves and nodes that are in the PVS for the current
cluster
===============
*/

#ifndef AMD3D

void GL_R_MarkLeaves (void)
{
	byte	*vis;
	byte	fatvis[MAX_MAP_LEAFS/8];
	mnode_t	*node;
	int		i, c;
	mleaf_t	*leaf;
	int		cluster;

	if (r_oldviewcluster == r_viewcluster && r_oldviewcluster2 == r_viewcluster2 && !r_novis->value && r_viewcluster != -1)
		return;

   // development aid to let you run around and see exactly where
   // the pvs ends
   if (gl_lockpvs->value)
      return;

   r_visframecount++;
	r_oldviewcluster = r_viewcluster;
	r_oldviewcluster2 = r_viewcluster2;

	if (r_novis->value || r_viewcluster == -1 || !r_worldmodel->vis)
	{
		// mark everything
		for (i=0 ; i<r_worldmodel->numleafs ; i++)
			r_worldmodel->leafs[i].visframe = r_visframecount;
		for (i=0 ; i<r_worldmodel->numnodes ; i++)
			r_worldmodel->nodes[i].visframe = r_visframecount;
		return;
	}

	vis = GL_Mod_ClusterPVS (r_viewcluster, r_worldmodel);
	// may have to combine two clusters because of solid water boundaries
	if (r_viewcluster2 != r_viewcluster)
	{
		FastMemCpy(fatvis, vis, (r_worldmodel->numleafs+7)/8);
		vis = GL_Mod_ClusterPVS (r_viewcluster2, r_worldmodel);
		c = (r_worldmodel->numleafs+31)/32;
		for (i=0 ; i<c ; i++)
			((int *)fatvis)[i] |= ((int *)vis)[i];
		vis = fatvis;
	}

	for (i=0,leaf=r_worldmodel->leafs ; i<r_worldmodel->numleafs ; i++, leaf++)
	{
		cluster = leaf->cluster;
		if (cluster == -1)
			continue;
		if (vis[cluster>>3] & (1<<(cluster&7)))
		{
			node = (mnode_t *)leaf;
			do
			{
				if (node->visframe == r_visframecount)
					break;
				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}

}

#endif



/*
=============================================================================

  LIGHTMAP ALLOCATION

=============================================================================
*/

static void LM_InitBlock( void )
{
	FastMemSet( gl_lms.allocated, 0, sizeof( gl_lms.allocated ) );
}

static void LM_UploadBlock( qboolean dynamic )
{
	int texture;
	int height = 0;

	if ( dynamic ) {
		texture = 0;
	} else {
		texture = gl_lms.current_lightmap_texture;
	}

	GL_Bind( gl_state.lightmap_textures + texture );
	qglTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	qglTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	if ( dynamic ) {
		int i;

		for ( i = 0; i < BLOCK_WIDTH; i++ ) {
			if ( gl_lms.allocated[i] > height ) {
				height = gl_lms.allocated[i];
			}
		}
#if defined(BURGER_MAC) && TARGET_API_MAC_CARBON
		qglPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE,1);
#endif
		qglTexSubImage2D2( GL_TEXTURE_2D,0,
			0, 0,
			BLOCK_WIDTH, height,
			GL_LIGHTMAP_FORMAT,
			GL_UNSIGNED_BYTE,
			gl_lms.lightmap_buffer );
	} else {
		qglTexImage2D( GL_TEXTURE_2D,0,gl_lms.internal_format,BLOCK_WIDTH, BLOCK_HEIGHT,0,GL_LIGHTMAP_FORMAT,GL_UNSIGNED_BYTE,gl_lms.lightmap_buffer );
		if ( ++gl_lms.current_lightmap_texture == MAX_LIGHTMAPS ) {
			ri.Sys_Error( ERR_DROP, "LM_UploadBlock() - MAX_LIGHTMAPS exceeded\n" );
		}
	}
}

// returns a texture number and the position inside it
static qboolean LM_AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;

	best = BLOCK_HEIGHT;

	for (i=0 ; i<BLOCK_WIDTH-w ; i++)
	{
		best2 = 0;

		for (j=0 ; j<w ; j++)
		{
			if (gl_lms.allocated[i+j] >= best)
				break;
			if (gl_lms.allocated[i+j] > best2)
				best2 = gl_lms.allocated[i+j];
		}
		if (j == w)
		{	// this is a valid spot
			*x = i;
			*y = best = best2;
		}
	}

	if (best + h > BLOCK_HEIGHT)
		return FALSE;

	for (i=0 ; i<w ; i++)
		gl_lms.allocated[*x + i] = best + h;

	return TRUE;
}

/*
================
GL_BuildPolygonFromSurface
================
*/
void GL_BuildPolygonFromSurface(msurface_t *fa)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	int			vertpage;
	float		*vec;
	float		s, t;
	glpoly_t	*poly;
	vec3_t		total;

// reconstruct the polygon
	pedges = currentmodel->edges;
	lnumverts = fa->numedges;
	vertpage = 0;

	VectorClear (total);
	//
	// draw texture
	//
	poly = (glpoly_t*)Hunk_Alloc (sizeof(glpoly_t) + (lnumverts-4) * VERTEXSIZE*sizeof(float));
	poly->next = fa->polys;
	poly->flags = fa->flags;
	fa->polys = poly;
	poly->numverts = lnumverts;

	for (i=0 ; i<lnumverts ; i++)
	{
		lindex = currentmodel->surfedges[fa->firstedge + i];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = currentmodel->vertexes[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = currentmodel->vertexes[r_pedge->v[1]].position;
		}
		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		s /= fa->texinfo->image->width;

		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
		t /= fa->texinfo->image->height;

		VectorAdd (total, vec, total);
		VectorCopy (vec, poly->verts[i]);
		poly->verts[i][3] = s;
		poly->verts[i][4] = t;

		//
		// lightmap texture coordinates
		//
		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		s -= fa->texturemins[0];
		s += fa->light_s*16;
		s += 8;
		s /= BLOCK_WIDTH*16; //fa->texinfo->texture->width;

		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
		t -= fa->texturemins[1];
		t += fa->light_t*16;
		t += 8;
		t /= BLOCK_HEIGHT*16; //fa->texinfo->texture->height;

		poly->verts[i][5] = s;
		poly->verts[i][6] = t;
	}

	poly->numverts = lnumverts;

}

/*
========================
GL_CreateSurfaceLightmap
========================
*/
void GL_CreateSurfaceLightmap (msurface_t *surf)
{
	int		smax, tmax;
	byte	*base;

    if (surf->flags & (SURF_DRAWSKY|SURF_DRAWTURB))
        return;

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;

	if ( !LM_AllocBlock( smax, tmax, &surf->light_s, &surf->light_t ) )
	{
		LM_UploadBlock( FALSE );
		LM_InitBlock();
		if ( !LM_AllocBlock( smax, tmax, &surf->light_s, &surf->light_t ) )
		{
			ri.Sys_Error( ERR_FATAL, "Consecutive calls to LM_AllocBlock(%d,%d) failed\n", smax, tmax );
		}
	}

	surf->lightmaptexturenum = gl_lms.current_lightmap_texture;

	base = gl_lms.lightmap_buffer;
	base += (surf->light_t * BLOCK_WIDTH + surf->light_s) * LIGHTMAP_BYTES;

	R_SetCacheState( surf );
	GL_R_BuildLightMap (surf, base, BLOCK_WIDTH*LIGHTMAP_BYTES);
}


/*
==================
GL_BeginBuildingLightmaps

==================
*/
void GL_BeginBuildingLightmaps (model_t *m)
{
	static lightstyle_t	lightstyles[MAX_LIGHTSTYLES];
	int i;
	static unsigned dummy[128*128];

	FastMemSet( gl_lms.allocated, 0, sizeof(gl_lms.allocated) );

	r_framecount = 1;		// no dlightcache

	GL_EnableMultitexture( TRUE );
	GL_SelectTexture( GL_TEXTURE1_SGIS );

	/*
	** setup the base lightstyles so the lightmaps won't have to be regenerated
	** the first time they're seen
	*/
	for (i=0 ; i<MAX_LIGHTSTYLES ; i++)
	{
		lightstyles[i].rgb[0] = 1;
		lightstyles[i].rgb[1] = 1;
		lightstyles[i].rgb[2] = 1;
		lightstyles[i].white = 3;
	}
	r_newrefdef.lightstyles = lightstyles;

	if (!gl_state.lightmap_textures)
	{
		gl_state.lightmap_textures	= TEXNUM_LIGHTMAPS;
//		gl_state.lightmap_textures	= gl_state.texture_extension_number;
//		gl_state.texture_extension_number = gl_state.lightmap_textures + MAX_LIGHTMAPS;
	}

	gl_lms.current_lightmap_texture = 1;

	/*
	** if mono lightmaps are enabled and we want to use alpha
	** blending (a,1-a) then we're likely running on a 3DLabs
	** Permedia2.  In a perfect world we'd use a GL_ALPHA lightmap
	** in order to conserve space and maximize bandwidth, however
	** this isn't a perfect world.
	**
	** So we have to use alpha lightmaps, but stored in GL_RGBA format,
	** which means we only get 1/16th the color resolution we should when
	** using alpha lightmaps.  If we find another board that supports
	** only alpha lightmaps but that can at least support the GL_ALPHA
	** format then we should change this code to use real alpha maps.
	*/
	if ( toupper( gl_monolightmap->CVarString[0] ) == 'A' )
	{
      gl_lms.internal_format = gl_tex_alpha_format;
	}
	/*
	** try to do hacked colored lighting with a blended texture
	*/
	else if ( toupper( gl_monolightmap->CVarString[0] ) == 'C' )
	{
      gl_lms.internal_format = gl_tex_alpha_format;
	}
	else if ( toupper( gl_monolightmap->CVarString[0] ) == 'I' )
	{
		gl_lms.internal_format = GL_INTENSITY8;
	}
	else if ( toupper( gl_monolightmap->CVarString[0] ) == 'L' )
	{
		gl_lms.internal_format = GL_LUMINANCE8;
	}
	else
	{
      gl_lms.internal_format = gl_tex_solid_format;
	}

	/*
	** initialize the dynamic lightmap texture
	*/
	GL_Bind( gl_state.lightmap_textures + 0 );
	qglTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	qglTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	qglTexImage2D( GL_TEXTURE_2D,0,gl_lms.internal_format,BLOCK_WIDTH, BLOCK_HEIGHT,0,GL_LIGHTMAP_FORMAT,GL_UNSIGNED_BYTE,dummy );
}

/*
=======================
GL_EndBuildingLightmaps
=======================
*/
void GL_EndBuildingLightmaps (void)
{
	LM_UploadBlock( FALSE );
	GL_EnableMultitexture( FALSE );
}

#ifdef SIN
/*
=======================
GL_UpdateLightmap
=======================
*/
void GL_UpdateLightmap( msurface_t * surf )
   {
//	unsigned *	temp = alloca(sizeof(unsigned) * 128*128);
	static Word32 temp[128*128];
	int			smax, tmax;

	GL_EnableMultitexture( TRUE );
	GL_SelectTexture( GL_TEXTURE1_SGIS );

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;

	GL_R_BuildLightMap( surf, (Word8 *)temp, smax*4 );
	R_SetCacheState( surf );

	GL_MBind( GL_TEXTURE1_SGIS, gl_state.lightmap_textures + surf->lightmaptexturenum );

#if defined(BURGER_MAC) && TARGET_API_MAC_CARBON
	qglPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE,1);
#endif
	qglTexSubImage2D2( GL_TEXTURE_2D, 0,
					  surf->light_s, surf->light_t,
					  smax, tmax,
					  GL_LIGHTMAP_FORMAT,
					  GL_UNSIGNED_BYTE, temp );
	GL_EnableMultitexture( FALSE );
	}
#endif