//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/snd_mem.c                                              $
// $Revision:: 3                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 8/8/01 4:49 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// sound caching
//

#include "client.h"
#include "snd_loc.h"

#ifdef SIN
#include "wave.h"
#endif

int			cache_full_cycle;

byte *S_Alloc (int size);

#ifdef SIN
/*
================
ReformatStereoIMAADPCM_Sfx
================
*/
static void ReformatStereoIMAADPCM_Sfx(sfx_t *sfx, byte *data)
   {
   int      blockcount;
   int      subblockcount;
	sfxcache_t	*sc;
   byte     *dest;
	
	sc = sfx->cache;
	if (!sc)
		return;

   blockcount = sc->real_length / sc->blockalign;
   if (blockcount * sc->blockalign != sc->real_length)
      {
		Com_DPrintf ("%s is not an even block multiple\n",sfx->name);
      }
   subblockcount = (sc->blockalign - (sizeof(unsigned) * 2)) >> 3;
   dest = sc->data;
   while( blockcount--)
      {
      int i;
      int j;
      //
      // copy over the block header
      //
      FastMemCpy( dest, data, sizeof(unsigned) * 2 );
      data += sizeof(unsigned) * 2;
      dest += sizeof(unsigned) * 2;
      for( j=0; j < subblockcount; j++)
         {
         unsigned dwLeft;
         unsigned dwRight;
         //
         // grab each 8 samples
         //
         dwLeft   = LittleLong(*(unsigned *)data);
         data   += sizeof(unsigned);
         dwRight  = LittleLong(*(unsigned *)data);
         data   += sizeof(unsigned);

         for( i=8; i>0; i-- )
            {
            int leftcode;
            int rightcode;

            leftcode  = (int)(dwLeft & 0x0F);
            rightcode  = (int)(dwRight & 0x0F);
            *(dest++) = leftcode + (rightcode<<4);

            dwLeft >>= 4;
            dwRight >>= 4;
            }
         }
      }
   }
#endif

/*
================
ResampleSfx
================
*/
static void ResampleSfx (sfx_t *sfx, int inrate, int inwidth, byte *data)
{
#ifdef SIN
	int		outcount;
	int		i;
	int		sample;
	sfxcache_t	*sc;
	
	sc = sfx->cache;
	if (!sc)
		return;

   outcount = sc->real_length;
   if (sc->type == WAV_IMA_ADPCM_FORMAT)
      {
      if (sc->stereo)
         {
         ReformatStereoIMAADPCM_Sfx(sfx, data);
//         FastMemCpy( sc->data, data, outcount );
         }
      else
         {
         // no need to reformat
         FastMemCpy( sc->data, data, outcount );
         }
      return;
      }
   else 
      {
	   if (s_loadas8bit->value)
		   sc->width = 1;
	   else
		   sc->width = inwidth;

      // resample / decimate to the current source rate
	   if (inwidth == 1 && sc->width == 1)
	      {
         // fast special case
		   for (i=0 ; i<outcount ; i++)
			   ((signed char *)sc->data)[i]
			   = (int)( (unsigned char)(data[i]) - 128);
	      }
	   else
	      {
         // general case
		   for (i=0 ; i<outcount ; i++)
		      {
			   if (inwidth == 2)
				   sample = LittleShort ( ((signed short *)data)[i] );
			   else
				   sample = (int)( (unsigned char)(data[i]) - 128) << 8;
			   if (sc->width == 2)
				   ((signed short *)sc->data)[i] = sample;
			   else
				   ((signed char *)sc->data)[i] = sample >> 8;
		      }
	      }
      }
#else
	int		outcount;
	int		srcsample;
	float	stepscale;
	int		i;
	int		sample, samplefrac, fracstep;
	sfxcache_t	*sc;
	
	sc = sfx->cache;
	if (!sc)
		return;

	stepscale = (float)inrate / dma.speed;	// this is usually 0.5, 1, or 2

	outcount = sc->length / stepscale;
	sc->length = outcount;
	if (sc->loopstart != -1)
		sc->loopstart = sc->loopstart / stepscale;
	sc->speed = dma.speed;

	if (s_loadas8bit->value)
		sc->width = 1;
	else
		sc->width = inwidth;
	sc->stereo = 0;

// resample / decimate to the current source rate

	if (stepscale == 1 && inwidth == 1 && sc->width == 1)
	{
// fast special case
		for (i=0 ; i<outcount ; i++)
			((signed char *)sc->data)[i]
			= (int)( (unsigned char)(data[i]) - 128);
	}
	else
	{
// general case
		samplefrac = 0;
		fracstep = stepscale*256;
		for (i=0 ; i<outcount ; i++)
		{
			srcsample = samplefrac >> 8;
			samplefrac += fracstep;
			if (inwidth == 2)
				sample = LittleShort ( ((short *)data)[srcsample] );
			else
				sample = (int)( (unsigned char)(data[srcsample]) - 128) << 8;
			if (sc->width == 2)
				((short *)sc->data)[i] = sample;
			else
				((signed char *)sc->data)[i] = sample >> 8;
		}
	}
#endif
}

//=============================================================================
#ifdef SIN
/*
==============
S_LoadRaw
==============
*/
sfxcache_t *S_LoadRaw( int khz, int bits, int channels, int length, byte * data )
   {
   sfx_t       sfx;
	wavinfo_t	info;
	int		   len;
   int         real_length;
	sfxcache_t	*sc;

   info.loopstart = -1;
   info.rate = khz;
   info.blockalign = 0;
   info.format = WAV_PCM_FORMAT;
   info.channels = channels;
   info.width = bits / 8;
   info.samples = length;

   len = info.samples;
   info.samples /= info.width;
   real_length = info.samples;
   info.samples /= info.channels;

	sc = (sfxcache_t *)AllocClear(len + sizeof(sfxcache_t));
	if (!sc) {
		return NULL;
	} 
	
	sc->length = info.samples;
	sc->loopstart = info.loopstart;
	sc->speed = info.rate;
	sc->width = info.width;
	sc->stereo = info.channels-1;
	sc->type = info.format;
	sc->blockalign = info.blockalign;
	sc->real_length = real_length;
	
	sfx.cache = sc;
	ResampleSfx ( &sfx, sc->speed, sc->width, data );
	
	return sc;
   }


/*
==============
S_FreeSound
==============
*/
void S_FreeSound (sfx_t *s)
{
	Free( s->cache );
	s->cache = NULL;
}
#endif

/*
==============
S_LoadSound
==============
*/
static sfxcache_t *S_LoadSoundPath (sfx_t *s
#ifdef SIN
, char * namebuffer
#endif
)
{
#ifndef SIN
    char	namebuffer[MAX_QPATH];
#endif
	byte	*data;
	wavinfo_t	info;
	int		len;
#ifdef SIN
   int      real_length;
#else
	float	stepscale;
#endif
	sfxcache_t	*sc;
	int		size;

// see if still in memory
	sc = s->cache;
	if (sc)
		return sc;

//Com_Printf ("S_LoadSound: %x\n", (int)stackbuf);
// load it in
#ifndef SIN
	Com_sprintf (namebuffer, sizeof(namebuffer), "%s/%s", s_sounddir->string, s->name);
#endif

//	Com_Printf ("loading %s\n",namebuffer);

	size = FS_LoadFile (namebuffer, (void **)&data);

	if (!data)
	{
		Com_DPrintf ("Couldn't load %s\n", namebuffer);
		return NULL;
	}

	info = GetWavinfo (s->name, data, size);

//#ifdef SIN
//	if (info.channels != 1 && info.format != WAV_IMA_ADPCM_FORMAT)
//#else
#ifndef SIN
	if (info.channels != 1)
	{
		Com_Printf ("%s is a stereo sample\n",s->name);
		FS_FreeFile (data);
		return NULL;
	}
#else
   if ( !info.format || !info.channels )
      {
      return NULL;
      }
#endif

#ifdef SIN
	if (info.format == WAV_PCM_FORMAT)
      {
      len = info.samples;
      info.samples /= info.width;
      real_length = info.samples;
      info.samples /= info.channels;
      }
   else if (info.format == WAV_IMA_ADPCM_FORMAT)
      {
      int numblocks;

      len = info.samples;
      real_length = info.samples;
      // determine number of blocks
      numblocks = info.samples/info.blockalign;
      // subtrack block headers from real_length
      info.samples -= numblocks*info.channels*sizeof(unsigned);
      info.samples *= 2/info.channels;
      }
#else
	stepscale = (float)info.rate / dma.speed;	
	len = info.samples / stepscale;
	len = len * info.width * info.channels;
#endif

	sc = s->cache = (sfxcache_t*)Alloc(len + sizeof(sfxcache_t));
	if (!sc) {
		FS_FreeFile (data);
		return NULL;
	}
	
	sc->length = info.samples;
	sc->loopstart = info.loopstart;
	sc->speed = info.rate;
	sc->width = info.width;
#ifdef SIN
	sc->stereo = info.channels-1;
	sc->type = info.format;
	sc->blockalign = info.blockalign;
	sc->real_length = real_length;
#else
	sc->stereo = info.channels;
#endif

	ResampleSfx (s, sc->speed, sc->width, data + info.dataofs);

	FS_FreeFile (data);

	return sc;
}

#ifdef SIN
sfxcache_t *S_LoadSound (sfx_t *s)
   {  
   char	namebuffer[MAX_QPATH];

// see if still in memory
	if ( s->cache )
		return s->cache;
	Com_sprintf (namebuffer, sizeof(namebuffer), "%s/%s", s_sounddir->CVarString, s->name);
   S_LoadSoundPath( s, namebuffer );
   return s->cache;
   }

sfxcache_t *S_LoadSong (sfx_t *s)
   {  
   char	namebuffer[MAX_QPATH];

// see if still in memory
	if ( s->cache )
		return s->cache;
	Com_sprintf (namebuffer, sizeof(namebuffer), "%s/%s", s_musicdir->CVarString, s->name);
   S_LoadSoundPath( s, namebuffer );
   return s->cache;
   }
#endif

