//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/sv_world.c                                             $
// $Revision:: 4                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 8/8/01 4:49 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
// world query functions
//

#include "server.h"
#ifdef SIN
#include "sinmdl.h"
#include "linklist.h"
#endif

/*
===============================================================================

ENTITY AREA CHECKING

FIXME: this use of "area" is different from the bsp file use
===============================================================================
*/

// (type *)STRUCT_FROM_LINK(link_t *link, type, member)
// ent = STRUCT_FROM_LINK(link,entity_t,order)
// FIXME: remove this mess!
#define	STRUCT_FROM_LINK(l,t,m) ((t *)((byte *)l - (int)&(((t *)0)->m)))

#define	EDICT_FROM_AREA(l) STRUCT_FROM_LINK(l,edict_t,area)

typedef struct areanode_s
{
	int		axis;		// -1 = leaf node
	float	dist;
	struct areanode_s	*children[2];
	link_t	trigger_edicts;
	link_t	solid_edicts;
} areanode_t;

#define	AREA_DEPTH	4
#define	AREA_NODES	32

areanode_t	sv_areanodes[AREA_NODES];
int			sv_numareanodes;

float	*area_mins, *area_maxs;
edict_t	**area_list;
int		area_count, area_maxcount;
int		area_type;

int SV_HullForEntity (edict_t *ent);


// ClearLink is used for new headnodes
static void ClearLink (link_t *l)
{
	l->prev = l->next = l;
}

static void RemoveLink (link_t *l)
{
	l->next->prev = l->prev;
	l->prev->next = l->next;
}

static void InsertLinkBefore (link_t *l, link_t *before)
{
	l->next = before;
	l->prev = before->prev;
	l->prev->next = l;
	l->next->prev = l;
}

/*
===============
SV_CreateAreaNode

Builds a uniformly subdivided tree for the given world size
===============
*/
static areanode_t *SV_CreateAreaNode (int depth, vec3_t mins, vec3_t maxs)
{
	areanode_t	*anode;
	vec3_t		size;
	vec3_t		mins1, maxs1, mins2, maxs2;

	anode = &sv_areanodes[sv_numareanodes];
	sv_numareanodes++;

	ClearLink (&anode->trigger_edicts);
	ClearLink (&anode->solid_edicts);

	if (depth == AREA_DEPTH)
	{
		anode->axis = -1;
		anode->children[0] = anode->children[1] = NULL;
		return anode;
	}

	VectorSubtract (maxs, mins, size);
	if (size[0] > size[1])
		anode->axis = 0;
	else
		anode->axis = 1;

	anode->dist = 0.5 * (maxs[anode->axis] + mins[anode->axis]);
	VectorCopy (mins, mins1);
	VectorCopy (mins, mins2);
	VectorCopy (maxs, maxs1);
	VectorCopy (maxs, maxs2);

	maxs1[anode->axis] = mins2[anode->axis] = anode->dist;

	anode->children[0] = SV_CreateAreaNode (depth+1, mins2, maxs2);
	anode->children[1] = SV_CreateAreaNode (depth+1, mins1, maxs1);

	return anode;
}

/*
===============
SV_ClearWorld

===============
*/
void SV_ClearWorld (void)
{
	FastMemSet (sv_areanodes, 0, sizeof(sv_areanodes));
	sv_numareanodes = 0;
	SV_CreateAreaNode (0, sv.models[1]->mins, sv.models[1]->maxs);
}


/*
===============
SV_UnlinkEdict

===============
*/
void SV_UnlinkEdict (edict_t *ent)
{
	if (!ent->area.prev)
		return;		// not linked in anywhere
	RemoveLink (&ent->area);
	ent->area.prev = ent->area.next = NULL;
}


/*
===============
SV_LinkEdict

===============
*/
#define MAX_TOTAL_ENT_LEAFS		128
void SV_LinkEdict (edict_t *ent)
{
	areanode_t	*node;
	int			leafs[MAX_TOTAL_ENT_LEAFS];
	int			clusters[MAX_TOTAL_ENT_LEAFS];
	int			num_leafs;
#ifdef SIN
	int			x, y, i, j, k;
#else
	int			i, j, k;
#endif
	int			area;
	int			topnode;

	if (ent->area.prev)
		SV_UnlinkEdict (ent);	// unlink from old position

	if (ent == ge->edicts)
		return;		// don't add the world

	if (!ent->inuse)
		return;

	// set the size
	VectorSubtract (ent->maxs, ent->mins, ent->size);

	// encode the size into the entity_state for client prediction
#ifdef SIN
	if (ent->solid == SOLID_BBOX && !(ent->svflags & SVF_DEADMONSTER) && !(ent->svflags & SVF_SHOOTABLE) )
	{	
		gravityaxis_t grav;

		grav = gravity_axis[ EFFECTS_TO_GRAVITYAXIS( ent->s.effects ) ];
		// x radius
		x = ent->maxs[ grav.x ];
		if (x<1)
			x = 1;
		if (x>255)
			x = 255;

		// y radius
		y = ent->maxs[ grav.y ];
		if (y<1)
			y = 1;
		if (y>255)
			y = 255;

		// z is not symetric
		j = ((ent->mins[ grav.z ] * grav.sign)+128)/8;
		if (j<1)
			j = 1;
		if (j>63)
			j = 63;

		// and z maxs can be negative...
		k = ((ent->maxs[ grav.z ] * grav.sign)+64);
		if (k<1)
			k = 1;
		if (k>511)
			k = 511;

		ent->s.solid = (k<<22) | (j<<16) | (y<<8) | x;
	}
#else
	if (ent->solid == SOLID_BBOX && !(ent->svflags & SVF_DEADMONSTER) )
	{	// assume that x/y are equal and symetric
		i = ent->maxs[0]/8;
		if (i<1)
			i = 1;
		if (i>31)
			i = 31;

		// z is not symetric
		j = (-ent->mins[2])/8;
		if (j<1)
			j = 1;
		if (j>31)
			j = 31;

		// and z maxs can be negative...
		k = (ent->maxs[2]+32)/8;
		if (k<1)
			k = 1;
		if (k>63)
			k = 63;

		ent->s.solid = (k<<10) | (j<<5) | i;
	}
#endif
	else if (ent->solid == SOLID_BSP)
	{
		ent->s.solid = 31;		// a solid_bbox will never create this value
	}
	else
		ent->s.solid = 0;

	// set the abs box
	if (ent->solid == SOLID_BSP &&
	(ent->s.angles[0] || ent->s.angles[1] || ent->s.angles[2]) )
	{	// expand for rotation
#ifdef SIN
      VectorCopy( ent->mins, ent->absmin );
      VectorCopy( ent->maxs, ent->absmax );
      CalculateRotatedBounds2( ent->s.mat, ent->absmin, ent->absmax );
      VectorAdd( ent->absmin, ent->s.origin, ent->absmin );
      VectorAdd( ent->absmax, ent->s.origin, ent->absmax );
#else
		float		max, v;
		int			i;

		max = 0;
		for (i=0 ; i<3 ; i++)
		{
			v =Burger::FPMath::Abs( ent->mins[i]);
			if (v > max)
				max = v;
			v =Burger::FPMath::Abs( ent->maxs[i]);
			if (v > max)
				max = v;
		}
		for (i=0 ; i<3 ; i++)
		{
			ent->absmin[i] = ent->s.origin[i] - max;
			ent->absmax[i] = ent->s.origin[i] + max;
		}
#endif
	}
	else
	{	// normal
		VectorAdd (ent->s.origin, ent->mins, ent->absmin);
		VectorAdd (ent->s.origin, ent->maxs, ent->absmax);
	}

	// because movement is clipped an epsilon away from an actual edge,
	// we must fully check even when bounding boxes don't quite touch
	ent->absmin[0] -= 1;
	ent->absmin[1] -= 1;
	ent->absmin[2] -= 1;
	ent->absmax[0] += 1;
	ent->absmax[1] += 1;
	ent->absmax[2] += 1;

// link to PVS leafs
	ent->num_clusters = 0;
	ent->areanum = 0;
	ent->areanum2 = 0;

	//get all leafs, including solids
	num_leafs = CM_BoxLeafnums (ent->absmin, ent->absmax,
		leafs, MAX_TOTAL_ENT_LEAFS, &topnode);

	// set areas
	for (i=0 ; i<num_leafs ; i++)
	{
		clusters[i] = CM_LeafCluster (leafs[i]);
		area = CM_LeafArea (leafs[i]);
		if (area)
		{	// doors may legally straggle two areas,
			// but nothing should evern need more than that
			if (ent->areanum && ent->areanum != area)
			{
				if (ent->areanum2 && ent->areanum2 != area && sv.state == ss_loading)
					Com_DPrintf ("Object touching 3 areas at %f %f %f\n",
					ent->absmin[0], ent->absmin[1], ent->absmin[2]);
				ent->areanum2 = area;
			}
			else
				ent->areanum = area;
		}
	}

	if (num_leafs >= MAX_TOTAL_ENT_LEAFS)
	{	// assume we missed some leafs, and mark by headnode
		ent->num_clusters = -1;
		ent->headnode = topnode;
	}
	else
	{
		ent->num_clusters = 0;
		for (i=0 ; i<num_leafs ; i++)
		{
			if (clusters[i] == -1)
				continue;		// not a visible leaf
			for (j=0 ; j<i ; j++)
				if (clusters[j] == clusters[i])
					break;
			if (j == i)
			{
				if (ent->num_clusters == MAX_ENT_CLUSTERS)
				{	// assume we missed some leafs, and mark by headnode
					ent->num_clusters = -1;
					ent->headnode = topnode;
					break;
				}

				ent->clusternums[ent->num_clusters++] = clusters[i];
			}
		}
	}

	// if first time, make sure old_origin is valid
	if (!ent->linkcount)
	{
		VectorCopy (ent->s.origin, ent->s.old_origin);
	}
	ent->linkcount++;

	if (ent->solid == SOLID_NOT)
		return;

// find the first node that the ent's box crosses
	node = sv_areanodes;
	while (1)
	{
		if (node->axis == -1)
			break;
		if (ent->absmin[node->axis] > node->dist)
			node = node->children[0];
		else if (ent->absmax[node->axis] < node->dist)
			node = node->children[1];
		else
			break;		// crosses the node
	}

	// link it in
	if (ent->solid == SOLID_TRIGGER)
		InsertLinkBefore (&ent->area, &node->trigger_edicts);
	else
		InsertLinkBefore (&ent->area, &node->solid_edicts);

}


/*
====================
SV_AreaEdicts_r

====================
*/
static void SV_AreaEdicts_r (areanode_t *node)
{
	link_t		*l, *next, *start;
	edict_t		*check;
	int			count;

	count = 0;

	// touch linked edicts
	if (area_type == AREA_SOLID)
		start = &node->solid_edicts;
	else
		start = &node->trigger_edicts;

	for (l=start->next  ; l != start ; l = next)
	{
		next = l->next;
		check = EDICT_FROM_AREA(l);

		if (check->solid == SOLID_NOT)
			continue;		// deactivated
		if (check->absmin[0] > area_maxs[0]
		|| check->absmin[1] > area_maxs[1]
		|| check->absmin[2] > area_maxs[2]
		|| check->absmax[0] < area_mins[0]
		|| check->absmax[1] < area_mins[1]
		|| check->absmax[2] < area_mins[2])
			continue;		// not touching

		if (area_count == area_maxcount)
		{
         Com_Printf ("SV_AreaEdicts: MAXCOUNT\n");
			return;
		}

		area_list[area_count] = check;
		area_count++;
	}

	if (node->axis == -1)
		return;		// terminal node

	// recurse down both sides
	if ( area_maxs[node->axis] > node->dist )
		SV_AreaEdicts_r ( node->children[0] );
	if ( area_mins[node->axis] < node->dist )
		SV_AreaEdicts_r ( node->children[1] );
}

/*
================
SV_AreaEdicts
================
*/
int SV_AreaEdicts (vec3_t mins, vec3_t maxs, edict_t **list,
	int maxcount, int areatype)
{
	area_mins = mins;
	area_maxs = maxs;
	area_list = list;
	area_count = 0;
	area_maxcount = maxcount;
	area_type = areatype;

	SV_AreaEdicts_r (sv_areanodes);

	return area_count;
}


//===========================================================================

/*
=============
SV_PointContents
=============
*/
int SV_PointContents (vec3_t p)
{
	edict_t		*touch[MAX_EDICTS], *hit;
	int			i, num;
	int			contents, c2;
	int			headnode;
	const float		*angles;

	// get base contents from world
	contents = CM_PointContents (p, sv.models[1]->headnode);

	// or in contents from all the other entities
	num = SV_AreaEdicts (p, p, touch, MAX_EDICTS, AREA_SOLID);

	for (i=0 ; i<num ; i++)
	{
		hit = touch[i];

		// might intersect, so do an exact clip
		headnode = SV_HullForEntity (hit);
		angles = hit->s.angles;
		if (hit->solid != SOLID_BSP)
			angles = vec3_origin;	// boxes don't rotate

		c2 = CM_TransformedPointContents (p, headnode, hit->s.origin, hit->s.angles);

		contents |= c2;
	}

	return contents;
}



typedef struct
{
	vec3_t boxmins, boxmaxs;// enclose the test object along entire move
	const float *mins, *maxs;	// size of the moving object
	vec3_t mins2, maxs2;	// size when clipping against mosnters
	const float *start;
	const float *end;
	trace_t trace;
	edict_t *passedict;
	int contentmask;
} moveclip_t;

#ifdef SIN
typedef struct clipent_s {
	edict_t * ent;
	trace_t trace;
	struct clipent_s *prev;
	struct clipent_s *next;
} clipent_t;

static clipent_t *clip_freelist = NULL;
static clipent_t *clip_candidatelist = NULL;
static qboolean	clip_fullpolytrace = FALSE;
static float   	clip_fullpolyradius;

void SV_ClipRelease(void)
{
	clipent_t *WorkPtr;
	clipent_t *Next;
	clipent_t *Root;
	WorkPtr = clip_freelist;
	if (WorkPtr) {
		clip_freelist = 0;
		Root = WorkPtr;
		do {
			Next = WorkPtr->next;
			Free(WorkPtr);
			WorkPtr = Next;
		} while (WorkPtr!=Root);
	}
	WorkPtr = clip_candidatelist;
	if (WorkPtr) {
		clip_candidatelist = 0;
		Root = WorkPtr;
		do {
			Next = WorkPtr->next;
			Free(WorkPtr);
			WorkPtr = Next;
		} while (WorkPtr!=Root);
	}
}

/*
====================
SV_FreeClipCandidates
====================
*/
static void SV_FreeClipCandidates( void )
   {
   clipent_t *next, *clip;

   if ( !clip_candidatelist )
      return;
   if ( !clip_freelist )
      {
      LL_New( clip_freelist, clipent_t, next, prev );
      }
   for ( clip = clip_candidatelist->next; clip != clip_candidatelist; clip = next )
      {
      next = clip->next;
      LL_Remove( clip, next, prev );
      LL_Add( clip_freelist, clip, next, prev );
      }
   }

/*
====================
SV_AddClipCandidate
====================
*/
static void SV_AddClipCandidate( edict_t * ent, trace_t * trace )
{
	clipent_t * clip;
	
	if ( !clip_candidatelist ) {
		LL_New( clip_candidatelist, clipent_t, next, prev );
	}
	if ( clip_freelist && clip_freelist->next != clip_freelist ) {
		clip = clip_freelist->next;
		LL_Remove( clip, next, prev );
	} else {
		LL_New( clip, clipent_t, next, prev );
	}
	clip->ent = ent;
	clip->trace = *trace;
	LL_SortedInsertion( clip_candidatelist, clip, next, prev, clipent_t, trace.fraction );
}

#endif


/*
================
SV_HullForEntity

Returns a headnode that can be used for testing or clipping an
object of mins/maxs size.
Offset is filled in to contain the adjustment that must be added to the
testing object's origin to get a point to use with the returned hull.
================
*/
int SV_HullForEntity (edict_t *ent)
{
	cmodel_t	*model;

// decide which clipping hull to use, based on the size
	if (ent->solid == SOLID_BSP)
	{	// explicit hulls in the BSP model
		model = sv.models[ ent->s.modelindex ];

#ifdef SIN
		if (!model)
			Com_Error (ERR_FATAL, "SOLID_BSP with a non bsp model");
#else
		if (!model)
			Com_Error (ERR_FATAL, "MOVETYPE_PUSH with a non bsp model");
#endif

		return model->headnode;
	}

	// create a temp hull from bounding box sizes

#ifdef SIN
   if ( clip_fullpolytrace )
      {
	   return CM_HeadnodeForBox (ent->fullmins, ent->fullmaxs);
      }
   else
      {
	   return CM_HeadnodeForBox (ent->mins, ent->maxs);
      }
#else
	return CM_HeadnodeForBox (ent->mins, ent->maxs);
#endif
}


//===========================================================================


/*
====================
SV_ClipMoveToEntities

====================
*/
static void SV_ClipMoveToEntities ( moveclip_t *clip )
{
	int			i, num;
	edict_t		*touchlist[MAX_EDICTS], *touch;
	trace_t		trace;
	int			headnode;
	const vec3_t		*angles;

	num = SV_AreaEdicts (clip->boxmins, clip->boxmaxs, touchlist
		, MAX_EDICTS, AREA_SOLID);

	// be careful, it is possible to have an entity in this
	// list removed before we get to it (killtriggered)
	for (i=0 ; i<num ; i++)
	{
		touch = touchlist[i];
		if (touch->solid == SOLID_NOT)
			continue;
		if (touch == clip->passedict)
			continue;
		if (clip->trace.allsolid)
			return;
		if (clip->passedict)
		{
		 	if (touch->owner == clip->passedict)
				continue;	// don't clip against own missiles
			if (clip->passedict->owner == touch)
				continue;	// don't clip against owner
		}

		if ( !(clip->contentmask & CONTENTS_DEADMONSTER)
		&& (touch->svflags & SVF_DEADMONSTER) )
				continue;
#ifdef SIN
		if ( !(clip->contentmask & CONTENTS_SHOOTABLE)
		&& (touch->svflags & SVF_SHOOTABLE) )
				continue;

      if (touch->svflags & (SVF_MONSTERCLIP|SVF_PLAYERCLIP) )
         {
         int hit;

   		if ( !( clip->contentmask & (CONTENTS_MONSTERCLIP|CONTENTS_PLAYERCLIP) ) )
            {
            continue;
            }

         hit = 0;
         if ( touch->svflags & SVF_MONSTERCLIP)
            {
            hit++;
            }

         if ( touch->svflags & SVF_PLAYERCLIP )
            {
            hit++;
            }

         if ( (touch->svflags & SVF_MONSTERCLIP) && !(clip->contentmask & CONTENTS_MONSTERCLIP) )
            {
            hit--;
            }

         if ( (touch->svflags & SVF_PLAYERCLIP) && !(clip->contentmask & CONTENTS_PLAYERCLIP) )
            {
            hit--;
            }

         if ( !hit )
            {
            continue;
            }
         }
#endif

		// might intersect, so do an exact clip
		headnode = SV_HullForEntity (touch);
		angles = &touch->s.angles;
		if (touch->solid != SOLID_BSP)
			angles = &vec3_origin;	// boxes don't rotate

#ifdef SIN
      if ( clip_fullpolytrace && touch->solid != SOLID_BSP )
         {
         float length;
         vec3_t delta;

         VectorSubtract( touch->s.origin, clip->start, delta );
         length = VectorLength( delta );
         if ( length < touch->fullradius )
            {
            trace.fraction = 0;
            trace.surface = NULL;
            trace.startsolid = FALSE;
            trace.allsolid = FALSE;
			   trace.ent = touch;
            trace.intersect.valid = FALSE;
            VectorCopy( clip->trace.dir, trace.dir );
            SV_AddClipCandidate( touch, &trace );
            continue;
            }
         }
#endif
		if (touch->svflags & SVF_MONSTER)
			trace = CM_TransformedBoxTrace (clip->start, clip->end,
				clip->mins2, clip->maxs2, headnode, clip->contentmask,
				touch->s.origin, angles);
		else
			trace = CM_TransformedBoxTrace (clip->start, clip->end,
				clip->mins, clip->maxs, headnode,  clip->contentmask,
				touch->s.origin, angles);

		if (trace.allsolid || trace.startsolid ||
		trace.fraction < clip->trace.fraction)
		{
#ifdef SIN
		if ( clip_fullpolytrace && touch )
            {
            trace.intersect.valid = FALSE;
			   trace.ent = touch;
            VectorCopy( clip->trace.dir, trace.dir );
            SV_AddClipCandidate( touch, &trace );
            }
         else
            {
            trace.intersect.valid = FALSE;
			   trace.ent = touch;
		 	   if (clip->trace.startsolid)
			      {
				   clip->trace = trace;
				   clip->trace.startsolid = TRUE;
			      }
			   else
				   clip->trace = trace;
            }
#else
			trace.ent = touch;
		 	if (clip->trace.startsolid)
			{
				clip->trace = trace;
				clip->trace.startsolid = TRUE;
			}
			else
				clip->trace = trace;
#endif
		}
		else if (trace.startsolid)
			clip->trace.startsolid = TRUE;
	}
#ifdef SIN
	if ( clip_fullpolytrace && clip_candidatelist )
      {
      int defhandle;
      sinmdl_intersection_t *intersect;
      clipent_t *clipent;

      for ( clipent = clip_candidatelist->next; clipent != clip_candidatelist; clipent = clipent->next )
         {
         touch = clipent->ent;

         //
         // this shouldn't happen but just make sure
         //
         if ( !touch )
            continue;

         defhandle = SINMDL_Handle_For_ModelIndex( touch->s.modelindex );
         if ( !( touch->svflags & SVF_USEBBOX ) && ( defhandle >= 0 ) )
            {
            intersect = SINMDL_RayIntersect
               (
               defhandle,
               touch->s.anim,
               touch->s.frame,
               touch->s.scale,
               touch->s.origin,
               touch->s.mat,
               clip->start,
               clip->end,
               clip_fullpolyradius
               );
            if (intersect)
               {
               float len;
               float fulllen;
               float fraction;
               vec3_t delta;
               vec3_t fulldelta;

               VectorSubtract( intersect->position, clip->start, delta );
               VectorSubtract( clip->end, clip->start, fulldelta );
               fulllen = VectorLength( fulldelta );
               len = VectorLength( delta );
               fraction = len / fulllen;
               if ( fraction < clip->trace.fraction )
                  {
                  clipent_t *next;
                  next = clipent->next;
                  // if the next entity in the list is a B-model
                  // make sure that our fraction is in deed closer 
                  // than it
                  if ( 
                        ( next != clip_candidatelist ) && 
                        ( next->ent->solid == SOLID_BSP ) && 
                        ( fraction > next->trace.fraction )
                     )
                     continue;
                  clip->trace = clipent->trace;
                  clip->trace.intersect = *intersect;
                  VectorCopy( intersect->normal, clip->trace.plane.normal );
                  VectorMA( clip->start, len, clip->trace.dir, clip->trace.endpos );
                  break;
                  }
               }
            }
         else
            {
            clip->trace = clipent->trace;
            break;
            }
         }
      SV_FreeClipCandidates();
      }
#endif
}


/*
==================
SV_TraceBounds
==================
*/
static void SV_TraceBounds (const vec3_t start,const vec3_t mins,const vec3_t maxs,const vec3_t end, vec3_t boxmins, vec3_t boxmaxs)
{
#if 0
// debug to test against everything
boxmins[0] = boxmins[1] = boxmins[2] = -9999;
boxmaxs[0] = boxmaxs[1] = boxmaxs[2] = 9999;
#else
	int		i;

	for (i=0 ; i<3 ; i++)
	{
		if (end[i] > start[i])
		{
			boxmins[i] = start[i] + mins[i] - 1;
			boxmaxs[i] = end[i] + maxs[i] + 1;
		}
		else
		{
			boxmins[i] = end[i] + mins[i] - 1;
			boxmaxs[i] = start[i] + maxs[i] + 1;
		}
	}
#endif
}

/*
==================
SV_Trace

Moves the given mins/maxs volume through the world from start to end.

Passedict and edicts owned by passedict are explicitly not checked.

==================
*/
trace_t SV_Trace (const vec3_t start,const vec3_t mins,const vec3_t maxs,const vec3_t end, edict_t *passedict, int contentmask)
{
	moveclip_t	clip;

#ifdef SIN
	clip_fullpolytrace = FALSE;
#endif

	if (!mins)
		mins = vec3_origin;
	if (!maxs)
		maxs = vec3_origin;
	
	FastMemSet( &clip, 0, sizeof ( moveclip_t ) );
	
	// clip to world
	clip.trace = CM_BoxTrace (start, end, mins, maxs, 0, contentmask);
	clip.trace.intersect.valid = FALSE;
	clip.trace.ent = ge->edicts;
	if (clip.trace.fraction == 0)
		return clip.trace;		// blocked by the world
	
	clip.contentmask = contentmask;
	clip.start = start;
	clip.end = end;
	clip.mins = mins;
	clip.maxs = maxs;
	clip.passedict = passedict;
	
	VectorCopy (mins, clip.mins2);
	VectorCopy (maxs, clip.maxs2);
	
	// create the bounding box of the entire move
	SV_TraceBounds ( start, clip.mins2, clip.maxs2, end, clip.boxmins, clip.boxmaxs );
	
	// clip to other solid entities
	SV_ClipMoveToEntities ( &clip );
	
	return clip.trace;
}

#ifdef SIN

trace_t SV_FullTrace (const vec3_t start,const vec3_t mins,const vec3_t maxs,const vec3_t end, float radius, edict_t *passedict, int contentmask)
	{
	moveclip_t	clip;

	clip_fullpolytrace = TRUE;
	clip_fullpolyradius = radius;

	if (!mins)
		mins = vec3_origin;
	if (!maxs)
		maxs = vec3_origin;

	FastMemSet( &clip, 0, sizeof ( moveclip_t ) );

	// clip to world
	clip.trace = CM_BoxTrace (start, end, mins, maxs, 0, contentmask);
   clip.trace.intersect.valid = FALSE;
	clip.trace.ent = ge->edicts;
	if (clip.trace.fraction == 0)
		{
		clip_fullpolytrace = FALSE;
		return clip.trace;		// blocked by the world
		}

	clip.contentmask = contentmask;
	clip.start = start;
	clip.end = end;
	clip.mins = mins;
	clip.maxs = maxs;
	clip.passedict = passedict;

	VectorCopy (mins, clip.mins2);
	VectorCopy (maxs, clip.maxs2);

	// create the bounding box of the entire move
	SV_TraceBounds ( start, clip.mins2, clip.maxs2, end, clip.boxmins, clip.boxmaxs );

	// clip to other solid entities
	SV_ClipMoveToEntities ( &clip );

	clip_fullpolytrace = FALSE;

	return clip.trace;
	}

#endif