//-----------------------------------------------------------------------------
//
//  $Logfile:: /win32/cin_win.cpp                                             $
// $Revision:: 1                                                              $
//   $Author:: Burger                                                         $
//     $Date:: 12/27/00 7:56p                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// $Log:: /win32/cin_win.cpp                                                  $
 * 
 * 1     12/27/00 7:56p Burger
// 
// 21    8/07/99 1:53p Markd
// fixed a bunch of more code for version 1.06
// 
// 20    7/27/99 4:37p Markd
// moved check for file to top of loading function so if cinematics weren't
// present it wouldn't do the flash thing
// 
// 19    11/17/98 8:26p Markd
// Changed s_register functions so that they only do status when a passed in
// boolean is true, made the cinematic do a sound restart rather than a S_Init
// 
// 18    10/27/98 10:35p Markd
// fixed cinematics AGAIN!
// 
// 17    10/27/98 10:35p Markd
// fixed fullscreen problems
// 
// 16    10/26/98 3:39p Markd
// put in cl_forcecinematic
// 
// 15    10/26/98 2:13a Markd
// fixed mode switching twice
// 
// 14    10/26/98 12:51a Markd
// fixed a bug with finish cinematic
// 
// 13    10/25/98 9:36p Markd
// made cinematics switch to software 16-bit 320x240 fullscreen by default
// 
// 12    10/25/98 12:19a Aldie
// Adjusted some things for cinematics to work with menus
// 
// 11    10/24/98 3:50a Markd
// fixed cinematics
// 
// 10    10/23/98 12:22a Aldie
// Added finish cinematic to goto nextserver
// 
// 9     10/21/98 6:40p Markd
// made cinematics use relative path names
// 
// 8     10/19/98 12:15a Markd
// tweaked some Dshow stuff
// 
// 7     9/28/98 5:28p Markd
// Put error messages into cinematic stuff
// 
// 6     9/28/98 4:15p Markd
// took out the SetFormat command
// 
// 5     9/27/98 5:10p Markd
// Rewrote a lot of the code, now works in current window hardware and software
// 
// 4     9/27/98 3:30p Markd
// intermediate check-in
// 
// 3     9/27/98 3:19p Markd
// intermediat check in of new cinematic code
// 
// 2     6/06/98 5:38p Markd
// Added windows cinematic stuff
// 
// 1     6/06/98 2:14p Markd
// 
// 1     6/06/98 1:17p Markd
// 
// DESCRIPTION:
// Directshow player
//

#include <windows.h>
#include "ddraw.h"
#include "mmstream.h"
#include "amstream.h"
#include "ddstream.h"
#include "errors.h"
#include <assert.h>
#include "..\client\client.h"
#include "winquake.h"

static IDirectDraw *pDD = NULL;
static IMediaStream *pPrimaryVidStream = NULL;
static IMediaStream *pPrimaryAudStream = NULL;
static IDirectDrawMediaStream *pDDStream = NULL;
static IDirectDrawSurface *pSurface = NULL;
static IDirectDrawStreamSample *pSample = NULL;
static IDirectDrawSurface *pPrimarySurface = NULL;
static DDSURFACEDESC movie_ddsd;
static IMultiMediaStream *pMMStream = NULL;
static int vid_modified = 0;
static int swmode_modified = 0;
static int ref_modified = 0;
static int fullscreen_modified = 0;

extern "C"
{
HWND cl_hwnd;
qboolean cinematic_active = false;
qboolean cinematic_paused = false;

#define RELEASE(x) if (x) (x)->Release();

HRESULT OpenMMStream(const char * pszFileName, IDirectDraw *pDD)
   {
   IAMMultiMediaStream *pAMStream;
   HRESULT hr;

   pMMStream = NULL;
   hr = CoCreateInstance(CLSID_AMMultiMediaStream, NULL, CLSCTX_INPROC_SERVER, IID_IAMMultiMediaStream, (void **)&pAMStream);
   if (SUCCEEDED(hr)) 
      {
      hr = pAMStream->Initialize(STREAMTYPE_READ, AMMSF_NOGRAPHTHREAD, NULL);
      if (SUCCEEDED(hr)) 
         {
         hr = pAMStream->AddMediaStream(pDD, &MSPID_PrimaryVideo, 0, NULL);
         if (SUCCEEDED(hr)) 
            {
            hr = pAMStream->AddMediaStream(NULL, &MSPID_PrimaryAudio, AMMSF_ADDDEFAULTRENDERER, NULL);
            if (SUCCEEDED(hr)) 
               {
               WCHAR       wPath[MAX_PATH];
               MultiByteToWideChar(CP_ACP, 0, pszFileName, -1, wPath, sizeof(wPath)/sizeof(wPath[0]));

               hr = pAMStream->OpenFile(wPath, 0);
               if (SUCCEEDED(hr)) 
                  {
                  pMMStream = pAMStream;
                  pAMStream->AddRef();
                  }
               }
            }
         }
      }
    if ( !SUCCEEDED( hr ) )
       {
       char buf[128];
       AMGetErrorText( hr, buf, sizeof( buf )-1 );
       Com_Printf("OpenMMStream Error: %s\n", buf );
       }
    RELEASE(pAMStream);
    return hr;
    }

/*
==================
SCR_StopCinematic
==================
*/
void SCR_StopCinematic (void)
   {
   if ( cl.cinematictime )
      {
	   cl.cinematictime = 0;	// done
      re.EndCinematic();
      RELEASE(pMMStream);
      RELEASE(pPrimaryAudStream);
      RELEASE(pPrimaryVidStream);
      RELEASE(pDDStream);
      RELEASE(pSample);
      RELEASE(pSurface);
      RELEASE(pDD);
      CoUninitialize();
      SCR_FinishCinematic();
      }

   if ( !sound_started )
		CL_Snd_Restart_f ();
   if ( cl_forcecinematic->value )
      {
      //
      // restore our settings
      //
      if ( ref_modified )
         {
         ref_modified = 0;
         Cvar_Set( "vid_ref", "gl" );
         vid_ref->modified = true;
         }
      if ( vid_modified )
         {
         Cvar_SetValue( "gl_ext_palettedtexture", vid_modified );
         vid_modified = 0;
         vid_ref->modified = true;
         }
      if ( swmode_modified )
         {
         Cvar_SetValue( "sw_mode", swmode_modified );
         swmode_modified = 0;
         vid_ref->modified = true;
         }
      if ( fullscreen_modified )
         {
         Cvar_SetValue( "vid_fullscreen", 0 );
         fullscreen_modified = 0;
         vid_ref->modified = true;
         }
      }
   }

/*
====================
SCR_FinishCinematic

Called when either the cinematic completes, or it is aborted
====================
*/
void SCR_FinishCinematic (void)
   {
   //
   // make sure we stop it
   //
   SCR_StopCinematic();

	// tell the server to advance to the next map / cinematic
	MSG_WriteByte (&cls.netchan.message, clc_stringcmd);
	SZ_Print (&cls.netchan.message, va("nextserver %i\n", cl.servercount));
   }

/*
==================
SCR_RunCinematic
decode next frame
==================
*/
void SCR_RunCinematic (void)
   {
   if (cl.cinematictime <= 0)
	   {
		return;
	   }

   if ( cl.cinematictime )
      {
      //
      // we render until their are no more updates
      // not true right now
      //
      if (pSample->Update(0, NULL, NULL, 0) != S_OK) 
         {
   		SCR_StopCinematic ();
	   	return;
	      }
      }
   }
   
/*
==================
SCR_DrawCinematic

Returns true if a cinematic is active, meaning the view rendering
should be skipped
==================
*/
qboolean SCR_DrawCinematic (void)
   {
   unsigned short * ptr;

	if (cl.cinematictime <= 0)
	   {
		return false;
	   }

	pSurface->Lock( NULL, &movie_ddsd, DDLOCK_WAIT, NULL );

	ptr = ( unsigned short * )movie_ddsd.lpSurface;
   re.DrawCinematicFrame(0, 0, viddef.width, viddef.height, movie_ddsd.dwWidth, movie_ddsd.dwHeight, movie_ddsd.lPitch, ptr);

   pSurface->Unlock( ptr );

	return true;
   }

/*
==================
SCR_PlayCinematic

==================
*/
void SCR_PlayCinematic ( const char *arg )
   {
   char name[ MAX_OSPATH ];
   HRESULT hr;
   FILE * fp;

   // 
   // try the cd first
   //
   if ( !strstr( arg, ":" ) )
      {
      Com_sprintf( name, sizeof( name ), "%s/movies/%s", FS_CDdir(), arg );
      }
   else
      {
      strncpy( name, arg, sizeof( name ) - 1 );
      }
	fp = fopen( name, "rb" );
   if ( !fp )
      {
      // 
      // try the hardrive next
      //
      Com_sprintf( name, sizeof( name ), "%s/movies/%s", FS_Gamedir(), arg );
   	fp = fopen( name, "rb" );
      if ( !fp )
         {
         // we didn't find it so lets return and not do the silly video switches
         SCR_FinishCinematic();
         return;
         }
      else
         {
         fclose( fp );
         }
      }
   else
      {
      fclose( fp );
      }

   if ( cl.cinematictime > 0 )
      SCR_StopCinematic();

   cl.cinematictime = 0;
   vid_modified = 0;
   ref_modified = 0;

   if ( cl_forcecinematic->value )
      {
      //
      // make sure we are in 16-bit software period
      //
      if ( !Burger::StringCaseCompare( Cvar_VariableString( "vid_ref" ), "soft" ) )
         {
         //
         // a value of 0 is 16-bit any thing else is an 8-bit variant 
         //
         vid_modified = Cvar_VariableValue( "gl_ext_palettedtexture" );
         swmode_modified = Cvar_VariableValue( "sw_mode" );
         fullscreen_modified = Cvar_VariableValue( "vid_fullscreen" );
         if ( (vid_modified) || (swmode_modified != 0) || (fullscreen_modified != 1) )
            {
            if ( !fullscreen_modified )
               fullscreen_modified = 1;
            else
               fullscreen_modified = 0;
            Cvar_SetValue( "gl_ext_palettedtexture", 0 );
            Cvar_SetValue( "sw_mode", 0 );
            Cvar_SetValue( "vid_fullscreen", 1 );
            vid_ref->modified = true;
            VID_CheckChanges();
            }
         else
            {
            fullscreen_modified = 0;
            }
         }
      //
      // else we are in hardware
      //
      else 
         {
         ref_modified = 1;
         vid_modified = Cvar_VariableValue( "gl_ext_palettedtexture" );
         swmode_modified = Cvar_VariableValue( "sw_mode" );
         fullscreen_modified = Cvar_VariableValue( "vid_fullscreen" );
         if ( (vid_modified) || (swmode_modified != 0) || (fullscreen_modified != 1) )
            {
            if ( !fullscreen_modified )
               fullscreen_modified = 1;
            else
               fullscreen_modified = 0;
            Cvar_Set( "vid_ref", "soft" );
            Cvar_SetValue( "gl_ext_palettedtexture", 0 );
            Cvar_SetValue( "sw_mode", 0 );
            Cvar_SetValue( "vid_fullscreen", 1 );
            vid_ref->modified = true;
            VID_CheckChanges();
            }
         else
            {
            fullscreen_modified = 0;
            }
         }
      }

   // make sure CD isn't playing music
	CDAudio_Stop();

   S_Shutdown();

   CoInitialize(NULL);

   hr = DirectDrawCreate(NULL, &pDD, NULL);
   if (SUCCEEDED(hr)) 
      {
      pDD->SetCooperativeLevel( cl_hwnd, DDSCL_NORMAL);
         {
         hr = OpenMMStream( name, pDD );
         if (SUCCEEDED(hr)) 
            {
            RECT rect;

            re.BeginCinematic();

	         memset( &movie_ddsd, 0, sizeof( movie_ddsd ) );
	         movie_ddsd.dwSize = sizeof( movie_ddsd );

            hr = pMMStream->GetMediaStream(MSPID_PrimaryAudio, &pPrimaryAudStream);
            hr = pMMStream->GetMediaStream(MSPID_PrimaryVideo, &pPrimaryVidStream);
            if (SUCCEEDED(hr)) 
               {
               hr = pPrimaryVidStream->QueryInterface(IID_IDirectDrawMediaStream, (void **)&pDDStream);
               if (SUCCEEDED(hr)) 
                  {
                  hr = pDDStream->GetFormat(&movie_ddsd, NULL, NULL, NULL);
                  if (SUCCEEDED(hr)) 
                     {
                     hr = pDDStream->CreateSample(NULL, NULL, DDSFF_PROGRESSIVERENDER, &pSample);
                     if (SUCCEEDED(hr)) 
                        {
                        hr = pSample->GetSurface(&pSurface, &rect);
                        if (SUCCEEDED(hr)) 
                           {
                           hr = pMMStream->SetState(STREAMSTATE_RUN);
                           if (SUCCEEDED(hr)) 
                              {
                              cls.state = ca_active;
                           	cl.cinematictime = Sys_Milliseconds ();
                              cl.cinematicframe = 0;
                              }
                           }
                        }
                     }
                  }
               }
            }
         }
      }
   if ( !SUCCEEDED( hr ) )
      {
      char buf[128];
      AMGetErrorText( hr, buf, sizeof( buf )-1 );
      Com_Printf("SCR_PlayCinematic Error: %s\n", buf );
      }
   if ( cl.cinematictime <= 0 )
      {
      SCR_FinishCinematic();
		Com_Printf ("Unable to start cinematic: %s\n", arg);
      }
   }
}



